--::::::::::
--binfile.spc
--::::::::::
-- **************************************************
-- *                                                *
-- *  Binary_File                                   *  SPEC
-- *                                                *
-- **************************************************
with CS_Parts_Types;  -- for BYTE type
use  CS_Parts_Types;
package Binary_File is
--| Purpose
--| Binary_File provides a convenient mechanism for accessing
--| binary files, implemented as an abstract data type.  The
--| binary file may be read or written one byte at a time.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| 7/15/90  Rick Conn  Initial Design and Code

  type FILE_TYPE is limited private;

  type FILE_MODE is (IN_FILE, OUT_FILE);
  type BLOCK is array (INTEGER range <>) of BYTE;

  Data_Error,      -- full BLOCK could not be read
  Device_Error,    -- problem with underlying system
  End_Error,       -- read attempted into end of file
  Mode_Error,      -- read attempted from output file, etc.
  Name_Error,      -- invalid file/dir name
  Status_Error,    -- file already open
  Use_Error,       -- write to read/only file, others
  Unexpected_Error
    : exception;

  -- ...................................................
  -- .                                                 .
  -- .  Binary_File.Create                             .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Create (File : in out FILE_TYPE;
                    Name : in STRING);
  --| Purpose
  --| Create a binary file and open it for output.
  --|
  --| Exceptions
  --|   Device_Error   -- raised if file cannot be created
  --|                  -- due to a hardware error
  --|   Name_Error     -- raised if Name is not a valid file
  --|                  -- or directory reference
  --|   Status_Error   -- raised if file Name is already
  --|                  -- open
  --|   Use_Error      -- raised if file Name exists and is
  --|                  -- read/only
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  Binary_File.Open                               .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Open   (File : in out FILE_TYPE;
                    Name : in STRING);
  --| Purpose
  --| Open an existing binary file for input.
  --|
  --| Exceptions
  --|   Device_Error   -- raised if file cannot be opened
  --|                  -- due to a hardware error
  --|   Name_Error     -- raised if Name is not a valid file
  --|                  -- or directory reference
  --|   Status_Error   -- raised if file Name is already
  --|                  -- open
  --|   Use_Error      -- raised if file Name is write/only
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  Binary_File.Close                              .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Close (File : in out FILE_TYPE);
  --| Purpose
  --| Close the indicated file.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  Binary_File.Reset                              .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Reset (File : in out FILE_TYPE;
                   Mode : in FILE_MODE := IN_FILE);
  --| Purpose
  --| Close the indicated file and reopen it (at the
  --| beginning) for input or output.
  --|
  --| Exceptions
  --|   Device_Error   -- raised if file cannot be accessed
  --|                  -- due to a hardware error
  --|   Name_Error     -- raised if Name is not a valid file
  --|                  -- or directory reference
  --|   Status_Error   -- raised if file Name is already
  --|                  -- open
  --|   Use_Error      -- raised if file Name exists and is
  --|                  -- read/only
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  Binary_File.Mode                               .  SPEC
  -- .                                                 .
  -- ...................................................
  function Mode (File : in FILE_TYPE) return FILE_MODE;
  --| Purpose
  --| Return the mode (IN_FILE or OUT_FILE) of the
  --| indicated File.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  Binary_File.Name                               .  SPEC
  -- .                                                 .
  -- ...................................................
  function Name (File : in FILE_TYPE) return STRING;
  --| Purpose
  --| Return the name of the indicated File.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  Binary_File.Is_Open                            .  SPEC
  -- .                                                 .
  -- ...................................................
  function Is_Open (File : in FILE_TYPE) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff the indicated File is open.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  Binary_File.Is_End                             .  SPEC
  -- .                                                 .
  -- ...................................................
  function Is_End  (File : in FILE_TYPE) return BOOLEAN;
  --| Purpose
  --| Return TRUE if the next byte to be returned from
  --| the indicated File is beyond the end of the file.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  Binary_File.Read                               .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Read  (File : in FILE_TYPE;
                   Item : out BYTE);
  --| Purpose
  --| Read the next byte from an OPENed File.
  --|
  --| Exceptions
  --|   Device_Error   -- raised if File cannot be accessed
  --|                  -- due to a hardware error
  --|   End_Error      -- raised if the next byte to be
  --|                  -- returned is beyond the end of
  --|                  -- the File
  --|   Mode_Error     -- raised if File is opened for
  --|                  -- output (mode OUT_FILE)
  --|   Status_Error   -- raised if File has not been
  --|                  -- OPENed
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  Binary_File.Read                               .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Read  (File : in FILE_TYPE;
                   Item : out BLOCK);
  --| Purpose
  --| Read the next block from an OPENed File.
  --|
  --| Exceptions
  --|   Data_Error     -- raised if a full BLOCK could
  --|                  -- not be read from the file
  --|   Device_Error   -- raised if File cannot be accessed
  --|                  -- due to a hardware error
  --|   End_Error      -- raised if the next byte to be
  --|                  -- returned is beyond the end of
  --|                  -- the File
  --|   Mode_Error     -- raised if File is opened for
  --|                  -- output (mode OUT_FILE)
  --|   Status_Error   -- raised if File has not been
  --|                  -- OPENed
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  Binary_File.Write                              .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Write (File : in FILE_TYPE;
                   Item : in BYTE);
  --| Purpose
  --| Write the next byte to a CREATEed File.
  --|
  --| Exceptions
  --|   Device_Error   -- raised if File cannot be accessed
  --|                  -- due to a hardware error
  --|   Mode_Error     -- raised if File is opened for
  --|                  -- input (mode IN_FILE)
  --|   Status_Error   -- raised if File has not been
  --|                  -- CREATEd
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  Binary_File.Write                              .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Write (File : in FILE_TYPE;
                   Item : in BLOCK);
  --| Purpose
  --| Write the next block to a CREATEed File.
  --|
  --| Exceptions
  --|   Device_Error   -- raised if File cannot be accessed
  --|                  -- due to a hardware error
  --|   Mode_Error     -- raised if File is opened for
  --|                  -- input (mode IN_FILE)
  --|   Status_Error   -- raised if File has not been
  --|                  -- CREATEd
  --|
  --| Notes (none)

private
  type FILE_OBJECT;  -- deferred to body
  type FILE_TYPE is access FILE_OBJECT;
end Binary_File;
