--::::::::::
--bintree.spc
--::::::::::
-- ********************************************************
-- *                                                      *
-- *  BINARY_TREES_PKG                                    *  SPEC
-- *                                                      *
-- ********************************************************
generic
    type VALUE_TYPE is private;
    with function Difference(P, Q: VALUE_TYPE) return integer is <>;
        -- Must return a value > 0 if P > Q, 0 if P = Q, and less than
        -- zero otherwise.
package Binary_Trees_Pkg is
--| Purpose
--| This package is an efficient implementation of unbalanced binary trees.
--| These trees have the following properties:
--|
--|  1. Inserting a value is cheap (log n Differences per insertion).
--|  2. Finding a value is cheap (log n Differences per querey).
--|  3. Can iterate over the values in sorted order in linear time.
--|  4. Space overhead is moderate (2 "pointers" per value stored).
--|
--| They are thus useful both for sorting sequences of indeterminate size
--| and for lookup tables.
--| 
--| Initialization Exceptions (none)
--| Notes
--| The following example shows how to use this package where nodes in
--| the tree are labeled with a String_Type value (for which a natural
--| Difference function is not available).
--|-
--|     package SP renames String_Pkg;
--| 
--|     type my_Value is record
--|       label: SP.string_type;
--|       value: integer;
--|     end record;
--| 
--|     function differ_label(P, Q: SP.string_type) return integer is
--|     begin
--|       if SP."<"(P, Q) then return -1;
--|       elsif SP."<"(Q, P) then return 1;
--|       else return 0;
--|       end if;
--|     end differ_label;
--| 
--|     package my_Tree is new Binary_Trees_pkg(my_Value, differ_Label);
--| 
--| Note that the required Difference function may be easily written in terms
--| of "<" if that is available, but that frequently two comparisons must
--| be done for each Difference.  However, both comparisons would have
--| to be done internally by this package for every instantiation if the
--| generic parameter were "<" instead of Difference.
--| 
--| PERFORMANCE
--|
--| Every node can be visited in the tree in linear time.  The cost
--| of creating an iterator is small and independent of the size
--| of the tree.
--| 
--| Recognizing that comparing values can be expensive, this package
--| takes a Difference function as a generic parameter.  If it took
--| a comparison function such as "<", then two comparisons would be
--| made per node visited during a search of the tree.  Of course this
--| is more costly when "<" is a trivial operation, but in those cases,
--| Difference can be bound to "-" and the overhead in negligable.
--| 
--| Two different kinds of iterators are provided.  The first is the 
--| commonly used set of functions Make_Iter, More, and Next.  The second
--| is a generic procedure called Visit.  The generic parameter to Visit is
--| a procedure which is called once for each value in the tree.  Visit
--| is more difficult to use and results in code that is not quite as clear,
--| but its overhead is about 20% of the More/Next style iterator.  It
--| is therefore recommended for use only in time critical inner loops.
--|
--| Modifications
--| Author: Bill Toscano and Michael Gordon, Intermetrics, Inc.

-- Exceptions --

  Duplicate_Value: exception;
      -- Raised on attempt to insert a duplicate node into a tree.

  Not_Found: exception;
      -- Raised on attempt to find a node that is not in a tree.

  No_More: exception;
      -- Raised on attempt to bump an iterator that has already scanned the
      -- entire tree.

  Out_Of_Order: exception;
      -- Raised if a problem in the ordering of a tree is detected.

  Invalid_Tree: exception;
      -- Value is not a tree or was not properly initialized.

-- Types --

  type SCAN_KIND is (INORDER, PREORDER, POSTORDER);
  -- Used to specify the order in which values should be scanned from a tree:
  --
  -- inorder: Left, Node, Right (nodes visited in increasing order)
  -- preorder: Node, Left, Right (top down)
  -- postorder: Left, Right, Node (bottom up)

  type TREE is private;
  type ITERATOR is private;

-- Operations --

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.CREATE                          .  SPEC
  -- .                                                   .
  -- .....................................................
  Function Create return TREE;
  --| Purpose
  --| Create and return an empty tree.  Note that this allocates
  --| a small amount of storage which can only be reclaimed through 
  --| a call to Destroy.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.INSERT                          .  SPEC
  -- .                                                   .
  -- .....................................................
  Procedure Insert (V: VALUE_TYPE;
                    T: TREE);
  --| Purpose
  --| Insert V into T in the proper place.  If a value equal
  --| to V (according to the Difference function) is already contained
  --| in the tree, the exception Duplicate_Value is raised.
  --| Caution: Since this package does not attempt to balance trees as
  --| values are inserted, it is important to remember that inserting
  --| values in sorted order will create a degenerate tree, where search
  --| and insertion is proportional to the N instead of to Log N.  If
  --| this pattern is common, use the Balanced_Tree function below.
  --|
  --| Exceptions
  --|   Duplicate_Value
  --|   Invalid_Tree
  --|
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.INSERT_IF_NOT_FOUND             .  SPEC
  -- .                                                   .
  -- .....................................................
  procedure Insert_if_not_Found (V         : VALUE_TYPE;
                                 T         : TREE;
                                 Found     : out BOOLEAN;
                                 Duplicate : out VALUE_TYPE);
  --| Purpose
  --| Insert V into T in the proper place.  If a value equal
  --| to V (according to the Difference function) is already contained
  --| in the tree, Found will be True and Duplicate will be the duplicate
  --| value.  This might be a sequence of values with the same key, and
  --| V can then be added to the sequence.
  --|
  --| Exceptions
  --|   Invalid_Tree.
  --|
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.REPLACE_IF_FOUND                .  SPEC
  -- .                                                   .
  -- .....................................................
  procedure Replace_if_Found (V         : VALUE_TYPE;
                              T         : TREE;
                              Found     : out BOOLEAN;
                              Old_Value : out VALUE_TYPE);
  --| Purpose
  --| Search for V in T.  If found, replace the old value with V,
  --| and return Found => True, Old_Value => the old value.  Otherwise,
  --| simply insert V into T and return Found => False.
  --|
  --| Exceptions
  --|   Invalid_Tree.
  --|
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.DESTROY                         .  SPEC
  -- .                                                   .
  -- .....................................................
  procedure Destroy (T: in out TREE);
  --| Purpose
  --| The space allocated to T is reclaimed.  The space occupied by
  --| the values stored in T is not however, recovered.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.DESTROY_DEEP                    .  SPEC
  -- .                                                   .
  -- .....................................................
  generic
    with procedure Free_Value(V: in out VALUE_TYPE) is <>;
  procedure Destroy_Deep (T: in out TREE);
  --| Purpose
  --| The space allocated to T is reclaimed.  The values stored
  --| in T are reclaimed using Free_Value, and the tree nodes themselves
  --| are then reclaimed (in a single walk of the tree).
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.BALANCED_TREE                   .  SPEC
  -- .                                                   .
  -- .....................................................
  generic
    with function Next_Value return VALUE_TYPE is <>;
    -- Each call to this procedure should return the next value to be
    -- inserted into the balanced tree being created.  If necessary,
    -- this function should check that each value is greater than the
    -- previous one, and raise Out_of_Order if necessary.  If values
    -- are not returned in strictly increasing order, the results are
    -- unpredictable.
  Function Balanced_Tree (Count: NATURAL) return TREE;
  --| Purpose
  --| Create a balanced tree by calling next_Value Count times.
  --| Each time Next_Value is called, it must return a value that compares
  --| greater than the preceeding value.  This function is useful for balancing
  --| an existing tree (next_Value iterates over the unbalanced tree) or
  --| for creating a balanced tree when reading data from a file which is
  --| already sorted.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.COPY_TREE                       .  SPEC
  -- .                                                   .
  -- .....................................................
  generic
    with function Copy_Value(V: VALUE_TYPE) return VALUE_TYPE is <>;
    -- This function is called to copy a value from the old tree to the
    -- new tree.
  Function Copy_Tree (T: TREE) return TREE;
  --| Purpose
  --| Create a balanced tree that is a copy of the tree T.
  --| The exception Invalid_Tree is raised if T is not a valid tree.
  --|
  --| Exceptions
  --|   Invalid_Tree
  --|
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.IS_EMPTY                        .  SPEC
  -- .                                                   .
  -- .....................................................
  Function Is_Empty (T: TREE) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff T is an empty tree or if T was not initialized.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.FIND                            .  SPEC
  -- .                                                   .
  -- .....................................................
  Function Find (V: VALUE_TYPE;
                 T: TREE) return VALUE_TYPE;
  --| Purpose
  --| Search T for a value that matches V.  The matching value is
  --| returned.  If no matching value is found, the exception Not_Found
  --| is raised.
  --|
  --| Exceptions
  --|   Not_Found
  --|   Invalid_Tree
  --|
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.FIND                            .  SPEC
  -- .                                                   .
  -- .....................................................
  Procedure Find (V     : VALUE_TYPE;
                  T     : TREE;
                  Found : out BOOLEAN;
                  Match : out VALUE_TYPE);
  --| Purpose
  --| Search T for a value that matches V.  On return, if Found is
  --| TRUE then the matching value is returned in Match.  Otherwise, Found
  --| is FALSE and Match is undefined.
  --|
  --| Exceptions
  --|   Invalid_Tree;
  --|
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.IS_FOUND                        .  SPEC
  -- .                                                   .
  -- .....................................................
  function Is_Found (V: VALUE_TYPE;
                     T: TREE) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff V is found in T.
  --|
  --| Exceptions
  --|   Invalid_Tree
  --|
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.SIZE                            .  SPEC
  -- .                                                   .
  -- .....................................................
  function Size (T: TREE) return NATURAL; 
  --| Purpose
  --| Return the number of values stored in T.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.VISIT                           .  SPEC
  -- .                                                   .
  -- .....................................................
  generic
    with procedure Process(V: VALUE_TYPE) is <>;
  procedure Visit (T     : TREE;
                   Order : SCAN_KIND);
  --| Purpose
  --| Invoke Process(V) for each value V in T.  The nodes are visited
  --| in the order specified by Order.  Although more limited than using
  --| an iterator, this function is also much faster.
  --|
  --| Exceptions
  --|   Invalid_Tree
  --|
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.MAKE_ITER                       .  SPEC
  -- .                                                   .
  -- .....................................................
  function Make_Iter (T: TREE) return ITERATOR;
  --| Purpose
  --| Create an iterator over a tree.
  --|
  --| Exceptions
  --|  Invalid_Tree
  --|
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.MORE                            .  SPEC
  -- .                                                   .
  -- .....................................................
  function More (I: ITERATOR) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff unscanned nodes remain in the tree being
  --| scanned by I.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  BINARY_TREES_PKG.NEXT                            .  SPEC
  -- .                                                   .
  -- .....................................................
  procedure Next (I: in out ITERATOR;
                  V: out VALUE_TYPE);
  --| Purpose
  --| Return the next value in the tree being scanned by I.
  --| The exception No_More is raised if there are no more values to scan.
  --|
  --| Exceptions
  --|   No_More
  --|
  --| Notes (none)

private

  type NODE;
  type NODE_PTR is access NODE;

  type NODE is 
    record
      Value : VALUE_TYPE;
      Less  : NODE_PTR;
      More  : NODE_PTR;
    end record;

  type TREE_HEADER is 
    record
      Count : NATURAL := 0;
      Root  : NODE_PTR := Null;
    end record;

  type TREE is access TREE_HEADER;

  type ITER_STATE is (LEFT, MIDDLE, RIGHT, DONE);

  type ITERATOR_RECORD;
  type ITERATOR is access ITERATOR_RECORD;

  type ITERATOR_RECORD is
    record
      State   : ITER_STATE;
      Parent  : ITERATOR;
      Subtree : NODE_PTR;
    end record;

end Binary_Trees_Pkg;
