--::::::::::
--bintree2.spc
--::::::::::
-- ***********************************************
-- *                                             *
-- *  BINARYTREES                                *  SPEC
-- *                                             *
-- ***********************************************
with Lists;
generic
    type ITEMTYPE is private; 
    with function "<" (X,Y: in ITEMTYPE) return BOOLEAN;
package BinaryTrees is 
--| Purpose
--| This package creates an ordered binary tree.  This will allow for 
--| quick insertion, and search.  
--|
--| The tree is organized such that 
--|  
--|  leftchild < root    root < rightchild
--| 
--| This means that by doing a left to right search of the tree will can
--| produce the nodes of the tree in ascending order.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| Author: Bill Toscano and Michael Gordon, Intermetrics

  type TREE is  private;     -- This is the type exported to represent the
                             -- tree.


  type TREEITER is private;  -- This is the type which is used to iterate
                             -- over the set.

  -- .................................................
  -- .                                               .
  -- .  BINARYTREES.CREATE                           .  SPEC
  -- .                                               .
  -- .................................................
  function Create return TREE;
  --| Purpose
  --| This creates a tree containing no information and no children. 
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .................................................
  -- .                                               .
  -- .  BINARYTREES.DEPOSIT                          .  SPEC
  -- .                                               .
  -- .................................................
  procedure Deposit (I : in ITEMTYPE; S : in TREE);
  --| Purpose
  --| This changes the information stored at the root of the tree S.
  --| It deposits the information I in the root of S.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .................................................
  -- .                                               .
  -- .  BINARYTREES.DESTROYTREE                      .  SPEC
  -- .                                               .
  -- .................................................
  procedure DestroyTree (T  :in out TREE);
  --| Purpose
  --| Destroys a tree and returns the space which it is occupying.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .................................................
  -- .                                               .
  -- .  BINARYTREES.INSERTNODE                       .  SPEC
  -- .                                               .
  -- .................................................
  Procedure Insertnode(N      : In Out ITEMTYPE; 
                       T      : In Out TREE;
                       Root   : Out TREE; 
                       Exists : out BOOLEAN); 
  --| Purpose
  --| This adds the node N to the tree T inserting in the proper position.
  --| Root is the root of the subtree which Node N heads (the position
  --| of Node N in T).
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .................................................
  -- .                                               .
  -- .  BINARYTREES.MAKETREEITER                     .  SPEC
  -- .                                               .
  -- .................................................
  function MakeTreeIter (T : in TREE) return TREEITER;
  --| Purpose
  --| Sets a variable to a position in the tree where the iteration is
  --| to begin.  In this case, the position is a pointer to the deepest
  --| leftmost leaf in the tree.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .................................................
  -- .                                               .
  -- .  BINARYTREES.MORE                             .  SPEC
  -- .                                               .
  -- .................................................
  function More (I : in TREEITER) return BOOLEAN;
  --| Purpose
  --| Returns TRUE iff there are more elements in the tree
  --| over which to iterate.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .................................................
  -- .                                               .
  -- .  BINARYTREES.NEXT                             .  SPEC
  -- .                                               .
  -- .................................................
  procedure Next (I    : in out TREEITER;
                  Info : out ITEMTYPE);    
  --| Purpose
  --| This is the iterator operation.  Given an Iter in the Tree, it
  --| returns the item Iter points to and updates the Iter.  If Iter
  --| is at the end of the Tree, More will indicate such.
  --|
  --| Exceptions (none)
  --| Notes (none)

private
   type NODE;
   type TREE is access NODE;
   type NODE is 
        record
            Info           : ITEMTYPE;
            LeftChild      : TREE;
            RightChild     : TREE;
        end record;

   package NodeOrder is new Lists (TREE);

   type TREEITER is
      record
          NodeList : NodeOrder.LIST;
          State    : NodeOrder.LISTITER;
      end record;

end BinaryTrees;
