--::::::::::
--bplustre.spc
--::::::::::
-- **********************************************************
-- *                                                        *
-- *  BP_Tree                                               *  SPEC
-- *                                                        *
-- **********************************************************
generic
  type KEY_TYPE is limited private;
  type NON_KEY_ITEM_TYPE is limited private;
  type NON_KEY_ITEM_CONTAINER is limited private;
  with function Empty
      (This_Non_Key_Item_Container : in NON_KEY_ITEM_CONTAINER)
    return Boolean is <>;
  with procedure Assign
      (To_Non_Key_Item_Container   : in out NON_KEY_ITEM_CONTAINER;
       From_Non_Key_Item_Container : in     NON_KEY_ITEM_CONTAINER)
    is <>;
  with procedure Insert
      (Container    : in out NON_KEY_ITEM_CONTAINER;
       Non_Key_Item : in     NON_KEY_ITEM_TYPE      ) is <>;
  with procedure Delete
      (Container    : in out NON_KEY_ITEM_CONTAINER;
       Non_Key_Item : in     NON_KEY_ITEM_TYPE      ) is <>;    
  with procedure Destroy_Contents
      (This_Non_Key_Item_Container : in out NON_KEY_ITEM_CONTAINER)
    is <>;
    -- This procedure must destroy everything in the container
    -- in preparation for the destruction of the container itself.
  with procedure Assign (Target_Key : in out KEY_TYPE;
                         Source_Key : in     KEY_TYPE) is <>;
  with function Less_Than (First_Key  : in KEY_TYPE;                       
                           Second_Key : in KEY_TYPE)
    return Boolean is <>;
  with function Equal (First_Key  : in KEY_TYPE;
                       Second_Key : in KEY_TYPE)
    return Boolean is <>;
package BP_Tree is
--| Purpose
--| Implement a B+ Tree class of objects.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| Author: William Thomas Wolfe, Clemson University

-- *******************************************************
-- This software is part of the Clemson University
-- Computer Science Department's Ada Software
-- Repository, and is copyrighted (C) 1989 by
-- Clemson University.  Permission to copy without
-- fee all or part of this software is granted,
-- provided that the copies are not made or
-- distributed for direct commercial advantage, and
-- that this copyright notice is not deleted or
-- modified.  To copy otherwise, or to republish,
-- requires a fee and/or specific permission.
-- *******************************************************

  type B_PLUS_TREE is limited private;

  Key_Does_Not_Exist_In_This_B_Plus_Tree            : EXCEPTION;
  No_Preceding_Key_Exists_In_This_B_Plus_Tree       : EXCEPTION;
  No_Following_Key_Exists_In_This_B_Plus_Tree       : EXCEPTION;
  No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree : EXCEPTION;

  type POINTER_TO_B_PLUS_TREE is access B_PLUS_TREE;

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Destroy                                 .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Destroy
      (Targeted_Object : in out POINTER_TO_B_PLUS_TREE);
  --| Purpose
  --| Unlike Unchecked_Deallocation, this procedure will properly
  --| destroy the B_Plus_Tree pointed to.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Insert_Item                             .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Insert_Item
      (Targeted_B_Plus_Tree : in out B_PLUS_TREE;
       Key_Value            : in     KEY_TYPE;
       Non_Key_Information  : in     NON_KEY_ITEM_TYPE);
  --| Purpose
  --| Insert an element into the Targeted_B_Plus_Tree.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Delete_Item                             .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Delete_Item
      (Targeted_B_Plus_Tree : in out B_PLUS_TREE;
       Key_Value            : in     KEY_TYPE;
       Non_Key_Information  : in     NON_KEY_ITEM_TYPE);
  --| Purpose
  --| Remove an element from the Targeted_B_Plus_Tree.
  --|
  --| Exceptions
  --|   Key_Does_Not_Exist_In_This_B_Plus_Tree 
  --|   No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree
  --|
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Key_Exists                              .  SPEC
  -- .                                                  .
  -- ....................................................
  function Key_Exists (Targeted_B_Plus_Tree : in B_PLUS_TREE;
                       Search_Key           : in KEY_TYPE)
    return BOOLEAN;
  --| Purpose
  --| Return TRUE iff Search_Key is found in Targeted_B_Plus_Tree.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Number_Of_Keys_Stored                   .  SPEC
  -- .                                                  .
  -- ....................................................
  function Number_Of_Keys_Stored
      (Targeted_B_Plus_Tree : in B_PLUS_TREE)
    return NATURAL;
  --| Purpose
  --| Return the Number_of_Keys_Stored in Targeted_B_Plus_Tree.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Get_Item_Container                      .  SPEC
  -- .                                                  .
  -- ....................................................
  function Get_Item_Container
      (Targeted_B_Plus_Tree : in B_PLUS_TREE;
       Search_Key           : in KEY_TYPE    )
    return NON_KEY_ITEM_CONTAINER;
  --| Purpose
  --| Return the NON_KEY_ITEM_CONTAINER associated with the
  --| Search_Key in Targeted_B_Plus_Tree.
  --|
  --| Exceptions
  --|   Key_Does_Not_Exist_In_This_B_Plus_Tree 
  --|   No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree
  --|
  --| Notes
  --| This procedure involves copying the entire container.  If
  --| your NON_KEY_ITEM_TYPE is quite large, it may be advisable
  --| to implement it as a pointer to the "real" structures,
  --| thus reducing the copying burden per instance of the
  --| NON_KEY_ITEM_TYPE to that of a single pointer.

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Get_First_Key                           .  SPEC
  -- .                                                  .
  -- ....................................................
  function Get_First_Key (Targeted_B_Plus_Tree : in B_PLUS_TREE)
    return KEY_TYPE;
  --| Purpose
  --| Return the first KEY_TYPE in Targeted_B_Plus_Tree.
  --|
  --| Exceptions
  --|   No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree
  --|
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Get_Last_Key                            .  SPEC
  -- .                                                  .
  -- ....................................................
  function Get_Last_Key (Targeted_B_Plus_Tree : in B_PLUS_TREE)
    return KEY_TYPE;
  --| Purpose
  --| Return the last KEY_TYPE in Targeted_B_Plus_Tree.
  --|
  --| Exceptions
  --|   No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree
  --|
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.A_Preceding_Key_Exists                  .  SPEC
  -- .                                                  .
  -- ....................................................
  function A_Preceding_Key_Exists
      (Targeted_B_Plus_Tree : in B_PLUS_TREE;
       Search_Key           : in KEY_TYPE)
    return Boolean;
  --| Purpose
  --| Determine if a key exists in the Targeted_B_Plus_Tree
  --| before the Search_Key.
  --|
  --| Exceptions
  --|   Key_Does_Not_Exist_In_This_B_Plus_Tree 
  --|   No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree
  --|
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Get_Preceding_Key                       .  SPEC
  -- .                                                  .
  -- ....................................................
  function Get_Preceding_Key
      (Targeted_B_Plus_Tree : in B_PLUS_TREE;
       Search_Key           : in KEY_TYPE)
    return KEY_TYPE;
  --| Purpose
  --| Obtain the preceding key in the Targeted_B_Plus_Tree
  --| before the Search_Key.
  --|
  --| Exceptions
  --|   Key_Does_Not_Exist_In_This_B_Plus_Tree 
  --|   No_Preceding_Key_Exists_In_This_B_Plus_Tree
  --|   No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree
  --|
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.A_Following_Key_Exists                  .  SPEC
  -- .                                                  .
  -- ....................................................
  function A_Following_Key_Exists
      (Targeted_B_Plus_Tree : in B_PLUS_TREE;
       Search_Key           : in KEY_TYPE)
    return Boolean;
  --| Purpose
  --| Determine if a key exists in the Targeted_B_Plus_Tree after
  --| the Search_Key.
  --|
  --| Exceptions
  --|   Key_Does_Not_Exist_In_This_B_Plus_Tree 
  --|   No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree
  --|
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Get_Following_Key                       .  SPEC
  -- .                                                  .
  -- ....................................................
  function Get_Following_Key
      (Targeted_B_Plus_Tree : in B_PLUS_TREE;
       Search_Key           : in KEY_TYPE)
    return KEY_TYPE;
  --| Purpose
  --| Obtain the following key in the Targeted_B_Plus_Tree
  --| before the Search_Key.
  --|
  --| Exceptions
  --|   Key_Does_Not_Exist_In_This_B_Plus_Tree 
  --|   No_Following_Key_Exists_In_This_B_Plus_Tree
  --|   No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree
  --|
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Delete_Key                              .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Delete_Key
      (Targeted_B_Plus_Tree : in out B_PLUS_TREE;
       Search_Key           : in     KEY_TYPE);
  --| Purpose
  --| Remove a Search_Key from the Targeted_B_Plus_Tree.
  --|
  --| Exceptions
  --|   Key_Does_Not_Exist_In_This_B_Plus_Tree 
  --|   No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree
  --|
  --| Notes
  --| The NON_KEY_ITEM_CONTAINER associated with this key will
  --| be emptied via the Destroy_Contents procedure.

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Exchange                                .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Exchange (First_B_Plus_Tree  : in out B_PLUS_TREE;
                      Second_B_Plus_Tree : in out B_PLUS_TREE);
  --| Purpose
  --| Exchanges the values of First_B_PLUS_TREE and
  --| Second_B_PLUS_TREE in O(1) time.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Assign                                  .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Assign (To_B_Plus_Tree   : in out B_PLUS_TREE;
                    From_B_Plus_Tree : in     B_PLUS_TREE);
  --| Purpose
  --| Replaces the contents of To_B_Plus_Tree with
  --| From_B_Plus_Tree.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Destroy                                 .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Destroy (Targeted_B_Plus_Tree : in out B_PLUS_TREE);
  --| Purpose
  --| Destroys all keys and all associated containers
  --| and renders the tree Empty.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  BP_Tree.Destroy                                 .  SPEC
  -- .                                                  .
  -- ....................................................
  function Empty (Targeted_B_Plus_Tree: in B_PLUS_TREE)
    return Boolean;
  --| Purpose
  --| Determine if Targeted_B_Plus_Tree is empty.
  --|
  --| Exceptions (none)
  --| Notes (none)

private
   type B_PLUS_TREE_DESCRIPTOR;
   type B_PLUS_TREE is access B_PLUS_TREE_DESCRIPTOR;   
end BP_Tree;
