--::::::::::
--clp.spc
--::::::::::
-- **********************************************
-- *                                            *
-- * COMMAND_LINE_PROCESSOR (CLP)               * SPEC
-- *                                            *
-- **********************************************
package Command_Line_Processor is
--| Purpose
--|   COMMAND_LINE_PROCESSOR is an abstract state machine
--| that allows the user to access a command line, which
--| may contain file references which are include files,
--| as a simple list of file names which can be accessed
--| via an interator and a Get function.  The command line
--| syntax is:
--|
--|      command input_file input_file ... output_file
--| or:
--|      command input_file input_file ... input_file
--|
--| where any "input_file" may be prefixed by an "@"
--| to make it an include file.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--| Modifications
--|   2/19/90  Rick Conn  Initial Design and Code

  type COMMAND_LINE_LAYOUT is (ALL_INPUT_FILES,
                               ONE_OUTPUT_FILE);
  -- the command line either contains only input
  -- files or a group of input files and one
  -- output file

  -- ..............................................
  -- .                                            .
  -- . CLP.INITIALIZE                             . SPEC
  -- .                                            .
  -- ..............................................
  procedure Initialize (Program_Name : in STRING;
                        Command_Kind : in COMMAND_LINE_LAYOUT
                                       := ONE_OUTPUT_FILE);
  --| Purpose
  --| Initialize the package, specifying a program
  --| name which may be used by the Command Line
  --| Interface
  --|
  --| Exceptions
  --|   ALLOCATION_PROBLEM
  --|   INIT_ERROR
  --|
  --| Notes
  --|   CALL INITIALIZE ONLY ONCE

  -- ..............................................
  -- .                                            .
  -- . CLP.RESET                                  . SPEC
  -- .                                            .
  -- ..............................................
  procedure Reset;
  --| Purpose
  --| Reset the iterator.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . CLP.IS_END                                 . SPEC
  -- .                                            .
  -- ..............................................
  function Is_End return BOOLEAN;
  --| Purpose
  --| Return TRUE if no more file names are
  --| available.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . CLP.FILE_NAME                              . SPEC
  -- .                                            .
  -- ..............................................
  function File_Name return STRING;
  --| Purpose
  --| Return the name of the next file.
  --|
  --| Exceptions
  --|   END_OF_FILE_LIST
  --|
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . CLP.OUTPUT_FILE_NAME                       . SPEC
  -- .                                            .
  -- ..............................................
  function Output_File_Name return STRING;
  --| Purpose
  --| Return the name of the output file.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . CLP.FILE_NAME_COUNT                        . SPEC
  -- .                                            .
  -- ..............................................
  function File_Name_Count return NATURAL;
  --| Purpose
  --| Number of file names in command line.

  ALLOCATION_PROBLEM : exception;
  END_OF_FILE_LIST   : exception;
  INIT_ERROR         : exception;
  UNEXPECTED_ERROR   : exception;  -- raised anytime

end Command_Line_Processor;
