--::::::::::
--cset.spc
--::::::::::
-- ******************************************************
-- *                                                    *
-- *  Character_Set                                     *  SPEC
-- *                                                    *
-- ******************************************************
package Character_Set is
--| Purpose
--| These routines test for the following subsets of package
--| ASCII:
--| Routine              Subset tested for
--| =======              =================
--| ALPHA                'a'..'z' | 'A'..'Z'
--| ALPHA_NUMERIC        ALPHA | '0'..'9'
--| CONTROL              < ' ' | DEL
--| DIGIT                '0'..'9'
--| GRAPHIC              ' ' < ch < DEL (does not include space)
--| HEXADECIMAL          DIGIT | 'A'..'F' | 'a'..'f'
--| LOWER                'a'..'z'
--| PRINTABLE            GRAPHIC | ' '
--| PUNCTUATION          GRAPHIC and not ALPHA_NUMERIC
--| SPACE                HT | LF | VT | FF | CR | ' '
--| UPPER                'A'..'Z'
--|
--| Initialization Exceptions (none)
--| Notes
--|     Most of the "functions" are actually arrays indexed by
--| CHARACTER, so they are remarkably efficient.
--|    Not all MIL-HDBK-1804 PDL annotations are
--| used in this package due to its simplicity.
--|
--| Modifications
--| Author: Richard Conn, TI
--| Modified by: Joseph M. Orost, Concurrent Computer Corp

  use  ASCII;

  type BIT_ARRAY is array (CHARACTER) of BOOLEAN;
  pragma PACK (BIT_ARRAY);

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.Is_Alpha                         .  SPEC
  -- .                                                 .
  -- ...................................................
  Is_Alpha         : constant BIT_ARRAY :=
      BIT_ARRAY'('a' .. 'z' => TRUE,
                 'A' .. 'Z' => TRUE,
                 others => FALSE);

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.Is_Alpha_Numeric                 .  SPEC
  -- .                                                 .
  -- ...................................................
  Is_Alpha_Numeric : constant BIT_ARRAY :=
    BIT_ARRAY'('a' .. 'z' => TRUE,
               'A' .. 'Z' => TRUE,
               '0' .. '9' => TRUE,
               others => FALSE   );

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.Is_Control                       .  SPEC
  -- .                                                 .
  -- ...................................................
  Is_Control       : constant BIT_ARRAY :=
    BIT_ARRAY'(NUL .. US => TRUE,
               DEL => TRUE,
               others => FALSE);

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.Is_Digit                         .  SPEC
  -- .                                                 .
  -- ...................................................
  Is_Digit         : constant BIT_ARRAY :=
    BIT_ARRAY'('0' .. '9' => TRUE,
               others => FALSE);

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.Is_Graphic                       .  SPEC
  -- .                                                 .
  -- ...................................................
  Is_Graphic       : constant BIT_ARRAY :=
    BIT_ARRAY'('!' .. '~' => TRUE,
               others => FALSE);

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.Is_Hexadecimal                   .  SPEC
  -- .                                                 .
  -- ...................................................
  Is_Hexadecimal   : constant BIT_ARRAY :=
    BIT_ARRAY'('0' .. '9' => TRUE,
               'A' .. 'F' => TRUE,
               'a' .. 'f' => TRUE,
               others => FALSE   );

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.Is_Lower                         .  SPEC
  -- .                                                 .
  -- ...................................................
  Is_Lower         : constant BIT_ARRAY :=
    BIT_ARRAY'('a' .. 'z' => TRUE,
               others => FALSE);

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.Is_Printable                     .  SPEC
  -- .                                                 .
  -- ...................................................
  Is_Printable     : constant BIT_ARRAY :=
    BIT_ARRAY'(' ' .. '~' => TRUE,
               others => FALSE);

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.Is_Punctuation                   .  SPEC
  -- .                                                 .
  -- ...................................................
  Is_Punctuation   : constant BIT_ARRAY :=
    BIT_ARRAY'('!' .. '/' => TRUE,
               ':' .. '@' => TRUE,
               '[' .. '`' => TRUE,
               '{' .. '~' => TRUE,
               others => FALSE   );

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.Is_Space                         .  SPEC
  -- .                                                 .
  -- ...................................................
  Is_Space         : constant BIT_ARRAY :=
    BIT_ARRAY'(HT  => TRUE,
               LF  => TRUE,
               VT  => TRUE,
               FF  => TRUE,
               CR  => TRUE,
               ' ' => TRUE,
               others => FALSE);

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.Is_Upper                         .  SPEC
  -- .                                                 .
  -- ...................................................
  Is_Upper         : constant BIT_ARRAY :=
    BIT_ARRAY'('A' .. 'Z' => TRUE,
               others => FALSE);

  type TRANSLATION_ARRAY is array (CHARACTER) of CHARACTER;
  pragma PACK (TRANSLATION_ARRAY);

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.Lower                            .  SPEC
  -- .                                                 .
  -- ...................................................
  Lower            : constant TRANSLATION_ARRAY :=
  --| Notes
  --|   LOWER can be used in place of TO_LOWER (Ada won't
  --| allow overloading of an object and a procedure).
    (NUL, SOH, STX, ETX, EOT, ENQ, ACK, BEL, BS , HT ,
     LF , VT , FF , CR , SO , SI , DLE, DC1, DC2, DC3,
     DC4, NAK, SYN, ETB, CAN, EM , SUB, ESC, FS , GS ,
     RS , US , ' ', '!', '"', ASCII.SHARP, '$', '%', '&', ''',
     '(', ')', '*', '+', ',', '-', '.', '/', '0', '1',
     '2', '3', '4', '5', '6', '7', '8', '9', ':', ';',
     '<', '=', '>', '?', '@', 'a', 'b', 'c', 'd', 'e',
     'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
     'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y',
     'z', '[', '\', ']', '^', '_', '`', 'a', 'b', 'c',
     'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm',
     'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w',
     'x', 'y', 'z', '{', '|', '}', '~', DEL);

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.To_Lower                         .  SPEC
  -- .                                                 .
  -- ...................................................
  function  To_Lower (Ch  : in CHARACTER) return CHARACTER;
  procedure To_Lower (Ch  : in out CHARACTER);
  procedure To_Lower (Str : in out STRING);

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.Upper                            .  SPEC
  -- .                                                 .
  -- ...................................................
  Upper            : constant TRANSLATION_ARRAY :=
  --| Notes
  --|   UPPER can be used in place of TO_UPPER.
  --|
    (NUL, SOH, STX, ETX, EOT, ENQ, ACK, BEL, BS , HT ,
     LF , VT , FF , CR , SO , SI , DLE, DC1, DC2, DC3,
     DC4, NAK, SYN, ETB, CAN, EM , SUB, ESC, FS , GS ,
     RS , US , ' ', '!', '"', ASCII.SHARP, '$', '%', '&', ''',
     '(', ')', '*', '+', ',', '-', '.', '/', '0', '1',
     '2', '3', '4', '5', '6', '7', '8', '9', ':', ';',
     '<', '=', '>', '?', '@', 'A', 'B', 'C', 'D', 'E',
     'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
     'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y',
     'Z', '[', '\', ']', '^', '_', '`', 'A', 'B', 'C',
     'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
     'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W',
     'X', 'Y', 'Z', '{', '|', '}', '~', DEL);

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.To_Upper                         .  SPEC
  -- .                                                 .
  -- ...................................................
  function  To_Upper (Ch  : in CHARACTER) return CHARACTER;
  procedure To_Upper (Ch  : in out CHARACTER);
  procedure To_Upper (Str : in out STRING);

  subtype CONTROL_CHARACTER_NAME_2 is STRING (1 .. 2);
  subtype CONTROL_CHARACTER_NAME_3 is STRING (1 .. 3);

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.CC_Name_2                        .  SPEC
  -- .                                                 .
  -- ...................................................
  function  CC_Name_2 (Ch : CHARACTER)
    return CONTROL_CHARACTER_NAME_2;
  --| Purpose
  --| Returns Ch as a two-character string.  If Ch is a control
  --| character, the string contains a caret (^) followed by
  --| the control character letter (e.g., ^H for ASCII.BS).
  --| If Ch is not a control character, the string contains a
  --| leading space and the character.

  -- ...................................................
  -- .                                                 .
  -- .  Character_Set.CC_Name_3                        .  SPEC
  -- .                                                 .
  -- ...................................................
  function  CC_Name_3 (Ch : CHARACTER)
    return CONTROL_CHARACTER_NAME_3;
  --| Purpose
  --| Returns Ch as a three-character string.  If Ch is a control
  --| character, the string contains the name given in ASCII (e.g.,
  --| "BS " for ^H).  If Ch is not a control character, the string
  --| contains two leading spaces and the character.

end Character_Set;
