--::::::::::
--cstrings.spc
--::::::::::
-- *********************************************************
-- *                                                       *
-- *  CStrings                                             *  SPEC
-- *                                                       *
-- *********************************************************
generic
  Max_String_Length : NATURAL := 400;  -- max length of a string
                                       -- including the trailing
                                       -- ASCII.NUL character
package CStrings is
--| Purpose
--|   CStrings provides a number of procedures and functions
--| which manipulate null-terminated strings (called C Strings)
--| and Ada strings (which contain no null character).
--| Type STRING is used to contain the C and Ada strings.
--| A C string contains a sequence of characters followed
--| by an ASCII.NUL; more characters may follow the ASCII.NUL
--| in the buffer, but they are ignored.  An Ada string is
--| a sequence of characters bound by the dimensions of the
--| buffer; all characters in the buffer are a part of the
--| string.
--|   The names of these procedures and functions were taken
--| from a listing of string-oriented C library functions.
--| The functionality of these routines is almost always
--| identical to the functionality of the original C routines.
--|
--| Initialization Exceptions (none)
--| Notes
--|   Reference Sun Release 4.0 man pages on "strings".
--|   Each string referenced in this specification is followed
--| by one of the following comments:
--|
--|    Comment      Meaning
--|    ===========  =========================================
--|    -- Ada       The string is an Ada String
--|    -- C         The string is a C String
--|    -- Ada or C  The string is an Ada String or a C String
--|
--| Modifications        Author: Richard Conn, MACA
--| 2/27/90  Richard Conn    Initial Version and Release

  type COMPARISON_RESULT is (LESS_THAN, EQUAL_TO, GREATER_THAN);

  -- Exceptions
  LENGTH_ERROR     : exception;  -- resulting string length
                                 -- is too long for buffer

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Make_Cstring                          .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Make_Cstring (From  : in STRING;   -- Ada or C
                          To    : out STRING); -- C
  --| Purpose
  --|   Place a copy of From into To.  Place
  --| the null terminator (ASCII.NUL) at the character
  --| in To corresponding to From(From'LAST+1).
  --|
  --| Exceptions
  --|   LENGTH_ERROR -- Destination is too short to hold
  --|                   the result or the result exceeds
  --|                   Max_String_Length characters

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Make_Cstring                          .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Make_Cstring (From_To : in out STRING;  -- Ada or C
                          Index   : in NATURAL);
  --| Purpose
  --|   Place a null into From_To on the indicated
  --| character.
  --|
  --| Exceptions
  --|   LENGTH_ERROR -- Index is out of bounds

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Ada_String                            .  SPEC
  -- .                                                 .
  -- ...................................................
  function Ada_String (From : in STRING) -- Ada or C
    return STRING;                       -- Ada
  --| Purpose
  --|   Return the slice of From up to but not including
  --| the ending NUL.  If From is an Ada string (no null),
  --| then the entire string is returned.
  --|
  --| Exceptions (none)

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strcat                                .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Strcat (To   : in out STRING;   -- C
                    From : in STRING);      -- Ada or C
  function Strcat (From_Part_1 : in STRING; -- Ada or C
                   From_Part_2 : in STRING) -- Ada or C
    return STRING;                          -- C
  --| Purpose
  --|   Strcat appends a copy of string Source to the end
  --| of string Destination.  The procedure Strcat modifies
  --| the string Destination, while the function Strcat
  --| does not modify the string Destination.
  --|
  --| Exceptions
  --|   LENGTH_ERROR -- Destination is too short to hold
  --|                   the result or the result exceeds
  --|                   Max_String_Length characters

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strncat                               .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Strncat (To     : in out STRING; -- C
                     From   : in STRING;     -- Ada or C
                     Length : in NATURAL);
  function Strncat  (To     : in STRING;     -- Ada or C
                     From   : in STRING;     -- Ada or C
                     Length : in NATURAL)
    return STRING;                           -- C
  --| Purpose
  --|   Strncat appends a copy of string From to the end
  --| of string To.  The procedure Strncat modifies
  --| the string To, while the function Strncat
  --| does not modify the string To.  At most Length
  --| characters are appended.
  --|
  --| Exceptions
  --|   LENGTH_ERROR -- Destination is too short to hold
  --|                   the result or the result exceeds
  --|                   Max_String_Length characters

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strcmp                                .  SPEC
  -- .                                                 .
  -- ...................................................
  function Strcmp (String1 : in STRING;  -- Ada or C
                   String2 : in STRING)  -- Ada or C
    return COMPARISON_RESULT;
  --| Purpose
  --|   Strcmp compares its arguments and returns the values
  --| LESS_THAN, EQUAL_TO, or GREATER_THAN accordingly as
  --| String1 is lexicographically less than, equal to, or
  --| greater than String2.
  --|
  --| Exceptions (none)

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strncmp                               .  SPEC
  -- .                                                 .
  -- ...................................................
  function Strncmp (String1 : in STRING;   -- Ada or C
                    String2 : in STRING;   -- Ada or C
                    Length  : in NATURAL)
    return COMPARISON_RESULT;
  --| Purpose
  --|   Strncmp compares its arguments and returns the values
  --| LESS_THAN, EQUAL_TO, or GREATER_THAN accordingly as
  --| String1 is lexicographically less than, equal to, or
  --| greater than String2.  Strncmp compares at most
  --| Length characters.
  --|
  --| Exceptions (none)

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strcasecmp                            .  SPEC
  -- .                                                 .
  -- ...................................................
  function Strcasecmp (String1 : in STRING;  -- Ada or C
                       String2 : in STRING)  -- Ada or C
    return COMPARISON_RESULT;
  --| Purpose
  --|   Strcasecmp compares its arguments and returns the values
  --| LESS_THAN, EQUAL_TO, or GREATER_THAN accordingly as
  --| String1 is lexicographically less than, equal to, or
  --| greater than String2.  Differences in case are ignored.
  --|
  --| Exceptions (none)

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strncasecmp                           .  SPEC
  -- .                                                 .
  -- ...................................................
  function Strncasecmp (String1 : in STRING;   -- Ada or C
                        String2 : in STRING;   -- Ada or C
                        Length  : in NATURAL)
    return COMPARISON_RESULT;
  --| Purpose
  --|   Strncasecmp compares its arguments and returns the values
  --| LESS_THAN, EQUAL_TO, or GREATER_THAN accordingly as
  --| String1 is lexicographically less than, equal to, or
  --| greater than String2.  Differences in case are ignored.
  --| At most Length characters are compared.
  --|
  --| Exceptions (none)

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strcpy                                .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Strcpy (From : in STRING;    -- Ada or C
                    To   : out STRING);  -- C
  --| Purpose
  --|   Strcpy copies From to To, stopping after
  --| the null character has been copied.
  --|
  --| Exceptions
  --|   LENGTH_ERROR -- Destination is too short to hold
  --|                   the result

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strncpy                               .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Strncpy (From   : in STRING;    -- Ada or C
                     To     : out STRING;   -- C
                     Length : in NATURAL);
  --| Purpose
  --|   Strncpy copies From to To, copying
  --| at most Length characters.  If there are more
  --| than Length characters in To, Length
  --| characters will be copied and a trailing null
  --| appended after the last character.
  --|
  --| Exceptions
  --|   LENGTH_ERROR -- Destination is too short to hold
  --|                   the result

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strlen                                .  SPEC
  -- .                                                 .
  -- ...................................................
  function Strlen (String1 : in STRING)  -- Ada or C
    return NATURAL;
  pragma inline (Strlen);
  --| Purpose
  --|   Strlen returns the number of characters in String1,
  --| not including the null-terminating character.
  --|
  --| Exceptions (none)

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strchr                                .  SPEC
  -- .                                                 .
  -- ...................................................
  function Strchr (String1 : in STRING;     -- Ada or C
                   Char1   : in CHARACTER)
    return NATURAL;
  --| Purpose
  --|   Strchr returns the index of the first occurrence
  --| of Char1 in the string String1 or the value 0 if
  --| Char1 does not occur in String1.  The null-terminating
  --| character is considered to be part of String1.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --|   This function is identical to the index and strchr
  --| functions in C.

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strrchr                               .  SPEC
  -- .                                                 .
  -- ...................................................
  function Strrchr (String1 : in STRING;    -- Ada or C
                    Char1   : in CHARACTER)
    return NATURAL;
  --| Purpose
  --|   Strrchr returns the index of the last occurrence
  --| of Char1 in the string String1 or the value 0 if
  --| Char1 does not occur in String1.  The null-terminating
  --| character is considered to be part of String1.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --|   This function is identical to the rindex and strrchr
  --| functions in C.

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strpbrk                               .  SPEC
  -- .                                                 .
  -- ...................................................
  function Strpbrk (String1 : in STRING;  -- Ada or C
                    String2 : in STRING)  -- Ada or C
    return NATURAL;
  --| Purpose
  --|   Strpbrk returns the index of the first occurrence in
  --| String1 of any character from String2 or the value 0 if
  --| no character from String2 exists in String1.
  --|
  --| Exceptions (none)

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strspn                                .  SPEC
  -- .                                                 .
  -- ...................................................
  function Strspn (String1 : in STRING;  -- Ada or C
                   String2 : in STRING)  -- Ada or C
    return NATURAL;
  --| Purpose
  --|   Strspn returns the length of the initial segment
  --| of String1 which consists entirely of characters
  --| from String2.
  --|
  --| Exceptions (none)

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strcspn                               .  SPEC
  -- .                                                 .
  -- ...................................................
  function Strcspn (String1 : in STRING;  -- Ada or C
                    String2 : in STRING)  -- Ada or C
    return NATURAL;
  --| Purpose
  --|   Strcspn returns the length of the initial segment
  --| of String1 which consists entirely of characters
  --| not from String2.
  --|
  --| Exceptions (none)

  -- ...................................................
  -- .                                                 .
  -- .  CStrings.Strtok                                .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Strtok (Target     : in STRING;      -- Ada or C
                    Start      : in out NATURAL;
                    Delimiters : in STRING;      -- Ada or C
                    Next_Token : out STRING);    -- C
  --| Purpose
  --|   Strtok considers the string Target to consist of a
  --| sequence of zero or more text tokens separated by spans
  --| of one or more characters from the separator string
  --| Delimiters.  A call to Strtok returns the first token
  --| in Target on or after the character indexed by Start.
  --| This token is returned in the string Next_Token with
  --| a null character immediately following the token.
  --| The separator string Delimiters may be different from
  --| call to call.
  --|
  --| Strtok must be called with Start's actual parameter
  --| being an initialized variable; generally, Start's
  --| initial value is Target'FIRST.
  --|
  --| Exceptions
  --|   LENGTH_ERROR -- Next_Token is too short to hold
  --|                   the result

end CStrings;
