--::::::::::
--darray.spc
--::::::::::
-- **************************************************
-- *                                                *
-- *  DARRAY_PKG                                    *  SPEC
-- *                                                *
-- **************************************************
generic
    type ELEM_TYPE is private;
    with function Equal (E1, E2: ELEM_TYPE)
        return BOOLEAN is "=";
package Darray_Pkg is
--| Purpose
--| This package provides the dynamic array (darray) abstract data type.
--| A darray has completely dynamic bounds, which change during runtime as
--| elements are added to/removed from the top/bottom. darrays are similar
--| to deques, differing only in that operations for indexing into the
--| structure are also provided.  A darray is indexed by integers that
--| fall within the current bounds.  The component type, elem_type, of a
--| darray is a generic formal parameter of this package, along with a
--| function, equal, that is assumed to form an equality relation over
--| over elem_type.
--|
--| The notation, <first, elts>, will be used to denote a darray.
--| first is the current low bound of the darray.  elts is the sequence
--| of elements contained in the darray.  For a given darray, d, the
--| dot selection mechanism is used to refer to these components, e.g.,
--| d.first and d.elts.  & is used for sequence concatenation, and also
--| for prepending/postpending a single element to a sequence.  |s| is
--| the number of elements in a sequence, s, and () is the null sequence.
--| Standard Ada array indexing notation is adopted for sequences.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| Programmer: Ron Kownacki, Intermetrics

-- Primary Types
  type DARRAY is private;
  type ARRAY_TYPE is array (INTEGER range <>) of ELEM_TYPE;

-- Storage Management Constants and Types  (see create procedure)
  Default_Predict : constant POSITIVE := 100;
  Default_High    : constant POSITIVE := 50;
  Default_Expand  : constant POSITIVE := 100;

-- Exceptions
  No_More       : exception;  -- Raised on incorrect use of an iterator.
  Out_of_Bounds : exception;  -- Raised on index out of current bounds.
  Uninitialized_Darray : exception;
        -- Raised on use of uninitialized darray by most operations.

-- Iterators
  type ELEMENTS_ITER is private;

-- Constructors

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.CREATE                              .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Create(First:          in     INTEGER := 1;
                   Predict:        in     POSITIVE := Default_Predict;
                   High_Percent:   in     POSITIVE := Default_High;
                   Expand_Percent: in     POSITIVE := Default_Expand;
                   D:              in out DARRAY);
  --| Purpose
  --| Sets d to <first, ()>.  If d has previously been initialized,
  --| then a destroy(d) is first performed.  The predict parameter
  --| specifies the initial space allocated.  (predict  = #elements).
  --| The high_percent parameter is the caller's expectation of the
  --| percentage of add_highs, out of total adds, to the darray.  For
  --| example, a caller would specify 100 if it was known that no
  --| add_lows would be performed.  The expand_percent parameter
  --| specifies the amount of additional space, as a percentage of
  --| currently allocated space, that is to be allocated whenever an
  --| expansion becomes necessary.  For example, 100 doubles the
  --| allocated space.
  --|
  --| Exceptions (none)
  --| Notes (none)
  
  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.ARRAY_TO_DARRAY                     .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Array_to_Darray(A:              in     ARRAY_TYPE;
                            First:          in     INTEGER := 1;
                            Predict:        in     POSITIVE;
                            High_Percent:   in     POSITIVE
                                                     := Default_High;
                            Expand_Percent: in     POSITIVE
                                                     := Default_Expand;
                            D:              in out DARRAY);
  --| Purpose
  --| Sets d to <first, a(a'first..a'last)>.  If d has previously
  --| been initialized, then an implicit destroy(d) is performed.
  --| The high_percent and expand_percent parameters are defined
  --| as for create.  Raises out_of_bounds iff predict < a'length.
  --|
  --| Exceptions
  --|   out_of_bounds
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.SET_FIRST                           .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Set_First(D:     in out DARRAY;
                      First: in     INTEGER);
  --| Purpose
  --| Sets d.first to first.
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.ADD_LOW                             .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Add_Low (D: in out DARRAY;
                     E: in     ELEM_TYPE);
  --| Purpose
  --| Sets d to <d.first - 1, e & d.elts>.
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.ADD_HIGH                            .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Add_High (D: in out DARRAY;
                      E: in     ELEM_TYPE);
  --| Purpose
  --| Sets d.elts to d.elts & e.
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.REMOVE_LOW                          .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Remove_Low (D: in out DARRAY);
  --| Purpose
  --| Sets d to <d.first + 1, d.elts(d.first + 1 .. last(d))>.
  --| Raises out_of_bounds iff is_empty(d).
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|   out_of_bounds
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.REMOVE_HIGH                         .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Remove_High (D: in out DARRAY);
  --| Purpose
  --| Sets d.elts to d.elts(d.first..last(d) - 1).
  --| Raises out_of_bounds iff is_empty(d).
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|   out_of_bounds
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.STORE                               .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Store (D: in out DARRAY;
                   I: in     INTEGER;
                   E: in     ELEM_TYPE);
  --| Purpose
  --| Replaces d.elts(i) with e.  Raises out_of_bounds iff
  --| either is_empty(d) or i is not in d.first..last(d).
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|   out_of_bounds
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.COPY                                .  SPEC
  -- .                                                 .
  -- ...................................................
  function Copy (D: DARRAY) return DARRAY;
  --| Purpose
  --| Returns a copy of d.  Subsequent changes to the structure of d
  --| will not be visible through the application of operations to
  --| the copy of d, and vice versa.  Assignment or parameter passing
  --| without using copy (or copy_deep, described below) will result
  --| in a single darray value being shared among objects.
  --| Raises uninitialized_darray if d has not been initialized.
  --| The assignment operation is used to transfer the values of
  --| the elem_type component objects of d; consequently, changes
  --| in these values may be observable through both darrays if
  --| elem_type is an access type, or contains access type
  --| components.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|
  --| Notes (none)
      
  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.COPY_DEEP                           .  SPEC
  -- .                                                 .
  -- ...................................................
  generic
    with function Copy (E: ELEM_TYPE) return ELEM_TYPE;
  function Copy_Deep (D: DARRAY) return DARRAY;
  --| Purpose
  --| Returns a copy of d.  Subsequent changes to the structure of d
  --| will not be visible through the application of operations to
  --| the copy of d, and vice versa.  Assignment or parameter passing
  --| without using copy_deep or copy will result in a single
  --| darray value being shared among objects.
  --| Raises uninitialized_darray if d has not been initialized.
  --| The transfer of elem_type component objects is accomplished by
  --| using the assignment operation in conjunction with the copy
  --| function.  Consequently, the user can prevent sharing of
  --| elem_type access components.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|
  --| Notes (none)
    
-- Query Operations

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.FETCH                               .  SPEC
  -- .                                                 .
  -- ...................................................
  function Fetch (D: DARRAY; I: INTEGER) return ELEM_TYPE;
  --| Purpose
  --| Returns d.elts(i).  Raises out_of_bounds iff either is_empty(d)
  --| or i is not in d.first..last(d).
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   out_of_bounds
  --|   uninitialized_darray
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.LOW                                 .  SPEC
  -- .                                                 .
  -- ...................................................
  function Low (D: in DARRAY) return ELEM_TYPE;
  --| Purpose
  --| Returns d.elts(d.first).  Raises out_of_bounds iff is_empty(d).
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   out_of_bounds
  --|   uninitialized_darray
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.HIGH                                .  SPEC
  -- .                                                 .
  -- ...................................................
  function High (D: in DARRAY) return ELEM_TYPE;
  --| Purpose
  --| Returns d.elts(last(d)).  Raises out_of_bounds iff is_empty(d).
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   out_of_bounds
  --|   uninitialized_darray
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.FIRST                               .  SPEC
  -- .                                                 .
  -- ...................................................
  function First (D: in DARRAY) return INTEGER;
  --| Purpose
  --| Returns d.first.
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.LAST                                .  SPEC
  -- .                                                 .
  -- ...................................................
  function Last (D: in DARRAY) return INTEGER;
  --| Purpose
  --| Returns d.first + |d.elts| - 1.
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.IS_EMPTY                            .  SPEC
  -- .                                                 .
  -- ...................................................
  function Is_Empty (D: in DARRAY) return BOOLEAN;
  --| Purpose
  --| Returns length(d) = 0, or equivalently, last(d) < d.first.
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.LENGTH                              .  SPEC
  -- .                                                 .
  -- ...................................................
  function Length (D: in DARRAY) return NATURAL;
  --| Purpose
  --| Returns |d.elts|.
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.EQUAL                               .  SPEC
  -- .                                                 .
  -- ...................................................
  function Equal (D1, D2: DARRAY) return BOOLEAN;
  --| Purpose
  --| Return (d1.first = d2.first and
  --|         last(d1) = last(d2) and
  --|         for each i in d1.first..last(d1),
  --|             equal(d1.elts(i), d2.elts(i)).
  --| Raises uninitialized_darray if either d1 or d2 has not been
  --| initialized.  Note that (d1 = d2) implies that equal(d1, d2)
  --| will always hold.  "=" is object equality, equal is state
  --| equality.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.DARRAY_TO_DARRAY                    .  SPEC
  -- .                                                 .
  -- ...................................................
  function Darray_to_Array (D: DARRAY) return ARRAY_TYPE;
  --| Purpose
  --| Let bounds_range be d.first..d.first + length(d) - 1.  If
  --| bounds_range is empty, then return an empty array with bounds
  --| of 1..0.  Otherwise, return bounds_range'(d.elts).
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|
  --| Notes (none)

-- Iterators

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.MAKE_ELEMENTS_ITER                  .  SPEC
  -- .                                                 .
  -- ...................................................
  function Make_Elements_Iter (D: DARRAY) return ELEMENTS_ITER;
  --| Purpose
  --| Create and return an elements itererator based on d.  This
  --| object can then be used in conjunction with the more function
  --| and the next procedure to iterate over the components of d.
  --| Raises uninitialized_darray if d has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_darray
  --|
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.MORE                                .  SPEC
  -- .                                                 .
  -- ...................................................
  function More (Iter: ELEMENTS_ITER) return BOOLEAN;
  --| Purpose
  --| Return true iff the elements iterator has not been exhausted.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.NEXT                                .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Next (Iter: in out ELEMENTS_ITER;
                  E:       out ELEM_TYPE);
  --| Purpose
  --| Let iter be based on the darray, d.  Successive calls of next
  --| will return, in e, successive elements of d.elts.  Each call
  --| updates the state of the elements iterator.  After all elements
  --| have been returned, an invocation of next will raise no_more.
  --| Requires:
  --| d must not be changed between the invocations of
  --| make_elements_iterator(d) and next.
  --|
  --| Exceptions
  --|   no_more
  --|
  --| Notes (none)

-- Heap Management

  -- ...................................................
  -- .                                                 .
  -- .  DARRAY_PKG.DESTROY                             .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Destroy (D: in out DARRAY);
  --| Purpose
  --| Return space consumed by the darray value associated with object
  --| d to the heap.  (If d is uninitialized, this operation does
  --| nothing.)  If other objects share the same darray value, then
  --| further use of these objects is erroneous.  Components of type
  --| elem_type, if they are access types, are not garbage collected.
  --| It is the user's responsibility to dispose of these objects.
  --| d is left in the uninitialized state.
  --|
  --| Exceptions (none)
  --| Notes (none)

private
    type ARRAY_PTR is access ARRAY_TYPE;
    type DARRAY_INFO is
        record
            First_Idx      : POSITIVE;
            Last_Idx       : NATURAL;
            First          : INTEGER;
            High_Percent   : POSITIVE;
            Expand_Percent : POSITIVE;
            Arr            : ARRAY_PTR := null;
        end record;
    type DARRAY is access DARRAY_INFO;

    -- Let r be an instance of the representation type.
    -- Representation Invariants:
    -- 1. r /= null, r.arr /= null (must be initialized to be valid.)
    -- 2. r.arr'first = 1 and
    --    r.arr'last >= 1
    -- 3. r.first_idx <= r.last_idx or
    --    r.first_idx = r.last_idx + 1
    -- 4. r.first_idx <= r.last_idx =>
    --        r.first_idx, r.last_idx in r.arr'range
    -- 5. r.expand_percent, r.high_percent get values at creation time,
    --    and these never change.
    --
    -- Abstraction Function:  (denoted by A(r))
    -- if r.last_idx < r.first_idx then
    --     <r.first, ()>
    -- else
    --     <r.first, (r.arr(r.first_idx),...,r.arr(r.last_idx))>
    --
    -- These properties follow:
    -- 1. length(A(r)) = r.last_idx - r.first_idx + 1
    -- 2. last(A(r)) = r.first + r.last_idx - r.first_idx
    -- 3. fetch(A(r), i) =
    --        if (i - r.first + r.first_idx) in r.first_idx..r.last_idx
    --            then r.arr(i - r.first + r.first_idx)
    --            else undefined.  (out_of_bounds)

    type ELEMENTS_ITER is
        record
            Last    : INTEGER := 0;
            Current : INTEGER := 1;
            Arr     : ARRAY_PTR;
        end record;

      -- Let d be the darray that an elements_iter, i, is based on.
      -- Initially, i.current = d.first_idx, i.last = d.last_idx, and
      -- i.arr = d.arr.
      -- more(i) = i.current <= i.last.
      -- next(i) = i.arr(current).  i.current incremented by next.
      -- Note that if an elements_iter object is not initialized, then
      -- more is false.

end Darray_Pkg;
