--::::::::::
--dlist.spc
--::::::::::
-- *****************************************************************
-- *                                                               *
-- *  DOUBLY_LINKED_LIST                                           *  SPEC
-- *                                                               *
-- *****************************************************************
generic
   type ELEMENT_OBJECT is private;
package Doubly_Linked_List is
--| Purpose   
--| DOUBLY_LINKED_LIST manipulates the abstract data type
--| LIST_ID, which is a linked list of objects.
--| DOUBLE_LIST provides routines to add objects to,
--| delete objects from, and extract objects from
--| the list.  DOUBLE_LIST also allows the user to
--| move about through the list and manipulate the
--| list in various ways.
--|
--| Initialization Exceptions (none)
--|
--| Notes
--| The number of list elements is restricted to 
--| INTEGER'LAST and the amount of memory or virtual
--| memory in the computer system.
--|
--| Modifications
--| Author: Richard Conn
   
-- Types
  type ELEMENT_POSITION is new INTEGER range 0 .. INTEGER'LAST;
  type LIST_ID is limited private;
   
-- Exceptions
  ADVANCE_PAST_END_OF_LIST          : exception;
  BACKUP_BEFORE_BEGINNING_OF_LIST   : exception;
  DYNAMIC_MEMORY_ALLOCATION_PROBLEM : exception;
  LIST_IS_EMPTY                     : exception;
  INVALID_INDEX                     : exception;
  UNEXPECTED_ERROR                  : exception;  -- raised anytime
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.INITIALIZE                            .  SPEC
  -- .                                                           .
  -- .............................................................
  procedure Initialize (ID : in out LIST_ID);
  --| Purpose
  --| Initialize the list to empty (the list is empty when
  --| first used); if the list contained any elements, they
  --| are deleted.
  --|
  --| Exceptions (none)
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.FIRST_ELEMENT                         .  SPEC
  -- .                                                           .
  -- .............................................................
  function First_Element (ID : in LIST_ID) return ELEMENT_OBJECT;
  --| Purpose
  --| Return the first element of the list.
  --|
  --| Exceptions
  --|   LIST_IS_EMPTY
  --|
  --| Notes (none)

  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.LAST_ELEMENT                          .  SPEC
  -- .                                                           .
  -- .............................................................
  function Last_Element (ID : in LIST_ID) return ELEMENT_OBJECT;
  --| Purpose
  --| Return the last element of the list.
  --|
  --| Exceptions
  --|   LIST_IS_EMPTY
  --|
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.CURRENT_ELEMENT                       .  SPEC
  -- .                                                           .
  -- .............................................................
  function Current_Element (ID : in LIST_ID) return ELEMENT_OBJECT;
  --| Purpose
  --| Return the current element of the list.
  --|
  --| Exceptions
  --|   LIST_IS_EMPTY
  --|
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.GOTO_FIRST                            .  SPEC
  -- .                                                           .
  -- .............................................................
  procedure Goto_First (ID : in out LIST_ID);
  --| Purpose
  --| Set the current element of the list to be the first
  --| element.
  --|
  --| Exceptions
  --|   LIST_IS_EMPTY
  --|
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.GOTO_LAST                             .  SPEC
  -- .                                                           .
  -- .............................................................
  procedure Goto_Last (ID : in out LIST_ID);
  --| Purpose
  --| Set the current element of the list to be the last
  --| element.
  --|
  --| Exceptions
  --|   LIST_IS_EMPTY
  --|
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.GOTO_ELEMENT                          .  SPEC
  -- .                                                           .
  -- .............................................................
  procedure Goto_Element (ID    : in out LIST_ID;
                          Index : in ELEMENT_POSITION);
  --| Purpose
  --| Set the current element of the list to be the Nth (INDEX)
  --| element.
  --|
  --| Exceptions
  --|   INVALID_INDEX
  --|   LIST_IS_EMPTY
  --|
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.CURRENT_INDEX                         .  SPEC
  -- .                                                           .
  -- .............................................................
  function Current_Index (ID : in LIST_ID) return ELEMENT_POSITION;
  --| Purpose
  --| Return the number of the current element.
  --|
  --| Exceptions
  --|   LIST_IS_EMPTY
  --|
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.LAST_INDEX                            .  SPEC
  -- .                                                           .
  -- .............................................................
  function Last_Index (ID : in LIST_ID) return ELEMENT_POSITION;
  --| Purpose
  --| Return the number of the last element.
  --|
  --| Exceptions
  --|   LIST_IS_EMPTY
  --|
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.ADVANCE                               .  SPEC
  -- .                                                           .
  -- .............................................................
  procedure Advance (ID : in out LIST_ID);
  --| Purpose
  --| Advance, setting the current element to be the next
  --| element.
  --|
  --| Exceptions
  --|   ADVANCE_PAST_END_OF_LIST
  --|   LIST_IS_EMPTY
  --|
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.BACKUP                                .  SPEC
  -- .                                                           .
  -- .............................................................
  procedure Backup (ID : in out LIST_ID);
  --| Purpose
  --| Backup, setting the current element to be the previous
  --| element.
  --|
  --| Exceptions
  --|   BACKUP_BEFORE_BEGINNING_OF_LIST
  --|   LIST_IS_EMPTY
  --|
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.IS_EMPTY                              .  SPEC
  -- .                                                           .
  -- .............................................................
  function Is_Empty (ID : in LIST_ID) return BOOLEAN;
  --| Purpose
  --| Return TRUE if the list is empty.
  --|
  --| Exceptions (none)
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.IS_END                                .  SPEC
  -- .                                                           .
  -- .............................................................
  function Is_End (ID : in LIST_ID) return BOOLEAN;
  --| Purpose
  --| Return TRUE if the end of the list has been passed.
  --|
  --| Exceptions (none)
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.IS_FIRST                              .  SPEC
  -- .                                                           .
  -- .............................................................
  function Is_First (ID : in LIST_ID) return BOOLEAN;
  --| Purpose
  --| Return TRUE if the current element is the first element.
  --|
  --| Exceptions (none)
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.APPEND_ELEMENT                        .  SPEC
  -- .                                                           .
  -- .............................................................
  procedure Append_Element (ID      : in out LIST_ID;
                            Element : ELEMENT_OBJECT);
  --| Purpose
  --| Append an element after the current element; set the current
  --| element to this new element.
  --|
  --| Exceptions
  --|   DYNAMIC_MEMORY_ALLOCATION_PROBLEM
  --|
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.INSERT_ELEMENT                        .  SPEC
  -- .                                                           .
  -- .............................................................
  procedure Insert_Element (ID      : in out LIST_ID;
                            Element : ELEMENT_OBJECT);
  --| Purpose
  --| Insert an element before the current element; the current
  --| element remains unchanged.
  --|
  --| Exceptions
  --|   DYNAMIC_MEMORY_ALLOCATION_PROBLEM
  --|
  --| Notes (none)
   
  -- .............................................................
  -- .                                                           .
  -- .  DOUBLY_LINKED_LIST.DELETE_ELEMENT                        .  SPEC
  -- .                                                           .
  -- .............................................................
  procedure Delete_Element (ID : in out LIST_ID);
  --| Purpose
  --| Delete the current element; the current element becomes the
  --| element following the current element.
  --|
  --| Exceptions
  --|   ADVANCE_PAST_END_OF_LIST
  --|   LIST_IS_EMPTY
  --|
  --| Notes (none)
   
private
   type ELEMENT;
   type ELEMENT_POINTER is access ELEMENT;
   type ELEMENT is 
      record
         Content  : ELEMENT_OBJECT;
         Next     : ELEMENT_POINTER;
         Previous : ELEMENT_POINTER;
      end record;
   type LIST_ID is 
      record
         First              : ELEMENT_POINTER  := null; -- first element
         Last               : ELEMENT_POINTER  := null; -- last element
         Current            : ELEMENT_POINTER  := null; -- current element
         Free               : ELEMENT_POINTER  := null; -- free element list
         Number_of_Elements : ELEMENT_POSITION := 0; -- number of elements
         Current_Index      : ELEMENT_POSITION := 0; -- index of current element
      end record;

end Doubly_Linked_List;
