--::::::::::
--dyn.spc
--::::::::::
-- *******************************************************
-- *                                                     *
-- *  DYN                                                *  SPEC
-- *                                                     *
-- *******************************************************
package Dyn is 
--| Purpose
--| Implement a dynamic string object class and provide operations
--| to manipulate objects of this class.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--|  This is a package of several string manipulation functions based on
--| a built-in dynamic STRING type DYN_STRING.  It is an adaptation and
--| extension of the package proposed by Sylvan Rubin of Ford Aerospace and
--| Communications Corporation in the Nov/Dec 1984 issue of the Journal of
--| Pascal, Ada and Modula-2.  Some new functions have been added, and much
--| of the body code has been rewritten.
--|
--| This package is derived from DSTR3.SRC in the Ada Software Repository
--| DSTR3.SRC was written by R.G. Cleaveland.  The derivation, done by
--| Richard Conn, was done to remove those general-purpose features of the
--| package not needed for the PTF project.

  Max_D_String_Length : constant POSITIVE := 100; 
    -- This is the maximum LENGTH of a dynamic string implemented with this
    -- package.  This value is "arbitrary" in that any reasonable number
    -- equal to or less than the maximum STRING LENGTH permitted by the
    -- compiler is acceptable.  The specific value above was chosen as a
    -- compromise between programmer convenience and memory space requirements.

  subtype DS_POS is INTEGER range 0..MAX_D_STRING_LENGTH;
  type DYN_STRING is private;

  STRING_TOO_SHORT: exception;

  -- ..................................................
  -- .                                                .
  -- .  DYN.D_STRING                                  .  SPEC
  -- .                                                .
  -- ..................................................
  function D_String (Char: CHARACTER)  return DYN_STRING;
  --| Purpose
  --| Creates a one-byte dynamic string of contents CHAR.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..................................................
  -- .                                                .
  -- .  DYN.D_STRING                                  .  SPEC
  -- .                                                .
  -- ..................................................
  function D_String (Str : STRING)  return DYN_STRING;
  --| Purpose
  --| Creates a dynamic string of contents STR.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..................................................
  -- .                                                .
  -- .  DYN.CHAR                                      .  SPEC
  -- .                                                .
  -- ..................................................
  function Char (Dstr  : DYN_STRING;
                 Posit : POSITIVE := 1) return CHARACTER;
  --| Purpose
  --| Return the Nth character of a dynamic string.
  --|
  --| Exceptions
  --|   STRING_TOO_SHORT
  --|
  --| Notes (none)

  -- ..................................................
  -- .                                                .
  -- .  DYN.STR                                       .  SPEC
  -- .                                                .
  -- ..................................................
  function Str (Dstr: DYN_STRING) return STRING;
  --| Purpose
  --| Return the string whose contents is the value of a dynamic
  --| string.
  --|
  --| Exceptions (none)
  --| Notes (none)
  
  -- ..................................................
  -- .                                                .
  -- .  DYN.LENGTH                                    .  SPEC
  -- .                                                .
  -- ..................................................
  function Length (Dstr: DYN_STRING)     return NATURAL;
  --| Purpose
  --| Returns the LENGTH of the dynamic string.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..................................................
  -- .                                                .
  -- .  DYN.CLEAR                                     .  SPEC
  -- .                                                .
  -- ..................................................
  procedure Clear (Dstr: in out DYN_STRING);
  --| Purpose
  --| Makes DSTR a null string.
  --|
  --| Exceptions (none)
  --| Notes (none)

private
  type DYN_STRING is
    record
      Size : INTEGER range 0..MAX_D_STRING_LENGTH;
      Data : STRING(1..MAX_D_STRING_LENGTH);
    end record;

end Dyn;
