--::::::::::
--fof.spc
--::::::::::
-- **********************************
-- *                                *
-- *  Formatted_Output_File (FOF)   *  SPEC
-- *                                *
-- **********************************
package Formatted_Output_File is
--| Purpose
--| Formatted_Output_File manipulates objects of type STRING (text),
--| placing text into the output file as it is received.
--| Formatted_Output_File is also used to define the format of the
--| text (number of lines per page, header, footer, etc.).
--|
--| Formatted_Output_File is a form of Report Generator.  Taking in
--| raw text and other directives (implemented by its procedures),
--| Formatted_Output_File creates reports (with header lines, footer
--| lines, page numbering, etc).
--|
--| Formatted_Output_File is also referred to as FOF.
--|
--| See the test programs for examples of the use of FOF.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| 04/22/90   Rick Conn    Initial version from PTF's FOF
--|                         package of 8/16/89

  type FILE is
    private;

  Maximum_Number_Of_Lines_On_Page
    : constant
      := 200;

  Maximum_Line_Length
    : constant
      := 200;

  Maximum_Number_Of_Header_Footer_Lines
    : constant
      := 8;

  Maximum_Number_Of_Pages
    : constant
      := 32000;


  type PAGE_ATTRIBUTE is
    ( TOP_MARGIN,       -- Number of lines before first header
      BOTTOM_MARGIN,    -- Number of lines after last footer
      LEFT_MARGIN,      -- Column num of the last col before the 1st char
      RIGHT_MARGIN,     -- Column number of the last char of the line
      LEFT_INDENT,      -- Number of columns to indent from LEFT_MARGIN
      RIGHT_INDENT,     -- Number of columns to indent from RIGHT_MARGIN
      TOTAL_LINES,      -- Number of lines on a page
      HEADER_LINES,     -- Number of lines in the header
      FOOTER_LINES,     -- Number of lines in the footer
      LINE_SPACING,     -- Number of blank lines after each text line
      PAGE_OFFSET,      -- Number of columns to offset each line
      TEMP_INDENT       -- Number of columns to indent next line only
                        -- (this is an absolute value, not influenced
                        --  by the LEFT_MARGIN or LEFT_INDENT settings)
    );

  type LINE_ATTRIBUTE is
    ( BOLD,                     -- Make words come out bold (overstrike)
      CENTER,                   -- Center lines (Put_Line with No Fill)
      FILL,                     -- Successively place words into an output
                                --  line until the next word will not fit
                                --  between the left and right margins
                                --  (with indents)
      FILL_STATE_BEFORE_CENTER, -- Save area for FILL
      JUSTIFY,                  -- Fill output line to RIGHT_MARGIN -
                                --  RIGHT_INDENT with spaces between words
      PAGING,                   -- Break output on page boundaries,
                                --  outputting footer, bottom margin,
                                --  top margin, and header
      UNDERLINE,                -- Underline words
      UNDERLINE_PUNCT,          -- If ON, underline punctuation
      USE_FORM_FEED             -- Use form feeds to eject pages
    );

  type PAGE_ATTRIBUTE_LIST is
    array (PAGE_ATTRIBUTE)
      of NATURAL;

  type OFF_ON is
    ( OFF, ON );

  type LINE_ATTRIBUTE_LIST is
    array (LINE_ATTRIBUTE)
      of OFF_ON;

  Page_Attribute_Defaults
    : constant PAGE_ATTRIBUTE_LIST
      := (
        TOP_MARGIN     => 4,
        BOTTOM_MARGIN  => 4,
        LEFT_MARGIN    => 12,
        RIGHT_MARGIN   => 90,
        LEFT_INDENT    => 0,
        RIGHT_INDENT   => 0,
        TOTAL_LINES    => 66,
        HEADER_LINES   => 2,
        FOOTER_LINES   => 2,
        LINE_SPACING   => 0,
        PAGE_OFFSET    => 0,
        TEMP_INDENT    => 0 );

  Line_Attribute_Defaults
    : constant LINE_ATTRIBUTE_LIST
      := (
        BOLD           => OFF,
        CENTER         => OFF,
        FILL           => ON,
        FILL_STATE_BEFORE_CENTER => ON,
        JUSTIFY        => ON,
        PAGING         => ON,
        UNDERLINE      => OFF,
        UNDERLINE_PUNCT => OFF,
        USE_FORM_FEED  => ON );


  Page_Number_Id_Default
    : constant CHARACTER
      := '#';

  type LINE_NUMBER is
    new INTEGER range 0 .. Maximum_Number_Of_Lines_On_Page;

  type HEADER_FOOTER_LINE is                     -- H/F line numbers
    new INTEGER range 1 .. Maximum_Number_Of_Header_Footer_Lines;

  type PAGE_NUMBER is
    new INTEGER range 0 .. Maximum_Number_Of_Pages;

  type STATUS is                                 -- for Open
    ( OK, NOT_OK );

  type PAGE_SIDE is                              -- for margins and indents
    ( LEFT_SIDE, RIGHT_SIDE );

  type PAGE_KIND is                              -- for headers and footers
    ( EVEN_PAGES, ODD_PAGES, ALL_PAGES );

  type NUMERIC_FORMAT is                         -- for page numbers
    ( ARABIC, LOWER_ROMAN, UPPER_ROMAN );

  Range_Error
    : exception;

  File_Not_Open
    : exception;

  -- ..................................
  -- .                                .
  -- .  FOF.Open                      .  SPEC
  -- .                                .
  -- ..................................
  procedure Open
    ( Item           : in out FILE;
      File_Name      : in STRING;
      Result         : out STATUS );
  --| Purpose
  --| Open the formatted output file for subsequent processing.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Close                     .  SPEC
  -- .                                .
  -- ..................................
  procedure Close
    ( Item           : in FILE );
  --| Purpose
  --| Close the formatted output file.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Put_Invisible_Word        .  SPEC
  -- .                                .
  -- ..................................
  procedure Put_Invisible_Word
    ( Item           : in FILE;
      What           : in STRING );
  --| Purpose
  --| Add a word to the current line and do not increment the
  --| character count.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Put_Word                  .  SPEC
  -- .                                .
  -- ..................................
  procedure Put_Word
    ( Item           : in FILE;
      What           : in STRING );
  --| Purpose
  --| Add a word to the current line.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Put_Line                  .  SPEC
  -- .                                .
  -- ..................................
  procedure Put_Line
    ( Item           : in FILE;
      What           : in STRING );
  --| Purpose
  --| Add a line to the current page.  If line break, insert blank
  --| lines as per LINE_SPACING.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Break_Line                .  SPEC
  -- .                                .
  -- ..................................
  procedure Break_Line
    ( Item           : in FILE );
  --| Purpose
  --| Break the current line (if it contains any words, output them).
  --| Insert blank lines as per the LINE_SPACING setting.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Current_Line              .  SPEC
  -- .                                .
  -- ..................................
  function Current_Line
    ( Item           : in FILE )
      return LINE_NUMBER;
  --| Purpose
  --| Return the number of the current line.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Skip                      .  SPEC
  -- .                                .
  -- ..................................
  procedure Skip
    ( Item           : in FILE;
      Number_Of_Lines : in LINE_NUMBER := 1 );
  --| Purpose
  --| Skip Number_Of_Lines in the output after first issuing a Break_Line.
  --| LINE_SPACING influences the actual number of lines skipped.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Break_Page                .  SPEC
  -- .                                .
  -- ..................................
  procedure Break_Page
    ( Item           : in FILE );
  --| Purpose
  --| If there is anything on the current page, output it and advance
  --| to the next page.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Break_Page                .  SPEC
  -- .                                .
  -- ..................................
  procedure Break_Page
    ( Item           : in FILE;
      New_Page_Num   : in PAGE_NUMBER );
  --| Purpose
  --| If there is anything on the current page, output it and advance
  --| to the next page.  Set the number of the next page to New_Page_Num.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Current_Page              .  SPEC
  -- .                                .
  -- ..................................
  function Current_Page
    ( Item           : in FILE )
      return PAGE_NUMBER;
  --| Purpose
  --| Return the number of the current page.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Current_Page              .  SPEC
  -- .                                .
  -- ..................................
  function Current_Page
    ( Item           : in FILE )
      return STRING;
  --| Purpose
  --| Return the number of the current page as a string.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Set_Page_Number_Format    .  SPEC
  -- .                                .
  -- ..................................
  procedure Set_Page_Number_Format
    ( Item           : in FILE;
      To             : in NUMERIC_FORMAT;
      Format_String  : in STRING );
  --| Purpose
  --| Set the format of the page number.  If the Format_String is not
  --| null, the page numbers in the headers and footers will appear as
  --| indicated (with the literal number substituted for # characters).
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Set_Page_Attribute        .  SPEC
  -- .                                .
  -- ..................................
  procedure Set_Page_Attribute
    ( Item           : in FILE;
      What           : in PAGE_ATTRIBUTE;
      To             : in NATURAL );
  --| Purpose
  --| Set a specified page attribute.
  --|
  --| Exceptions
  --| Range_Error    raised if To is outside the range for What
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Set_Line_Attribute        .  SPEC
  -- .                                .
  -- ..................................
  procedure Set_Line_Attribute
    ( Item           : in FILE;
      What           : in LINE_ATTRIBUTE;
      To             : in OFF_ON );
  --| Purpose
  --| Turn off or on the indicated attribute for the current and
  --| following lines.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Get_Page_Attribute        .  SPEC
  -- .                                .
  -- ..................................
  function Get_Page_Attribute
    ( Item           : in FILE;
      What           : in PAGE_ATTRIBUTE )
      return NATURAL;
  --| Purpose
  --| Get a specified page attribute.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Get_Line_Attribute        .  SPEC
  -- .                                .
  -- ..................................
  function Get_Line_Attribute
    ( Item           : in FILE;
      What           : in LINE_ATTRIBUTE )
      return OFF_ON;
  --| Purpose
  --| Get the indicated attribute for the current and
  --| following lines.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Test_Page                 .  SPEC
  -- .                                .
  -- ..................................
  function Test_Page
    ( Item           : in FILE;
      Number_Of_Lines : in LINE_NUMBER )
      return BOOLEAN;
  --| Purpose
  --| Return TRUE if Number_Of_Lines is remaining on the current page.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Set_Footer_Line           .  SPEC
  -- .                                .
  -- ..................................
  procedure Set_Footer_Line
    ( Item           : in FILE;
      Class          : in PAGE_KIND;
      Number         : in HEADER_FOOTER_LINE;
      Left_Text      : in STRING;
      Center_Text    : in STRING;
      Right_Text     : in STRING );
  --| Purpose
  --| Store a footer line for EVEN, ODD, or ALL pages.
  --| The footer line is dynamically adjusted, based on the left and right
  --| margin settings.  The strings Left, Center, and Right are left-
  --| justified, centered, and right-justified in the indicated footer
  --| line, respectively.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Set_Header_Line           .  SPEC
  -- .                                .
  -- ..................................
  procedure Set_Header_Line
    ( Item           : in FILE;
      Class          : in PAGE_KIND;
      Number         : in HEADER_FOOTER_LINE;
      Left_Text      : in STRING;
      Center_Text    : in STRING;
      Right_Text     : in STRING );
  --| Purpose
  --| Store a header line for EVEN, ODD, or ALL pages.
  --| The header line is dynamically adjusted, based on the left and right
  --| margin settings.  The strings Left, Center, and Right are left-
  --| justified, centered, and right-justified in the indicated header
  --| line, respectively.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Set_Page_Number_Id        .  SPEC
  -- .                                .
  -- ..................................
  procedure Set_Page_Number_Id
    ( Item           : in FILE;
      To             : in CHARACTER );
  --| Purpose
  --| Set the character used to represent the page number in the
  --| header and footer lines of the output file.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Set_Page_Number_Format    .  SPEC
  -- .                                .
  -- ..................................
  procedure Set_Page_Number_Format
    ( Item           : in FILE;
      To             : in NUMERIC_FORMAT );
  --| Purpose
  --| Set the format used to represent the page number in the
  --| header and footer lines of the output file.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  FOF.Page_Number_Format        .  SPEC
  -- .                                .
  -- ..................................
  function Page_Number_Format
    ( Item           : in FILE )
    return NUMERIC_FORMAT;
  --| Purpose
  --| Get the format used to represent the page number in the
  --| header and footer lines of the output file.
  --|
  --| Exceptions
  --| File_Not_Open
  --|
  --| Notes (none)

private -- Formatted_Output_File

  type FILE_OBJECT;
  type FILE is
    access FILE_OBJECT;

end Formatted_Output_File;
