--::::::::::
--hashmap.spc
--::::::::::
-- ******************************************************
-- *                                                    *
-- *  Hashed_Mapping_PKG                                *  SPEC
-- *                                                    *
-- ******************************************************
with lists;       -- Lists used in implementation.  (private)
pragma elaborate(lists);
generic
  type KEY_TYPE is private;
  with function Equal (K1, K2: KEY_TYPE) return BOOLEAN is "=";
  type BUCKET_RANGE is range <>;
    -- Defines the number of hash buckets, one for each member
    -- of BUCKET_RANGE.
  with function Hash (K: KEY_TYPE) return BUCKET_RANGE;
    -- Required property: equal(e1, e2) => hash(e1) = hash(e2).
    -- Best results if hash produces a uniform distribution
    -- over BUCKET_RANGE.
  type VALUE_TYPE is private;
package Hashed_Mapping_PKG is
--| Purpose
--| This package provides a mapping from one arbitrary type,
--| KEY_TYPE, to another arbitrary type, VALUE_TYPE.  These
--| types are generic formals to the package, along with an
--| equality relation on KEY_TYPE, an integer subtype that
--| determines the number of hash buckets, and a hashing
--| function on KEY_TYPE that maps to that integer subtype.
--|
--| For the purpose of specifying the operations in this
--| package, we will view a mapping as a set of bindings,
--| or key/value pairs.  This allows the use of set notation
--| in description.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| Author: Ron Kownacki, Intermetrics

  type MAPPING is private;

  No_More: exception;
    -- Raised on incorrect use of an iterator.
  Uninitialized_Mapping: exception;
    -- Raised on use of an unitialized MAPPING by most operations.
  Already_Bound: exception;
    -- Raised on attempt to rebind a key that is currently bound.
  Not_Bound: exception;
    -- Raised when a key that is expected to be bound is unbound.

  type KEYS_ITER is private;     -- Bound keys in arbitrary order.
  type VALUES_ITER is private;   -- Bound values in arbitrary order.
  type BINDINGS_ITER is private; -- Key,value pairs in arbitrary order

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Create                          .  SPEC
  -- .                                                     .
  -- .......................................................
  function Create return MAPPING;
  --| Purpose
  --| Return {}.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Bind                            .  SPEC
  -- .                                                     .
  -- .......................................................
  procedure Bind (Map:   in out MAPPING;
                  Key:   in     KEY_TYPE;
                  Value: in     VALUE_TYPE);
  --| Purpose
  --| Insert the binding, <key, value>, into map.  Raises
  --| already_bound iff a pair, <k', v'>, where equal(key, k'),
  --| is in map.  Raises Uninitialized_Mapping iff map has
  --| not been initialized.
  --|
  --| Exceptions
  --|   Already_Bound
  --|   Uninitialized_Mapping
  --|
  --| Notes (none)

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Unbind                          .  SPEC
  -- .                                                     .
  -- .......................................................
  procedure Unbind (Map: in out MAPPING;
                    Key: in     KEY_TYPE);
  --| Purpose
  --| If <k, v>, where equal(key, k), is in map, then removes
  --| <k, v> from map.  Raises not_bound if no such pair exists.
  --| Raises Uninitialized_Mapping iff map has not been initialized.
  --|
  --| Exceptions
  --|   Not_Bound
  --|   Uninitialized_Mapping
  --|
  --| Notes (none)

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Copy                            .  SPEC
  -- .                                                     .
  -- .......................................................
  function Copy (Map: MAPPING) return MAPPING;
  --| Purpose
  --| Returns a copy of map.  Subsequent changes to map will not be
  --| visible through applying operations to the copy of map.
  --| Assignment or parameter passing without copying will result
  --| in a single MAPPING value being shared among MAPPING objects.
  --| Raises Uninitialized_Mapping iff map has not been initialized.
  --| The assignment operation is used to transfer the values of the
  --| KEY_TYPE and VALUE_TYPE type COMPONENTs of map; consequently,
  --| changes in the values of these types may be observable through
  --| both MAPPINGs if these are access types, or if they contain
  --| COMPONENTs of an access type.
  --|
  --| Exceptions
  --|   Uninitialized_Mapping
  --|
  --| Notes (none)

-- Query Operations:

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Is_Empty                        .  SPEC
  -- .                                                     .
  -- .......................................................
  function Is_Empty (Map: MAPPING) return BOOLEAN;
  --| Purpose
  --| Return map = {}.
  --| Raises Uninitialized_Mapping iff map has not been
  --| initialized.
  --|
  --| Exceptions
  --|   Uninitialized_Mapping
  --|
  --| Notes (none)

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Size                            .  SPEC
  -- .                                                     .
  -- .......................................................
  function Size (Map: MAPPING) return NATURAL;
  --| Purpose
  --| Return |map|, the number of bindings in map.
  --| Raises Uninitialized_Mapping iff map has not been
  --| initialized.
  --|
  --| Exceptions
  --|   Uninitialized_Mapping
  --|
  --| Notes (none)

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Is_Bound                        .  SPEC
  -- .                                                     .
  -- .......................................................
  function Is_Bound (Map: MAPPING; Key: KEY_TYPE) return BOOLEAN;
  --| Purpose
  --| Return true iff equal(key, k) for some <k, v> in map.
  --| Raises Uninitialized_Mapping iff map has not been
  --| initialized.
  --|
  --| Exceptions
  --|   Uninitialized_Mapping
  --|
  --| Notes (none)

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Fetch                           .  SPEC
  -- .                                                     .
  -- .......................................................
  function Fetch (Map: MAPPING; Key: KEY_TYPE) return VALUE_TYPE;
  --| Purpose
  --| If <k, v>, where equal(key, k), is in map, then return v.
  --| Raises not_bound if no such <k, v> exists.
  --| Raises Uninitialized_Mapping iff map has not been
  --| initialized.
  --|
  --| Exceptions
  --|   Not_Bound
  --|   Uninitialized_Mapping
  --|
  --| Notes (none)

-- Iterators:

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Make_Keys_Iter                  .  SPEC
  -- .                                                     .
  -- .......................................................
  function Make_Keys_Iter (Map: MAPPING) return KEYS_ITER;
  --| Purpose
  --| Create and return a keys iterator based on map.  This
  --| object can then be used in conjunction with the more
  --| function and the next procedure to iterate over all keys
  --| that are bound in map. Raises Uninitialized_Mapping iff
  --| map has not been initialized.
  --|
  --| Exceptions
  --|   Uninitialized_Mapping
  --|
  --| Notes (none)

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.More                            .  SPEC
  -- .                                                     .
  -- .......................................................
  function More (Iter: KEYS_ITER) return BOOLEAN;
  --| Purpose
  --| Return true iff the keys iterator has not been exhausted.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Next                            .  SPEC
  -- .                                                     .
  -- .......................................................
  procedure Next (Iter: in out KEYS_ITER; Key: out KEY_TYPE);
  --| Purpose
  --| Let iter be based on the MAPPING, map.  Successive calls
  --| of next will return the bound keys of map in some
  --| arbitrary order. After all bound keys have been returned,
  --| then the procedure will raise no_more.
  --|
  --| Exceptions
  --|    No_More
  --|
  --| Notes
  --|   Map must not be changed between the invocations of
  --| Make_Keys_Iterator (Map) and Next.
      
  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Make_Values_Iter                .  SPEC
  -- .                                                     .
  -- .......................................................
  function Make_Values_Iter (Map: MAPPING) return VALUES_ITER;
  --| Purpose
  --| Create and return a values iterator based on map.  This
  --| object can then be used in conjunction with the more
  --| function and the next procedure to iterate over all values
  --| that are bound to keys in map.
  --| Raises Uninitialized_Mapping iff map has not been
  --| initialized.
  --|
  --| Exceptions
  --|   Uninitialized_Mapping
  --|
  --| Notes (none)

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.More                            .  SPEC
  -- .                                                     .
  -- .......................................................
  function More (Iter: VALUES_ITER) return BOOLEAN;
  --| Purpose
  --| Return true iff the values iterator has not been exhausted.
  --|
  --| Exceptions (none)
  --| Notes (none)
  
  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Next                            .  SPEC
  -- .                                                     .
  -- .......................................................
  procedure Next (Iter: in out VALUES_ITER; Val: out VALUE_TYPE);
  --| Purpose
  --| Let iter be based on the MAPPING, map.  Successive calls
  --| of next will return the bound values of map in some
  --| arbitrary order. After all bound values have been returned,
  --| then the procedure will raise no_more.
  --|
  --| Exceptions
  --|   No_More
  --|
  --| Notes
  --|     Map must not be changed between the invocations of
  --| Make_Values_Iterator (Map) and Next.

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Make_Keys_Iter                  .  SPEC
  -- .                                                     .
  -- .......................................................
  function Make_Bindings_Iter (Map: MAPPING) return BINDINGS_ITER;
  --| Purpose
  --| Create and return a bindings iterator based on map.
  --| This object can then be used in conjunction with the
  --| more function and the next procedure to iterate over
  --| all key/value pairs in map. Raises Uninitialized_Mapping
  --| iff map has not been initialized.
  --|
  --| Exceptions
  --|   Uninitialized_Mapping
  --|
  --| Notes (none)

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.More                            .  SPEC
  -- .                                                     .
  -- .......................................................
  function More (Iter: BINDINGS_ITER) return BOOLEAN;
  --| Purpose
  --| Return true iff the bindings iterator has not been exhausted.
  --|
  --| Exceptions (none)
  --| Notes (none)
    
  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Next                            .  SPEC
  -- .                                                     .
  -- .......................................................
  procedure Next (Iter: in out BINDINGS_ITER;
		  Key:  out    KEY_TYPE;
		  Val:  out    VALUE_TYPE);
  --| Purpose
  --| Let iter be based on the MAPPING, map.  Successive calls
  --| of next will return the key/value pairs of map in some
  --| arbitrary order. After all such pairs have been returned,
  --| then the procedure will raise no_more.
  --|
  --| Exceptions
  --|   No_More
  --|
  --| Notes
  --|     Map must not be changed between the invocations of
  --| Make_Bindings_Iterator (Map) and Next.

-- Heap management:

  -- .......................................................
  -- .                                                     .
  -- .  Hashed_Mapping_PKG.Destroy                         .  SPEC
  -- .                                                     .
  -- .......................................................
  procedure Destroy (M: in out MAPPING);
  --| Purpose
  --| Return space consumed by MAPPING value associated with
  --| object m to the heap.  (If m is uninitialized, this
  --| operation does nothing.)  If other objects share the
  --| same MAPPING value, the further use of these objects is
  --| erroneous.  COMPONENTs of type VALUE_TYPE, if they are
  --| access types, are not garbage collected. It is the user's
  --| responsibility to dispose of these objects. m is left in
  --| the uninitialized state.
  --|
  --| Exceptions (none)
  --| Notes (none)
	  
private
  type COMPONENT is record
    Key: KEY_TYPE;
    Val: VALUE_TYPE;
  end record;

  function Equal (C1, C2: COMPONENT) return BOOLEAN;
    -- Return true iff equal(c1.key, c2.key).

  package Bucket_PKG is new Lists (COMPONENT, Equal);
  use Bucket_PKG;

  type BUCKET_ARRAY is array (BUCKET_RANGE) of LIST;

  type MAPPING_REC is record
    Size    : NATURAL;
    Buckets : BUCKET_ARRAY;
  end record;
    
  type MAPPING is access MAPPING_REC;
  -- Representation Invariants:
  -- 1. r /= null.  (This would be the uninitialized case)
  -- 2. If for some i, a COMPONENT, c, is in bucket r.buckets(i),
  --    then hash(c.key) = i.
  -- 3. If a COMPONENT, c1, is in bucket, r.buckets(i), then there is
  --    no other c2 in r.buckets(i) such that equal(c1, c2).
  --    (Enforce one binding to a given key at any time.)
  -- 4. r.size equals the total number of COMPONENTs in buckets
  --    r.buckets(BUCKET_RANGE'first) through
  --    r.buckets(BUCKET_RANGE'last).
  --
  -- Abstraction Function:
  -- A(r) is the set consisting of all key, value pairs that appear as
  -- COMPONENTs in buckets r.buckets(BUCKET_RANGE'first) through
  -- r.buckets(BUCKET_RANGE'last).

  type GENERAL_ITER is record
    Map      : MAPPING;
    Current  : BUCKET_RANGE;
    Position : LIST;
  end record;

  -- For a given general_iter, i, the make, more and next operations
  -- have the following effects:
  -- make: Sets map field to the given MAPPING, sets i.current to the
  -- lowest idx of a nonempty bucket, and sets i.position to the head
  -- of that bucket.
  -- more: Returns not empty(i.position).
  -- next: key, val fields of first COMPONENT of i.position.
  -- Advances i.position to next COMPONENT in bucket, if it exists.
  -- Otherwise, increments i.current until a nonempty bucket, and sets
  -- i.position to this bucket.  When this fails, sets i.position to an
  -- empty bucket.

  type KEYS_ITER is new general_iter;
  type VALUES_ITER is new general_iter;
  type BINDINGS_ITER is new general_iter;

end Hashed_Mapping_PKG;
