--::::::::::
--lbintree.spc
--::::::::::
-- ********************************************************
-- *                                                      *
-- *  LABELED_BINARY_TREES_PKG                            *  SPEC
-- *                                                      *
-- ********************************************************
with Binary_Trees_Pkg;
generic
    type LABEL_TYPE is private; -- Type for labels stored in the tree.
    type VALUE_TYPE is private; -- Type for values stored in the tree.
    with function Difference (P, Q: LABEL_TYPE) return INTEGER is <>;
      -- Must return a value > 0 if P > Q, 0 if P = Q, and less than
      -- zero otherwise, where P and Q are labels.
package Labeled_Binary_Trees_Pkg is
--| Purpose
--| This package provides labeled binary trees, which are the same as
--| unlabeled binary trees except that when searching for or inserting
--| a value into the tree, only the label field is compared.
--| 
--| Initialization Exceptions (none)
--|
--| Notes
--| USAGE: (See Overview of Binary_Trees_Package)
--| 
--| PERFORMANCE: (See Overview of Binary_Trees_Package)
--|
--| Modifications
--| Author: Bill Toscano and Michael Gordon, Intermetrics, Inc.

-- This should be private (but cannot be)

  type LABEL_VALUE_PAIR is 
    record
      Label : LABEL_TYPE;
      Value : VALUE_TYPE;
    end record;

  function LV_Differ (P, Q: LABEL_VALUE_PAIR) return INTEGER;
  package LVT is new Binary_Trees_Pkg (LABEL_VALUE_PAIR, LV_Differ);

-- Exceptions --

  Duplicate_Value: exception renames LVT.Duplicate_Value;
  -- Raised on attempt to insert a duplicate label into a tree.

  Not_Found: exception renames LVT.Not_Found;
  -- Raised on attempt to find a label that is not in a tree.

  No_More: exception renames LVT.No_More;
  -- Raised on attempt to bump an iterator that has already scanned the
  -- entire tree.

  Out_Of_Order: exception renames LVT.Out_Of_Order;
  -- Raised if a problem in the ordering of a tree is detected.

  Invalid_Tree: exception renames LVT.Invalid_Tree;
  -- Value is not a tree or was not properly initialized.

-- Types --

  subtype SCAN_KIND is LVT.SCAN_KIND;

  --? function InOrder return LVT.Scan_Kind renames LVT.InOrder;

  InOrder   : constant SCAN_KIND := LVT.InOrder;
  PreOrder  : constant SCAN_KIND := LVT.PreOrder;
  PostOrder : constant SCAN_KIND := LVT.PostOrder;

  -- is (inorder, preorder, postorder);
  -- Used to specify the order in which values should be scanned from a tree:
  --
  -- inorder: Left, Node, Right (nodes visited in increasing order)
  -- preorder: Node, Left, Right (top down)
  -- postorder: Left, Right, Node (bottom up)

  subtype TREE is LVT.TREE;
  subtype ITERATOR is LVT.ITERATOR;

-- Operations --

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.CREATE                     .  SPEC
  -- .                                                      .
  -- ........................................................
  Function Create return TREE renames LVT.Create;
  --| Purpose
  --| Create and return an empty tree.  Note that this allocates
  --| a small amount of storage which can only be reclaimed through 
  --| a call to Destroy.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.INSERT                     .  SPEC
  -- .                                                      .
  -- ........................................................
  Procedure Insert (L: LABEL_TYPE;
                    V: VALUE_TYPE;
                    T: TREE);
  --| Purpose
  --| Insert (L, V) into T in the proper place.  If a label equal
  --| to L (according to the Difference function) is already contained
  --| in the tree, the exception Duplicate_Label is raised.
  --| Caution: Since this package does not attempt to balance trees as
  --| values are inserted, it is important to remember that inserting
  --| labels in sorted order will create a degenerate tree, where search
  --| and insertion is proportional to the N instead of to Log N.  If
  --| this pattern is common, use the Balanced_Tree function below.
  --|
  --| Exceptions
  --|   Duplicate_Value
  --|   Invalid_Tree
  --|
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.INSERT_IF_NOT_FOUND        .  SPEC
  -- .                                                      .
  -- ........................................................
  procedure Insert_if_not_Found (L         : LABEL_TYPE;
                                 V         : VALUE_TYPE;
                                 T         : TREE;
                                 Found     : out BOOLEAN;
                                 Duplicate : out VALUE_TYPE);
  --| Purpose
  --| Insert V into T in the proper place.  If a value equal
  --| to V (according to the Difference function) is already contained
  --| in the tree, Found will be True and Duplicate will be the duplicate
  --| value.  This might be a sequence of values with the same key, and
  --| V can then be added to the sequence.
  --|
  --| Exceptions
  --|   Invalid_Tree
  --|
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.REPLACE_IF_FOUND           .  SPEC
  -- .                                                      .
  -- ........................................................
  procedure Replace_if_Found (L         : LABEL_TYPE;
                              V         : VALUE_TYPE;
                              T         : TREE;
                              Found     : out BOOLEAN;
                              Old_Value : out VALUE_TYPE);
  --| Purpose
  --| Search for L in T.  If found, replace the old value with V,
  --| and return Found => True, Old_Value => the old value.  Otherwise,
  --| simply insert the L, V pair into T and return Found => False.
  --|
  --| Exceptions
  --|   Invalid_Tree
  --|
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.DESTROY                    .  SPEC
  -- .                                                      .
  -- ........................................................
  procedure Destroy (T: in out TREE) renames LVT.Destroy;
  --| Purpose
  --| The space allocated to T is reclaimed.  The space occupied by
  --| the values stored in T is not however, recovered.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.DESTROY_DEEP               .  SPEC
  -- .                                                      .
  -- ........................................................
  generic
    with procedure Free_Value (V: in out VALUE_TYPE) is <>;
    with procedure Free_Label (L: in out LABEL_TYPE) is <>;
  procedure Destroy_Deep (T: in out TREE);
  --| Purpose
  --| The space allocated to T is reclaimed.  The values and
  --| labels stored it T are reclaimed using Free_Label and
  --| Free_Value, and the tree nodes themselves
  --| are then reclaimed (in a single walk of the tree).
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.BALANCED_TREE              .  SPEC
  -- .                                                      .
  -- ........................................................
  generic
    with procedure Next_Pair (L: in out LABEL_TYPE; V: in out VALUE_TYPE)
        is <>;
    -- Each call to this procedure should return the next (Label, Value)
    -- pair to be
    -- inserted into the balanced tree being created.  If necessary,
    -- this function should check that each value is greater than the
    -- previous one, and raise Out_of_Order if necessary.  If values
    -- are not returned in strictly increasing order, the results are
    -- unpredictable.
  function Balanced_Tree (Count: NATURAL) return TREE;
  --| Purpose
  --| Create a balanced tree by calling next_Pair Count times.
  --| Each time Next_Pair is called, it must return a label that compares
  --| greater than the preceeding label.  This function is useful for balancing
  --| an existing tree (next_Pair iterates over the unbalanced tree) or
  --| for creating a balanced tree when reading data from a file which is
  --| already sorted.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.COPY_TREE                  .  SPEC
  -- .                                                      .
  -- ........................................................
  generic
    with function Copy_Label (L: LABEL_TYPE) return LABEL_TYPE is <>;
    with function Copy_Value (V: VALUE_TYPE) return VALUE_TYPE is <>;
    -- This function is called to copy a value from the old tree to the
    -- new tree.
  Function Copy_Tree (T: TREE) return TREE;
  --| Purpose
  --| Create a balanced tree that is a copy of the tree T.
  --| The exception Invalid_Tree is raised if T is not a valid tree.
  --|
  --| Exceptions
  --|   Invalid_Tree
  --|
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.IS_EMPTY                   .  SPEC
  -- .                                                      .
  -- ........................................................
  function Is_Empty (T: TREE) return BOOLEAN renames LVT.Is_Empty;
  --| Purpose
  --| Return TRUE iff T is an empty tree or if T was not initialized.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.FIND                       .  SPEC
  -- .                                                      .
  -- ........................................................
  Function Find (L: LABEL_TYPE;
                 T: TREE) return VALUE_TYPE;
  --| Purpose
  --| Search T for a label that matches L.  The corresponding value
  --| is returned.  If no matching label is found, the exception Not_Found
  --| is raised.
  --|
  --| Exceptions
  --|   Not_Found
  --|   Invalid_Tree
  --|
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.FIND                       .  SPEC
  -- .                                                      .
  -- ........................................................
  Procedure Find (L     : LABEL_TYPE;
                  T     : TREE;
                  Found : out BOOLEAN;
                  Match : out VALUE_TYPE);
  --| Purpose
  --| Search T for a label that matches L.  On return, if Found is
  --| TRUE then the corresponding value is returned in Match.  Otherwise,
  --| Found is FALSE and Match is undefined.
  --|
  --| Exceptions
  --|   Invalid_Tree
  --|
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.IS_FOUND                   .  SPEC
  -- .                                                      .
  -- ........................................................
  function Is_Found (L: LABEL_TYPE;
                     T: TREE) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff L is found in T.
  --|
  --| Exceptions
  --|   Invalid_Tree
  --|
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.SIZE                       .  SPEC
  -- .                                                      .
  -- ........................................................
  function Size (T: TREE) return NATURAL renames LVT.Size; 
  --| Purpose
  --| Return the number of values stored in T.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.VISIT                      .  SPEC
  -- .                                                      .
  -- ........................................................
  generic
    with procedure Process(L: LABEL_TYPE; V: VALUE_TYPE) is <>;
  procedure Visit (T     : TREE;
                   Order : SCAN_KIND);
  --| Purpose
  --| Invoke Process(V) for each value V in T.  The nodes are visited
  --| in the order specified by Order.  Although more limited than using
  --| an iterator, this function is also much faster.
  --|
  --| Exceptions
  --|   Invalid_Tree
  --|
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.MAKE_ITER                  .  SPEC
  -- .                                                      .
  -- ........................................................
  function Make_Iter (T: TREE) return ITERATOR renames LVT.Make_Iter;
  --| Purpose
  --| Create an iterator over a tree.
  --|
  --| Exceptions
  --|   Invalid_Tree
  --|
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.MORE                       .  SPEC
  -- .                                                      .
  -- ........................................................
  function More (I: ITERATOR) return BOOLEAN renames LVT.More;
  --| Purpose
  --| Return TRUE iff unscanned nodes remain in the tree being
  --| scanned by I.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ........................................................
  -- .                                                      .
  -- .  LABELED_BINARY_TREES_PKG.NEXT                       .  SPEC
  -- .                                                      .
  -- ........................................................
  procedure Next (I: in out ITERATOR;
                  L: out LABEL_TYPE;
                  V: out VALUE_TYPE);
  --| Purpose
  --| Return the next value in the tree being scanned by I.
  --| The exception No_More is raised if there are no more values to scan.
  --|
  --| Exceptions
  --|   No_More
  --|
  --| Notes (none)

end Labeled_Binary_Trees_Pkg;
