--::::::::::
--ltrees.spc
--::::::::::
-- *************************************************
-- *                                               *
-- *  LABELED_TREES                                *  SPEC
-- *                                               *
-- *************************************************
with Lists;
generic
  type LABEL_TYPE is private;
    -- This is used to identify nodes in the tree.
  type VALUE_TYPE is private; 
    -- Information being contained in a node of tree
  with function "<" (X: in LABEL_TYPE; Y: in LABEL_TYPE)
      return BOOLEAN is <> ;
    -- Function which defines ordering of nodes
    -- a < b -> not (b < a) and  (b /= a) for all a and b.
package Labeled_Trees is
--| Purpose
--| This package creates an ordered binary tree.  This will allow for 
--| quick insertion, and search.  
--|
--| The tree is organized such that 
--|  
--|  label (leftchild) < label (root)    label (root) < label (rightchild)
--| 
--| This means that by doing a left to right search of the tree will 
--| produce the nodes of the tree in ascending order.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| Author: Bill Toscano and Michael Gordon, Intermetrics, Inc.

  type TREE is  private;
  type TREE_ITER is private;

  Label_Already_Exists_In_Tree : exception;     
  Label_Not_Present            : exception;
  No_More                      : exception;
  Tree_Is_Empty                : exception;    

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.CREATE                            .  SPEC
  -- .                                                  .
  -- ....................................................
  function Create return TREE;
  --| Purpose
  --| This creates a tree containing no information and no children.  An 
  --| emptytree.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.DESTROY_DEEP_TREE                 .  SPEC
  -- .                                                  .
  -- ....................................................
  generic
    with procedure Dispose_Label (L :in out LABEL_TYPE);
    with procedure Dispose_Value (V :in out VALUE_TYPE);
  procedure Destroy_Deep_Tree (T : in out TREE);
  --| Purpose
  --| Destroys all nodes in a tree and the label and value associated
  --| with each node.
  --|
  --| Exceptions (none)
  --| Notes (none)
    
  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.DESTROY_TREE                      .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Destroy_Tree (T : in out TREE);
  --| Purpose
  --| Destroys a tree and returns the space which it is occupying.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.FETCH_VALUE                       .  SPEC
  -- .                                                  .
  -- ....................................................
  function Fetch_Value (T : in TREE;
                        L : in LABEL_TYPE) return VALUE_TYPE;   
  --| Purpose
  --| Get the value of the node with the given label.
  --| If the label is not present Label_Not_Present is raised.
  --|
  --| Exceptions
  --|   Label_Not_Present
  --|
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.FETCH_VALUE                       .  SPEC
  -- .                                                  .
  -- ....................................................
  function Fetch_Value (T : in TREE) return VALUE_TYPE;
  --| Purpose
  --| Return the value stored at the root node of the given tree.
  --| Raises Label_Not_Present if the tree T is empty.  
  --|
  --| Exceptions
  --|   Label_Not_Present
  --|
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.GET_TREE                          .  SPEC
  -- .                                                  .
  -- ....................................................
  function Get_Tree (T : in TREE;
                     L : in LABEL_TYPE) return TREE;
  --| Purpose
  --| Get the subtree whose root is labelled L.
  --|
  --| Exceptions
  --|   Label_Not_Present if the label L is not in T
  --|
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.FORWARD                           .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Forward (I : in out TREE_ITER);
  --| Purpose
  --| This is used to advance the iterator.  Typically this is used in
  --| conjunction with Node_Value and Node_Label.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.INSERT_NODE                       .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Insert_Node (T : in out TREE;
                         L : in LABEL_TYPE;
                         V : in VALUE_TYPE);
  --| Purpose
  --| Inserts a node into the specified tree.
  --| This adds the node with label L to the tree T.  Label_Already_Exists is 
  --| raised if L already exists in T.
  --|
  --| Exceptions
  --|   Label_Already_Exists
  --|
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.IS_EMPTY                          .  SPEC
  -- .                                                  .
  -- ....................................................
  function Is_Empty (T : in TREE) return BOOLEAN;
  --| Purpose
  --| Returns TRUE iff the tree is empty.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.IS_LABEL_IN_TREE                  .  SPEC
  -- .                                                  .
  -- ....................................................
  function Is_Label_In_Tree (T : in TREE;
                             L : in LABEL_TYPE) return BOOLEAN;
  --| Purpose
  --| Returns TRUE iff the given labels is in the given tree.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.IS_LABEL_IN_TREE                  .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Is_Label_In_Tree (T       : in TREE;
                              L       : in LABEL_TYPE;
                              Subtree : out TREE;
                              Present : out BOOLEAN);
  --| Purpose
  --| This operation can be used to see if a label is in the tree.
  --| It sets the variable Present to TRUE iff the given label is in
  --| the given tree.
  --| If it is, the Subtree out parameter can then be used to
  --| to update the value field of the label.  The sequence would be
  --| 
  --|  Is_Label_In_Tree (T, L, Subtree, Present);
  --|  if Present then
  --|     Store_Value (Subtree, SomeValue);
  --|  end if;
  --| 
  --| If the label is not Present, then Subtree is the root of the tree
  --| where the label would be stored if it were present.  Thus the following
  --| sequence would be useful.
  --|
  --| Is_Label_In_Tree (T, L, Subtree, Present);
  --| if not Present then
  --|    Insert_Node (Subtree, L, V);
  --| end if;
  --| 
  --| The advantage to this routine is that the tree need only be searched 
  --| once instead of twice once for the existence check and then once for
  --| the insertion.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.ITERATOR_LABEL                    .  SPEC
  -- .                                                  .
  -- ....................................................
  function Iterator_Label (I : in TREE_ITER) return LABEL_TYPE;
  --| Purpose
  --| Returns the label of the node corresponding to the iterator.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.ITERATOR_VALUE                    .  SPEC
  -- .                                                  .
  -- ....................................................
  function Iterator_Value (I : in TREE_ITER) return VALUE_TYPE;
  --| Purpose
  --| Returns the value of the node corresponding to the iterator.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.MAKE_TREE                         .  SPEC
  -- .                                                  .
  -- ....................................................
  function Make_Tree (L : in LABEL_TYPE;
                      V : in VALUE_TYPE) return TREE;
  --| Purpose
  --| Creates a tree whose root has the given label and value.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.MAKE_TREE_ITER_IN                 .  SPEC
  -- .                                                  .
  -- ....................................................
  function Make_Tree_Iter_In  (T : in TREE) return TREE_ITER;
  --| Purpose
  --| This sets up an iteration of the nodes of the tree in inorder.
  --| By using the Next operations the nodes of the tree are returned in
  --| in inorder. Inorder means return the left child then the node 
  --| then the right child.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.MAKE_TREE_ITER_POST               .  SPEC
  -- .                                                  .
  -- ....................................................
  function Make_Tree_Iter_Post (T : in TREE) return TREE_ITER;
  --| Purpose
  --| This sets up an iteration of the nodes of the tree in postorder.
  --| By using the Next operations the nodes of the tree are returned in
  --| post order. Post order means return the node first then its left child 
  --| and then its right child.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.MAKE_TREE_ITER_PRE                .  SPEC
  -- .                                                  .
  -- ....................................................
  function Make_Tree_Iter_Pre (T : in TREE) return TREE_ITER;
  --| Purpose
  --| This sets up an iteration of the nodes of the tree in preorder.
  --| By using the Next operations the nodes of the tree are returned in
  --| ascending order.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.MORE                              .  SPEC
  -- .                                                  .
  -- ....................................................
  function More (I : in TREE_ITER) return BOOLEAN;
  --| Purpose
  --| Returns TRUE iff there are more elements in the tree to
  --| iterate over.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.NEXT                              .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Next (I : in out TREE_ITER;
                  V : out VALUE_TYPE);    
  --| Purpose
  --| This returns the next element in the iteration and advances the iterator.
  --| No_More is raised when after the last element has been returned and
  --| an attempt is made to get another element.
  --|
  --| Exceptions
  --|   No_More
  --|
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.NEXT                              .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Next (I : in out TREE_ITER;
                  V : out VALUE_TYPE;
                  L : out LABEL_TYPE);    
  --| Purpose
  --| This iteration operation returns the label of a node as well as the 
  --| node's value.  No_More is raised if Next is called after the last
  --| element of the tree has been returned.
  --|
  --| Exceptions
  --|   No_More
  --|
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.STORE_VALUE                       .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Store_Value (T : in out TREE;
                         L : in LABEL_TYPE;
                         V : in VALUE_TYPE);
  --| Purpose
  --| Sets a new value V in the tree T at the node identified
  --| by the label L.  
  --| Label_Not_Present is raised if L is not in T.
  --|
  --| Exceptions
  --|   Label_Not_Present
  --|
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  LABELED_TREES.STORE_VALUE                       .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Store_Value (T : in out TREE;
                         V : in VALUE_TYPE);
  --| Purpose
  --| This stores the value V in the root node of the tree T.   
  --| Raises Label_Not_Present if T is empty.
  --|
  --| Exceptions
  --|   Label_Not_Present
  --|
  --| Notes (none)

private
   type NODE;
   type TREE is access NODE;

   type NODE is 
     record
       Label       : LABEL_TYPE;
       Value       : VALUE_TYPE;
       Left_Child  : TREE;
       Right_Child : TREE;
     end record;

   package NODE_ORDER is new Lists (TREE);

   type TREE_ITER is
      record
        Node_List : Node_Order.LIST;
        State     : Node_Order.LISTITER;
      end record;

end Labeled_Trees;
