--::::::::::
--matrix.spc
--::::::::::
-- ****************************************************************
-- *                                                              *
-- *  Matrix_Package                                              *  SPEC
-- *                                                              *
-- ****************************************************************
package MATRIX_PACKAGE is
--| Purpose
--| This package is a general purpose matrix package. It defines data
--| types VECTOR and MATRIX, and contains functions to perform general
--| matrix algebra operations.
--|
--| Initialization Exceptions (none)
--|
--| Notes
--|   Not all MIL-HDBK-1804 PDL annotations are used in this package
--| due to its simplicity.
--|
--| Modifications
--| Author: Dr. Roger Lee, Naval Air Development Center
--|         Art Adamson, Consultant

-- Types
  type VECTOR is array(integer range<>) of float ;
  subtype VEC2T is VECTOR (integer range 1..2) ;
  subtype VEC3T is VECTOR (integer range 1..3) ;
  type MATRIX is array(integer range<>,integer range <>) of float;
  type MATR2T is array(integer range<>) of VEC2T;

-- Exceptions
  INCOMPARABLE_DIMENSION :exception; -- the dimension of matrices
  -- or vectors to be operated are 
  -- incomparable
  SINGULAR : exception;    -- matrix to be inverted is singular

-- Operations
  function TRANSPOSE(A : MATRIX) return MATRIX ; -- transpose of matrix
  function TRANSPOSE(A : VECTOR) return VECTOR ; -- transpose of vector
  function "+" (A : VECTOR; B : VECTOR) return VECTOR ; -- sum of vector
  function "+" (A : MATRIX; B : MATRIX) return MATRIX ; -- sum of matrix
  function "+" (A : float;  B : VECTOR) return VECTOR ; 
  -- float added to, each term of matrix
  function "+" (A : VEC2T;  B : MATR2T) return MATR2T ;
  -- Vec2T added to, each term of MATR2T
  function "+" (A : MATR2T;  B : MATR2T) return MATR2T ;
  -- Corressponding terms added.
  function "-" (A : VECTOR; B : VECTOR) return VECTOR ; 
  -- difference of vector
  function "-" (A : MATRIX; B : MATRIX) return MATRIX ; 
  -- difference of matrix
  function "*" (A : float;  B : VECTOR) return VECTOR ; 
  -- scalar, vector multiplication
  function "*" (A : VECTOR; B : float)  return VECTOR ; 
  -- vector, scalar multiplication
  function "*" (A : VECTOR; B : VECTOR) return float ;  
  -- inner(dot) product of two vectors
  function "*" (A : MATRIX; B : VECTOR) return VECTOR ; 
  -- matrix,column vector multiplication
  function mat4mult(UL : MATRIX; UR : MATRIX; BL : MATRIX; BR : MATRIX;
                 B : VECTOR) return VECTOR ; 
  -- large matrix broken into 4 smaller ones, column vector multiplication 
  -- (upper left, upper right, bottom left, bottom right--all square)
  function "*" (A : VECTOR; B : MATRIX) return VECTOR ; 
  -- row vector,matrix multiplication
  function "*" (A : float;  B : MATRIX) return MATRIX ; 
  -- scalar, matrix multiplication
  function "*" (A : MATRIX; B : float)  return MATRIX ; 
  -- matrix, scalar multiplication
  function "*" (A : MATRIX; B : MATRIX) return MATRIX ; 
  -- matrix, matrix multiplication
  function "*" (A : float;  B : MATR2T) return MATR2T ;
  -- Multiplies each term of a MATR2T by a float
  function "*" (A : VEC2T;  B : MATR2T) return VECTOR ;
  -- Dot product of each term of MATR2T by a VEC2T, return array of floats
  function "*" (A : VECTOR;  B : MATR2T) return MATR2T ;
  --Multiplies each term of VEC2T by a corresponding float from a VECTOR
  function "**"(A : MATRIX; P : integer) return MATRIX; 
  -- square matrix raised to integer power
  -- if P = -1, we invert the matrix
  function "**" (A : VECTOR; B : VECTOR) return VECTOR ;
  -- A X B = ab sin(theta) a direction
  --perpendicular to plane of A & B.
  function JCROSS (A : VEC2T) return VEC2T ;
  --Rotates Vec2T 90 degrees CW.
  function JCROSS (A : MATR2T) return MATR2T ;
  --Rotates Vec2T's 90 degrees CW.
  function ROTX   (A : VEC2T) return VEC2T ;
  --Rotates Vec2T 180 degrees about the X axis.
  function ROTY   (A : VEC2T) return VEC2T ;
  --Rotates Vec2T 180 degrees about the Y axis.
  function aXbDOTj(A : VEC2T; B : VEC2T) return FLOAT;
  --Gets magnitude of A cross B for 2 2D vectors.
  function GETTAN (A : VEC2T; B : VEC2T) return FLOAT;
  --Gets TAN(THETA) between 2 2D vectors.

end MATRIX_PACKAGE;
