--::::::::::
--mlib.spc
--::::::::::
-- ***************************************************************
-- *                                                             *
-- *  FLOATING_CHARACTERISTICS                                   *  SPEC
-- *                                                             *
-- ***************************************************************
package Floating_Characteristics is
--| Purpose
--| This package is a floating mantissa definition of a binary FLOAT 
--| It was first used on the DEC-10 and the VAX but should work for any
--| since the parameters are obtained by initializing on the actual hardware.
--| Otherwise the parameters could be set in the spec if known.
--| This is a preliminary package that defines the properties 
--| of the particular floating point type for which we are going to
--| generate the math routines.
--| The constants are those required by the routines described in
--| "Software Manual for the Elementary Functions" W. Cody & W. Waite
--| Prentice-Hall 1980.
--| Actually most are needed only for the test programs
--| rather than the functions themselves, but might as well be here.
--| Most of these could be in the form of attributes if 
--| all the floating types to be considered were those built into the
--| compiler, but we also want to be able to support user defined types
--| such as software floating types of greater precision than
--| the hardware affords, or types defined on one machine to
--| simulate another.
--| So we use the Cody-Waite names and derive them from an adaptation
--| of the MACHAR routine as given by Cody-Waite in Appendix B.
--|
--| Initialization Exceptions (none)
--|
--| Notes
--|    Not all MIL-HDBK-1804 PDL annotations are
--| used in this package due to its simplicity.
--|
--| Modifications
--| Author: Major Terry Courtwright, World Wide Military Command and
--|  Control Information Systems Joint Program Management Office

    Ibeta : INTEGER;
    --  The radix of the floating-point representation

    It : INTEGER;
    --  The number of base IBETA digits in the DIS_FLOAT significand

    Irnd : INTEGER;
    --  TRUE (1) if floating addition rounds, FALSE (0) if truncates

    Ngrd : INTEGER;
    --  Number of guard digits for multiplication

    Machep : INTEGER;
    --  The largest negative integer such that
    --    1.0 + FLOAT(IBETA) ** MACHEP /= 1.0
    --  except that MACHEP is bounded below by -(IT + 3)

    Negep : INTEGER;
    --  The largest negative integer such that
    --    1.0 -0 FLOAT(IBETA) ** NEGEP /= 1.0
    --  except that NEGEP is bounded below by -(IT + 3)

    Iexp : INTEGER;
    --  The number of bits (decimal places if IBETA = 10)
    --  reserved for the representation of the exponent (including
    --  the bias or sign) of a floating-point number

    Minexp : INTEGER;
    --  The largest in magnitude negative integer such that
    --  FLOAT(IBETA) ** MINEXP is a positive floating-point number

    Maxexp : INTEGER;
    --  The largest positive exponent for a finite floating-point number

    Eps : FLOAT;
    --  The smallest positive floating-point number such that
    --                              1.0 + EPS /= 1.0
    --  In particular, if IBETA = 2 or IRND = 0,
    --  EPS = FLOAT(IBETA) ** MACHEP
    --  Otherwise, EPS = (FLOAT(IBETA) ** MACHEP) / 2

    Epsneg : FLOAT;
    --  A small positive floating-point number such that 1.0-EPSNEG /= 1.0

    Xmin : FLOAT;
    --  The smallest non-vanishing floating-point power of the radix
    --  In particular, XMIN = FLOAT(IBETA) ** MINEXP

    Xmax : FLOAT;
    --  The largest finite floating-point number

--  Here the structure of the floating type is defined.
--  I have assumed that the exponent is always some integer form.
--  The mantissa can vary.
--  Most often it will be a fixed type or the same floating type
--  depending on the most efficient machine implementation.
--  Most efficient implementation may require details of the machine hardware
--  In this version the simplest representation is used.
--  The mantissa is extracted into a FLOAT and uses the predefined operations.
  subtype EXPONENT_TYPE is INTEGER;    --  should be derived
  subtype MANTISSA_TYPE is FLOAT;     --   range -1.0..1.0;

--  A consequence of the rigorous constraints on MANTISSA_TYPE is that 
--  operations must be very carefully examined to make sure that no number
--  greater than one results.
--  Actually this limitation is important in constructing algorithms
--  which will also run when MANTISSA_TYPE is a fixed point type.

--  If we are not using the STANDARD type, we have to define all the 
--  operations at this point.
--  We also need PUT for the type if it is not otherwise available.

--  Now we do something strange.
--  Since we do not know in the following routines whether the mantissa
--  will be carried as a fixed or floating type, we have to make some
--  provision for dividing by two.
--  We cannot use the literals, since FIXED/2.0 and FLOAT/2 will fail.
--  We define a type-dependent factor that will work.
  Mantissa_Divisor_2 : constant FLOAT := 2.0;
  Mantissa_Divisor_3 : constant FLOAT := 3.0;

--  This will work for the MANTISSA_TYPE defined above.
--  The alternative of defining an operation "/" to take care of it
--  is too sweeping and would allow unAda-like errors.

  Mantissa_Half : constant MANTISSA_TYPE := 0.5;

-- Subprograms
  procedure Defloat (X : in FLOAT;
                     L : out EXPONENT_TYPE;
                     E : out MANTISSA_TYPE);
  procedure Refloat (N : in EXPONENT_TYPE;
                     F : in MANTISSA_TYPE; 
                     Z : out FLOAT);

--  Since the user may wish to define a floating type by some other name
--  CONVERT_TO_FLOAT is used rather than just FLOAT for explicit coersion.
  function Convert_to_Float (K : INTEGER) return FLOAT;

  --function CONVERT_TO_FLOAT(N : EXPONENT_TYPE) return FLOAT;
  function Convert_to_Float (F : MANTISSA_TYPE) return FLOAT;

end Floating_Characteristics;

-- ***************************************************************
-- *                                                             *
-- *  NUMERIC_PRIMITIVES                                         *  SPEC
-- *                                                             *
-- ***************************************************************
with Floating_Characteristics;
use  Floating_Characteristics;
package Numeric_Primitives is
--| Purpose
--| This package contains the definitions of several useful constants
--| and functions associated with FLOAT numbers.
--|
--| Initialization Exceptions (none)
--|
--| Notes
--| Due to the simplicity of this package, the MIL-HDBK-1804 PDL
--| annotations are not used in the rest of this specification.
--|
--| Modifications
--| Author: Major Terry Courtwright, World Wide Military Command and
--|  Control Information Systems Joint Program Management Office

  --  This may seem a little much but is put in this form to allow the
  --  same form to be used for a generic package.
  --  If that is not needed, simple litterals could be substituted.
  Zero  : FLOAT;
  One   : FLOAT;
  Two   : FLOAT;
  Three : FLOAT;
  Half  : FLOAT;

  --  The following "constants" are effectively deferred to
  --  the initialization part of the package body.
  --  This is in order to make it possible to generalize the floating type.
  --  If that capability is not desired, constants may be included here.
  PI            : FLOAT;
  One_Over_PI   : FLOAT;
  Two_Over_PI   : FLOAT;
  PI_Over_Two   : FLOAT;
  PI_Over_Three : FLOAT;
  PI_Over_Four  : FLOAT;
  PI_Over_Six   : FLOAT;

-- Subprograms
  function Sign (X, Y : FLOAT) return FLOAT;
    --  Returns the value of X with the sign of Y.

  function Max (X, Y :  FLOAT) return FLOAT;
    --  Returns the algebraicly larger of X and Y.

  function Truncate (X : FLOAT) return FLOAT;
    --  Returns the floating value of the integer no larger than X.
    --  AINT(X)

  function Round (X : FLOAT) return FLOAT;
    --  Returns the floating value nearest X.
    --  AINTRND(X)

  function Ran return FLOAT;
    --  This uses a portable algorithm and is included at this point.
    --  Algorithms that presume unique machine hardware information
    --  should be initiated in FLOATING_CHARACTERISTICS.

end Numeric_Primitives;

-- ***************************************************************
-- *                                                             *
-- *  CORE_FUNCTIONS                                             *  SPEC
-- *                                                             *
-- ***************************************************************
with Floating_Characteristics;
use  Floating_Characteristics;
package Core_Functions is
--| Purpose
--| This package contains the definitions of several fundamental
--| functions associated with FLOAT numbers.
--|
--| Initialization Exceptions (none)
--|
--| Notes
--| Due to the simplicity of this package, the MIL-HDBK-1804 PDL
--| annotations are not used in the rest of this specification.
--|
--| Modifications
--| Author: Major Terry Courtwright, World Wide Military Command and
--|  Control Information Systems Joint Program Management Office

  Exp_Large : FLOAT;
  Exp_Small : FLOAT;

-- Subprograms
  function SQRT(X : FLOAT) return FLOAT;

  function CBRT(X : FLOAT) return FLOAT;

  function LOG(X : FLOAT) return FLOAT;
  function LOG10(X : FLOAT) return FLOAT;

  function EXP(X : FLOAT) return FLOAT;

  function "**"(X, Y : FLOAT) return FLOAT;

end Core_Functions;

-- ***************************************************************
-- *                                                             *
-- *  TRIG_FUNCTIONS                                             *  SPEC
-- *                                                             *
-- ***************************************************************
package Trig_Functions is
--| Purpose
--| This package contains the definitions of several trigonometric
--| and hypertrigonometic functions associated with FLOAT numbers.
--|
--| Initialization Exceptions (none)
--|
--| Notes
--| Due to the simplicity of this package, the MIL-HDBK-1804 PDL
--| annotations are not used in the rest of this specification.
--|
--| Modifications
--| Author: Major Terry Courtwright, World Wide Military Command and
--|  Control Information Systems Joint Program Management Office

  function SIN(X : FLOAT) return FLOAT;
  function COS(X : FLOAT) return FLOAT;
  function TAN(X : FLOAT) return FLOAT;
  function COT(X : FLOAT) return FLOAT;

  function ASIN(X : FLOAT) return FLOAT;
  function ACOS(X : FLOAT) return FLOAT;
  function ATAN(X : FLOAT) return FLOAT;
  function ATAN2(V, U : FLOAT) return FLOAT;

  function SINH(X : FLOAT) return FLOAT;
  function COSH(X : FLOAT) return FLOAT;
  function TANH(X : FLOAT) return FLOAT;

end Trig_Functions;
