--::::::::::
--ordset.spc
--::::::::::
-- ****************************************************
-- *                                                  *
-- *  ORDEREDSETS                                     *  SPEC
-- *                                                  *
-- ****************************************************
with BinaryTrees;
generic
      type ITEMTYPE is private;
      with function "<" (X, Y : in ITEMTYPE) return BOOLEAN;
package OrderedSets is
--| Purpose
--| This abstractions is a counted ordered set.  This means that 
--| associated with each member of the set is a count of the number of
--| times it appears in the set.  The order part means that there is
--| an ordering associated with the members.  This allows fast insertion.
--| It also makes it easy to iterate over the set in order.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| Author: Bill Toscano and Michael Gordon, Intermetrics

      type SET is private;
      type SETITER is private;

  -- .....................................................
  -- .                                                   .
  -- .  ORDEREDSETS.CARDINALITY                          .  SPEC
  -- .                                                   .
  -- .....................................................
  function Cardinality (S : in SET) return NATURAL;
  --| Purpose
  --| Return the number of members in the set.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  ORDEREDSETS.CREATE                               .  SPEC
  -- .                                                   .
  -- .....................................................
  function Create return SET;
  --| Purpose
  --| Return the empty set.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  ORDEREDSETS.DESTROY                              .  SPEC
  -- .                                                   .
  -- .....................................................
  procedure Destroy (S : in out SET);
  --| Purpose
  --| Destroy a set and return its space.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  ORDEREDSETS.GETCOUNT                             .  SPEC
  -- .                                                   .
  -- .....................................................
  function GetCount (I : in SETITER) return NATURAL;
  --| Purpose
  --| Returns the count associated with the member corresponding to the
  --| current interator I.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  ORDEREDSETS.INSERT                               .  SPEC
  -- .                                                   .
  -- .....................................................
  procedure Insert (M : in ITEMTYPE;
                    S : in out SET);
  --| Purpose
  --| Insert a member M into set S.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  ORDEREDSETS.MAKESETITER                          .  SPEC
  -- .                                                   .
  -- .....................................................
  function MakeSetIter (S : in SET) return SETITER;
  --| Purpose
  --| Prepares a user for an iteration operation by returning
  --| a SetIter.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  ORDEREDSETS.MORE                                 .  SPEC
  -- .                                                   .
  -- .....................................................
  function More (I : in SETITER) return BOOLEAN;
  --| Purpose
  --| Returns TRUE if there are more elements in the set
  --| to iterate over.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .....................................................
  -- .                                                   .
  -- .  ORDEREDSETS.NEXT                                 .  SPEC
  -- .                                                   .
  -- .....................................................
  procedure Next (I : in out SETITER;
                  M : out ITEMTYPE);
  --| Purpose
  --| Returns the current member in the iteration and increments
  --| the iterator.
  --|
  --| Exceptions (none)
  --| Notes (none)

private 
   type MEMBER is 
       record 
         Info  : ITEMTYPE;
         Count : NATURAL;
       end record;

   function "<" (X: in MEMBER; Y: in MEMBER) return BOOLEAN;

   package TreePkg is new BinaryTrees (ITEMTYPE => MEMBER, "<" => "<");

   type SET is
       record 
         SetRep :TreePkg.TREE;
       end record;

   type SETITER is
       record
         Place : TreePkg.TREEITER;
         Count : NATURAL;
       end record;

end OrderedSets;
