--::::::::::
--out.spc
--::::::::::
-- **********************************
-- *                                *
-- *  Output_File                   *  SPEC
-- *                                *
-- **********************************
package Output_File is
--| Purpose
--| Output_File implements an abstract data type of an output file.
--| Output_File offers an abstraction that can be made more efficient
--| by not using Text_IO (and having its associated overhead imposed)
--| if possible and also offers the ability to suppress the output,
--| which may be desired if a caller is skipping over pages and just
--| wants to output to a null device during this process.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| 08/16/89  Rick Conn    Initial Version

  type FILE_TYPE is
    private;

  Cannot_Create_Output_File
    : exception;
  Write_Error
    : exception;

  -- ..................................
  -- .                                .
  -- .  Output_File.Already_Exists    .  SPEC
  -- .                                .
  -- ..................................
  function Already_Exists
    ( File_Name      : in STRING )
    return BOOLEAN;
  --| Purpose
  --| Determine if the FILE_TYPE object already exists.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  Output_File.Delete            .  SPEC
  -- .                                .
  -- ..................................
  function Delete
    ( File_Name      : in STRING )
    return BOOLEAN;
  --| Purpose
  --| Delete the FILE_TYPE object.  Return TRUE if successful.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  Output_File.Create            .  SPEC
  -- .                                .
  -- ..................................
  procedure Create
    ( Id             : in out FILE_TYPE;
      File_Name      : in STRING );
  --| Purpose
  --| Create creates a new FILE_TYPE object.
  --|
  --| Exceptions
  --|   Cannot_Create_Output_File
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  Output_File.Put               .  SPEC
  -- .                                .
  -- ..................................
  procedure Put
    ( Id             : in out FILE_TYPE;
      Item           : in CHARACTER );
  procedure Put
    ( Id             : in out FILE_TYPE;
      Item           : in STRING );
  --| Purpose
  --| Put writes an Item to the FILE_TYPE object.
  --|
  --| Exceptions
  --|   Write_Error
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  Output_File.Put_Line          .  SPEC
  -- .                                .
  -- ..................................
  procedure Put_Line
    ( Id             : in out FILE_TYPE;
      Item           : in STRING );
  --| Purpose
  --| Put_Line writes an Item to the FILE_TYPE object.  The Item is followed
  --| by a New_Line;
  --|
  --| Exceptions
  --|   Write_Error
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  Output_File.New_Line          .  SPEC
  -- .                                .
  -- ..................................
  procedure New_Line
    ( ID             : in out FILE_TYPE );
  --| Purpose
  --| New_Line writes an end-of-line sequence to the FILE_TYPE object.
  --|
  --| Exceptions
  --|   Write_Error
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  Output_File.New_Page          .  SPEC
  -- .                                .
  -- ..................................
  procedure New_Page
    ( ID             : in out FILE_TYPE );
  --| Purpose
  --| New_Page writes an end-of-page sequence to the FILE_TYPE object.
  --|
  --| Exceptions
  --|   Write_Error
  --|
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  Output_File.Enable_Output     .  SPEC
  -- .  Output_File.Disable_Output    .
  -- .                                .
  -- ..................................
  procedure Enable_Output
    ( ID             : in out FILE_TYPE );
  procedure Disable_Output
    ( ID             : in out FILE_TYPE );
  --| Purpose
  --| Enable_Output and Disable_Output enable and disable the output of
  --| Items and new lines to the FILE_TYPE object.  When created, output
  --| to a FILE_TYPE object is enabled.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..................................
  -- .                                .
  -- .  Output_File.Close             .  SPEC
  -- .                                .
  -- ..................................
  procedure Close
    ( ID             : in out FILE_TYPE );
  --| Purpose
  --| Close closes output to the FILE_TYPE object.
  --|
  --| Exceptions (none)
  --| Notes (none)

private -- Output_File
  type FILE_OBJECT;
  type FILE_TYPE is
    access FILE_OBJECT;

end Output_File;
