--::::::::::
--priqueue.spc
--::::::::::
-- ********************************************************
-- *                                                      *
-- *  PRIORITIZED_QUEUE                                   *  SPEC
-- *                                                      *
-- ********************************************************
generic
   type ENQUEUED_OBJECT is limited private;
   type PRIORITY_VALUE is (<>);
   with procedure Assign (Target : in out ENQUEUED_OBJECT;
                          Source : in     ENQUEUED_OBJECT) is <>;
   with function "=" (First_Object  : in ENQUEUED_OBJECT;
                      Second_Object : in ENQUEUED_OBJECT) return BOOLEAN is <>;
   with procedure Destroy (Targeted_Object : in out ENQUEUED_OBJECT) is <>;
   with function "<" (First_Object  : in PRIORITY_VALUE;
                      Second_Object : in PRIORITY_VALUE) return BOOLEAN is <>;
package Prioritized_Queue is
--| Purpose
--| Support prioritized queues.  Items may be added to removed
--| from these queues based on priority, as opposed to first
--| arrival.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| Author: Bill Wolfe, Clemson University

-- *******************************************************
-- This software is part of the Clemson University
-- Computer Science Department's Ada Software
-- Repository, and is copyrighted (C) 1989 by
-- Clemson University.  Permission to copy without
-- fee all or part of this software is granted,
-- provided that the copies are not made or
-- distributed for direct commercial advantage, and
-- that this copyright notice is not deleted or
-- modified.  To copy otherwise, or to republish,
-- requires a fee and/or specific permission.
-- *******************************************************

   type PRIORITY_QUEUE is limited private;
    -- requires O (n) space, where n is the NUMBER_OF_ITEMS in the queue.

   Requested_Item_Does_Not_Exist_In_This_Priority_Queue  : EXCEPTION;
   No_Items_Currently_Exist_In_This_Empty_Priority_Queue : EXCEPTION;

   type POINTER_TO_PRIORITY_QUEUE is access PRIORITY_QUEUE;

  -- ..........................................................
  -- .                                                        .
  -- .  PRIORITIZED_QUEUE.DESTROY                             .  SPEC
  -- .                                                        .
  -- ..........................................................
  procedure Destroy (Targeted_Object : in out POINTER_TO_PRIORITY_QUEUE);
  --| Purpose
  --| Remove the queue, freeing the space allocated to it.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --| Unlike UNCHECKED_DEALLOCATION, this procedure will properly
  --| destroy the PRIORITY_QUEUE being pointed to.  Works in O (n)
  --| time, where n is the NUMBER_OF_ITEMS in the PRIORITY_QUEUE
  --| being pointed to.

  -- ..........................................................
  -- .                                                        .
  -- .   PRIORITIZED_QUEUE.INSERT_ITEM                        .  SPEC
  -- .                                                        .
  -- ..........................................................
  procedure Insert_Item (Queue    : in out PRIORITY_QUEUE;
                         Object   : in     ENQUEUED_OBJECT;
                         Priority : in     PRIORITY_VALUE);
  --| Insert the indicated OBJECT into the QUEUE at the given PRIORITY.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --| The QUEUE can safely handle multiple instances of a given
  --| (OBJECT, PRIORITY) pair.  Works in O (log n) time, where n
  --| is the NUMBER_OF_ITEMS in the updated QUEUE.
  --| A series of consecutive initializing insertions uses O (n) time,
  --| where n is the number of insertions.

  -- ..........................................................
  -- .                                                        .
  -- .  PRIORITIZED_QUEUE.REMOVE_HIGHEST_PRIORITY_OBJECT      .  SPEC
  -- .                                                        .
  -- ..........................................................
  procedure Remove_Highest_Priority_Object
      (Highest_Priority_Object : in out ENQUEUED_OBJECT;
       Queue                   : in out PRIORITY_QUEUE); 
  --| Purpose
  --| Remove the highest priority object.  If there are several objects
  --| of the same highest priority, the first object entered will be
  --| removed.
  --|
  --| Exceptions
  --|   No_Items_Currently_Exist_In_This_Empty_Priority_Queue
  --|
  --| Notes
  --| Works in O (log n) time, where n is the NUMBER_OF_ITEMS
  --| originally in the QUEUE.  Raises
  --| No_Items_Currently_Exist_In_This_Empty_Priority_Queue
  --| if the QUEUE is EMPTY.

  -- ..........................................................
  -- .                                                        .
  -- .  PRIORITIZED_QUEUE.REMOVE_HIGHEST_PRIORITY_OBJECT      .  SPEC
  -- .                                                        .
  -- ..........................................................
  procedure Remove_Highest_Priority_Object
      (Highest_Priority_Object : in out ENQUEUED_OBJECT;
       Priority_of_the_Object  :    out PRIORITY_VALUE;
       Queue                   : in out PRIORITY_QUEUE); 
  --| Purpose
  --| Remove the highest priority object in a queue, returning both
  --| the object and its priority.
  --|
  --| Exceptions
  --|   No_Items_Currently_Exist_In_This_Empty_Priority_Queue
  --|
  --| Notes
  --| Works in O (log n) time, where n is the NUMBER_OF_ITEMS
  --| originally in the QUEUE.
  --| Raises No_Items_Currently_Exist_In_This_Empty_Priority_Queue
  --| if the QUEUE is EMPTY.

  -- ..........................................................
  -- .                                                        .
  -- .  PRIORITIZED_QUEUE.DELETE_ITEM                         .  SPEC
  -- .                                                        .
  -- ..........................................................
  procedure Delete_Item (Queue    : in out PRIORITY_QUEUE;
                         Object   : in     ENQUEUED_OBJECT;
                         Priority : in     PRIORITY_VALUE);
  --| Purpose
  --| Delete an item in the queue given the item and its priority.
  --|
  --| Exceptions
  --|   Requested_Item_Does_Not_Exist_In_This_Priority_Queue
  --|   No_Items_Currently_Exist_In_This_Empty_Priority_Queue
  --|
  --| Notes
  --| Works in O (n) time, where n is the NUMBER_OF_ITEMS originally
  --| in the QUEUE.
  --|
  --| If multiple occurrences of the specified OBJECT and PRIORITY
  --| exist, the first such occurrence found will be deleted, and
  --| all others will be left undisturbed.
  --|   PURGE_ITEM should be used if you wish to eliminate all such
  --| occurrences.
  --|
  --| If no occurrences of the specified OBJECT and PRIORITY exist,
  --| and the queue is not EMPTY, raises
  --| Requested_Item_Does_Not_Exist_In_This_Priority_Queue.
  --|
  --| If the QUEUE is EMPTY, raises
  --| No_Items_Currently_Exist_In_This_Empty_Priority_Queue.

  -- ..........................................................
  -- .                                                        .
  -- .  PRIORITIZED_QUEUE.PURGE_ITEM                          .  SPEC
  -- .                                                        .
  -- ..........................................................
  procedure Purge_Item (Queue  : in out PRIORITY_QUEUE;
                        Object : in     ENQUEUED_OBJECT);
  --| Purpose
  --| Remove all instances of an OBJECT regardless of its priority.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --| Works in O (n) time, where n is the NUMBER_OF_ITEMS originally
  --| in the QUEUE.
  --|
  --| Will terminate normally, even if the QUEUE was already EMPTY... 

  -- ..........................................................
  -- .                                                        .
  -- .  PRIORITIZED_QUEUE.PURGE_ITEM                          .  SPEC
  -- .                                                        .
  -- ..........................................................
  procedure Purge_Item (Queue    : in out PRIORITY_QUEUE;
                        Object   : in     ENQUEUED_OBJECT;
                        Priority : in     PRIORITY_VALUE);
  --| Purpose
  --| Remove all instances of an OBJECT at a given PRIORITY.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --| Works in O (n) time, where n is the NUMBER_OF_ITEMS originally
  --| in the QUEUE.
  --|
  --| Will terminate normally, even if the QUEUE was already EMPTY... 

  -- ..........................................................
  -- .                                                        .
  -- .  PRIORITIZED_QUEUE.PURGE_PRIORITY                      .  SPEC
  -- .                                                        .
  -- ..........................................................
  procedure Purge_Priority (Queue    : in out PRIORITY_QUEUE;
                            Priority : in     PRIORITY_VALUE);
  --| Purpose
  --| Removes all objects of a given PRIORITY.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --| Works in O (n) time, where n is the NUMBER_OF_ITEMS originally
  --| in the QUEUE.
  --|
  --| Will terminate normally, even if the QUEUE was already EMPTY... 

  -- ..........................................................
  -- .                                                        .
  -- .  PRIORITIZED_QUEUE.PURGE_PRIORITY_RANGE                .  SPEC
  -- .                                                        .
  -- ..........................................................
  procedure Purge_Priority_Range (Queue         : in out PRIORITY_QUEUE;
                                  From_Priority : in     PRIORITY_VALUE; 
                                  To_Priority   : in     PRIORITY_VALUE);
  --| Purpose
  --| Remove all objects with priorities between FROM_PRIORITY and
  --| TO_PRIORITY, inclusive.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --| Works in O (n) time, where n is the NUMBER_OF_ITEMS originally
  --| in the QUEUE.
  --|
  --| Will terminate normally, even if the QUEUE was already EMPTY... 

  -- ..........................................................
  -- .                                                        .
  -- .  PRIORITIZED_QUEUE.MERGE                               .  SPEC
  -- .                                                        .
  -- ..........................................................
  procedure Merge (Target_Queue : in out PRIORITY_QUEUE;
                   Source_Queue : in PRIORITY_QUEUE);
  --| Purpose
  --| Merge two queues.  The objects which were in the SOURCE_QUEUE
  --| are merged into the TARGET_QUEUE; the SOURCE_QUEUE
  --| is left EMPTY.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --| Works in O (log n) time, where n is the NUMBER_OF_ITEMS
  --| in the newly merged queue. 

  -- ..........................................................
  -- .                                                        .
  -- .  PRIORITIZED_QUEUE.CHANGE_PRIORITY                     .  SPEC
  -- .                                                        .
  -- ..........................................................
  procedure Change_Priority (Queue        : in out PRIORITY_QUEUE;
                             Object       : in     ENQUEUED_OBJECT;
                             Old_Priority : in     PRIORITY_VALUE;
                             New_Priority : in     PRIORITY_VALUE);
  --| Purpose
  --| Change the priority of an object in a queue.
  --|
  --| Exceptions
  --|   Requested_Item_Does_Not_Exist_In_This_Priority_Queue
  --|   No_Items_Currently_Exist_In_This_Empty_Priority_Queue
  --|
  --| Notes
  --| Works in O (n) time, where n is the NUMBER_OF_ITEMS in the QUEUE.
  --|
  --| If multiple occurrences of the specified OBJECT and OLD_PRIORITY
  --| exist in the QUEUE, the first such occurrence found will be
  --| modified, and all others will be left undisturbed.
  --|
  --| If no occurrences of the specified OBJECT and OLD_PRIORITY exist
  --| in the QUEUE, and the QUEUE is not EMPTY, raises
  --| Requested_Item_Does_Not_Exist_In_This_Priority_Queue.
  --|
  --| If the QUEUE is EMPTY, raises
  --| No_Items_Currently_Exist_In_This_Empty_Priority_Queue.

  -- ..........................................................
  -- .                                                        .
  -- .  PRIORITIZED_QUEUE.EMPTY                               .  SPEC
  -- .                                                        .
  -- ..........................................................
  function Empty (Queue : in PRIORITY_QUEUE) return BOOLEAN;
  --| Purpose
  --| Determine if a queue is empty.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --| Works in O (1) time.

  -- ..........................................................
  -- .                                                        .
  -- .  PRIORITIZED_QUEUE.NUMBER_OF_ITEMS                     .  SPEC
  -- .                                                        .
  -- ..........................................................
  function Number_of_Items (Queue : in PRIORITY_QUEUE)
      return NATURAL;
  --| Purpose
  --| Determines the number of items in a queue.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --| Works in O (n) time, where n is the NUMBER_OF_ITEMS in the QUEUE.

  -- ..........................................................
  -- .                                                        .
  -- .  PRIORITIZED_QUEUE.ASSIGN                              .  SPEC
  -- .                                                        .
  -- ..........................................................
  procedure Assign (Target_Object : in out PRIORITY_QUEUE;
                    Source_Object : in     PRIORITY_QUEUE);
  --| Purpose
  --| Assign one queue to another, replacing the TARGET_OBJECT.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --| Works in O (n) time, where n is the maximum of the
  --| NUMBER_OF_ITEMS to be destroyed in the TARGET_OBJECT
  --| and the NUMBER_OF_ITEMS in the SOURCE_OBJECT.

  -- ..........................................................
  -- .                                                        .
  -- .  PRIORITIZED_QUEUE.DESTROY                             .  SPEC
  -- .                                                        .
  -- ..........................................................
  procedure Destroy (Targeted_Object : in out PRIORITY_QUEUE);
  --| Purpose
  --| Destroy a queue, freeing its contents.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --| Works in O (n) time, where n is the NUMBER_OF_ITEMS in the QUEUE.

private
  type PRIORITY_QUEUE_NODE;
  type PRIORITY_QUEUE is access PRIORITY_QUEUE_NODE;   

end PRIORITIZED_QUEUE;
