--::::::::::
--scanners.spc
--::::::::::
-- **************************************************
-- *                                                *
-- *  Scanners                                      *  SPEC
-- *                                                *
-- **************************************************
package Scanners is
--| Purpose
--| This package is used to break strings into tokens in a
--| very simple but efficient manner.  For maximum efficiency,
--| the scanner type is not private so that it can be used
--| directly.  The following conventions are adopted to allow
--| the Ada string handling primitives to be used to maximum
--| advantage:
--|
--|  1. Strings are never copied.  The scanner type contains
--|     First and Last components so that slices may be used
--|     to obtain the desired tokens (substrings).
--| 
--|  2. The scanner type does not include a copy of the
--|     string being scanned, also to avoid copying strings.
--| 
--|  3. The Length component of a scanner is always set to the
--|     length of the item scanned.  If it is zero it means
--|     that no such item was found, either because it wasn't
--|     there or because the scanner is exhausted. The is_Empty
--|     operation may be used to determint if a scanner is
--|     exhausted (usually before attempting to scan something).
--|
--|  4. All operations have well defined behavior for any
--|     consistent input. There are no exceptions declared in
--|     this package or raised directly by the operations in
--|     the package.
--|
--| Initialization Exceptions (none)
--|
--| Notes
--|    Not all MIL-HDBK-1804 PDL annotations are
--| used in this package due to its simplicity.
--|    No exceptions are raised by this package.
--|
--| Modifications
--| Author: Bill Toscano and Michael Gordon, Intermetrics

-- Types:
  type SCANNER_TYPE is record
    Index     : NATURAL;  -- Index of next char to be scanned
    Max_Index : NATURAL;  -- Index of last scannable char
    First     : NATURAL;  -- Index of 1st char of the result of a scan
    Last      : NATURAL;  -- Index of last char of the result of a scan
    Length    : NATURAL;  -- Length of the item scanned
  end record;

-- Constructors:

  -- ...............................................
  -- .                                             .
  -- .  Scanners.Start_Scanner                     .  SPEC
  -- .                                             .
  -- ...............................................
  procedure Start_Scanner (
    Scanner : in out SCANNER_TYPE;
    S       : in STRING;
    Last    : in NATURAL);
  --| Purpose
  --| Initialize Scanner for scanning S from S'FIRST to Last.
  --| S and Last are typically obtained by calling
  --| Text_IO.Get_Line.

  -- ...............................................
  -- .                                             .
  -- .  Scanners.Is_Empty                          .  SPEC
  -- .                                             .
  -- ...............................................
  function Is_Empty (Scanner: in SCANNER_TYPE)
    return BOOLEAN;
  pragma inline(is_Empty);
  --| Purpose
  --| Return True iff Scanner.Index > Scanner.Max_Index.
  --| Return TRUE iff there are more characters to scan.

  -- ...............................................
  -- .                                             .
  -- .  Scanners.Is_Alpha                          .  SPEC
  -- .                                             .
  -- ...............................................
  function Is_Alpha (Scanner : in SCANNER_TYPE;
                     S       : in STRING)
    return BOOLEAN;
  pragma inline(is_Alpha);
  --| Purpose
  --| Return True iff S(Scanner.Index) is an alphabetic character.

  -- ...............................................
  -- .                                             .
  -- .  Scanners.Is_Alpha                          .  SPEC
  -- .                                             .
  -- ...............................................
  function Is_Digit (Scanner : in SCANNER_TYPE;
                     S       : in string)
    return BOOLEAN;
  pragma inline(is_Digit);
  --| Purpose
  --| Return True iff S(Scanner.Index) is a decimal digit.

  -- ...............................................
  -- .                                             .
  -- .  Scanners.Is_Sign                           .  SPEC
  -- .                                             .
  -- ...............................................
  function Is_Sign (Scanner : in SCANNER_TYPE;
                    S       : in STRING)
    return BOOLEAN;
  pragma inline(is_Sign);
  --| Purpose
  --| Return True iff S(Scanner.Index) is '+' or '-'

  -- ...............................................
  -- .                                             .
  -- .  Scanners.Is_Digit_or_Sign                  .  SPEC
  -- .                                             .
  -- ...............................................
  function Is_Digit_or_Sign (Scanner : in SCANNER_TYPE;
                             S       : in string)
    return BOOLEAN;
  pragma inline(is_Digit_or_Sign);
  --| Purpose
  --| Return True iff S(Scanner.Index) is '+', '-', or a decimal digit.

  -- ...............................................
  -- .                                             .
  -- .  Scanners.Skip_Blanks                       .  SPEC
  -- .                                             .
  -- ...............................................
  procedure Skip_Blanks (Scanner : in out SCANNER_TYPE;
                         S       : in STRING);
  --| Purpose
  --| Increment Scanner.Index until S(Scanner.Index) is
  --| neither a blank nor a tab character, or until it is
  --| greater than Scanner.Max_Index.

  -- ...............................................
  -- .                                             .
  -- .  Scanners.Trim_Blanks                       .  SPEC
  -- .                                             .
  -- ...............................................
  procedure Trim_Blanks (Scanner : in out SCANNER_TYPE;
                         S       : in STRING);
  --| Purpose
  --| Adjust Scanner.First and Scanner.Last such that 
  --| S(Scanner.First..Scanner.Last) contains neither leading
  --| nor trailing blanks or tabs.  Scanner.Length is adjusted
  --| accordingly.  This is useful to remove blanks after a
  --| call to Scan_Delimited, Scan_Quoted, Scan_Until, etc.

  -- ...............................................
  -- .                                             .
  -- .  Scanners.Scan_Until                        .  SPEC
  -- .                                             .
  -- ...............................................
  procedure Scan_Until (Scanner : in out SCANNER_TYPE;
                        S       : in STRING;
                        C       : in CHARACTER);
  --| Purpose
  --| Scan in string S starting at Scanner.Index until the
  --| character C is encountered or the string ends.  On
  --| return, if Scanner.Length > 0 then
  --| S(Scanner.First..Scanner.Last) contains the characters that
  --| appeared before C and Scanner(Index) = C.  If C was
  --| not found, then the scanner is not affected except to
  --| set Scanner.Length to 0.

  -- ...............................................
  -- .                                             .
  -- .  Scanners.Scan_Word                         .  SPEC
  -- .                                             .
  -- ...............................................
  procedure Scan_Word (Scanner : in out SCANNER_TYPE;
                       S       : in STRING);
  --| Purpose
  --| Scan in string S for a sequence of non-blank characters,
  --| starting at Scanner.Index.  On return, if
  --| Scanner.Length > 0 then S(Scanner.First..Scanner.Last)
  --| is a word and Scanner.Index is just past the end of the
  --| word (Scanner.Last+1), ready to scan the next item.  

  -- ...............................................
  -- .                                             .
  -- .  Scanners.Scan_Number                       .  SPEC
  -- .                                             .
  -- ...............................................
  procedure Scan_Number (Scanner : in out SCANNER_TYPE;
                         S       : in STRING);
  --| Purpose
  --| Scan in string S for a sequence of numeric characters,
  --| optionally preceeded by a sign (+/-), starting at
  --| Scanner.Index.  On return, if Scanner.Length > 0 then
  --| S(Scanner.First..Scanner.Last) is a number and
  --| Scanner.Index is just past the end of the number
  --| (Scanner.Last+1), ready to scan the next item.

  -- ...............................................
  -- .                                             .
  -- .  Scanners.Scan_Delimited                    .  SPEC
  -- .                                             .
  -- ...............................................
  procedure Scan_Delimited (Scanner : in out SCANNER_TYPE;
                            S       : in STRING);
  --| Purpose
  --| The character S(Scanner.Index) is considered a "quote".  
  --| Scanner.First is set to the Scanner.Index+1, and
  --| Scanner.Index is incremented until another "quote"
  --| is encountered or the end of the string is reached.
  --| On return, Scanner.Last is the index of the closing
  --| "quote" or the last character in S if no closing "quote"
  --| was found.

  -- ...............................................
  -- .                                             .
  -- .  Scanners.Scan_Quoted                       .  SPEC
  -- .                                             .
  -- ...............................................
  procedure Scan_Quoted (Scanner : in out SCANNER_TYPE;
                         S       : in out STRING);
  --| Purpose
  --| The character S(Scanner.Index) is considered a "quote".  
  --| The string S is scanned for a closing "quote".  During
  --| the scan, two quotes in a row are replaced by a single
  --| quote.  On return, Scanner.First is the first character
  --| of the quoted string, and Scanner.Last is the last
  --| character.  (The outermost quotes are not included.)
  --| Scanner.Index is the first character after the
  --| closing quote, Scanner.Length is the number of characters
  --| in the quoted string.  Note that the string being scanned
  --| (S) is modified by this routine (to remove the extra quotes,
  --| if any).

end Scanners;
