--::::::::::
--search.spc
--::::::::::
-- *****************************************************
-- *                                                   *
-- *  SEARCH_UTILITIES                                 *  SPEC
-- *                                                   *
-- *****************************************************
with System;
generic
  type COMPONENT_TYPE is limited private;  -- type of component to search for
  type INDEX_TYPE     is (<>);             -- type of array index
  type ARRAY_TYPE is array (INDEX_TYPE range <>) of COMPONENT_TYPE;
  with function "<"(Left, Right : in COMPONENT_TYPE) return BOOLEAN is <>;
  with function "="(Left, Right : in COMPONENT_TYPE) return BOOLEAN is <>;
package Search_Utilities is
--| Purpose
--| Search_Utilities is a generic searching package. The Search subprograms
--| will search a one dimensional array of any data type
--| indexed by discrete type components.
--|
--| Note that the component type of the array is not restricted to simple
--| types. An array of records or allocators can be searched. If the
--| component type is a record or allocator, then the generic formal
--| subprogram parameter "<" below must be specified as a selector
--| function.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| Author: Geoff Mendal, Stanford University

  type DATA_ORDER_TYPE is (ORDERED, NOT_ORDERED);
  -- This type should be used to specify how the data is
  -- ordered.  The default is Not_Ordered.  However, significant CPU time
  -- can be saved if the data is ordered and the default, Not_Ordered,
  -- is overridden.
  --
  -- If the data are ordered, then if two or more components in the array
  -- can match the search component provided, then the component location
  -- returned by Search should be thought of as an arbitrary selection
  -- from amongst those possible match-components.
  --
  -- If the data are not ordered, then if two or more components in the
  -- array can match the search component provided, then the component
  -- location returned by Search will be the one closest to
  -- Search_Array'FIRST.

  type PERFORMANCE_INSTRUMENTATION_TYPE is range -1 .. System.Max_Int;
  -- This type declaration should be used to specify the
  -- instrumentation analysis data that can be returned by the
  -- Search procedure below.  -1 is only returned if an overflow in
  -- calculations has occurred.  The Search subprograms will not terminate
  -- if an overflow in instrumentation analysis data calculations has
  -- occurred.

  -- ....................................................
  -- .                                                  .
  -- .  SEARCH_UTILITIES.VERSION                        .  SPEC
  -- .                                                  .
  -- ....................................................
  function Version return STRING;
  --| Purpose
  --| Returns the version number of this package.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  SEARCH_UTILITIES.SEARCH                         .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Search (
    Component                        : in     COMPONENT_TYPE;
    Search_Array                     : in     ARRAY_TYPE;
    Location_Found                   :    out INDEX_TYPE;
    Component_Found                  :    out BOOLEAN;
    Number_of_Comparisons            :    out PERFORMANCE_INSTRUMENTATION_TYPE;
    Order_Strategy                   : in     DATA_ORDER_TYPE := NOT_ORDERED;
    No_Match_Index                   : in     INDEX_TYPE      := INDEX_TYPE'LAST);
  --| Purpose
  --| This procedure will search a one dimensional array of
  --| components.  It can search an ordered or unordered array.  If
  --| an ordered array is specified, it defaults to an ascending
  --| order (which can be overridden by the user).  The array components
  --| must only support equality, inequality, and assignment (private
  --| types).  The array indices can be of any discrete type.  The number
  --| of comparisons can also be returned.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  SEARCH_UTILITIES.SEARCH                         .  SPEC
  -- .                                                  .
  -- ....................................................
  procedure Search (
    Component                        : in     COMPONENT_TYPE;
    Search_Array                     : in     ARRAY_TYPE;
    Location_Found                   :    out INDEX_TYPE;
    Component_Found                  :    out BOOLEAN;
    Order_Strategy                   : in     DATA_ORDER_TYPE := NOT_ORDERED;
    No_Match_Index                   : in     INDEX_TYPE := INDEX_TYPE'LAST);
  --| Purpose
  --| This overloading of procedure Search should be used when
  --| no instrumentation analysis data are required.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  SEARCH_UTILITIES.SEARCH                         .  SPEC
  -- .                                                  .
  -- ....................................................
  function Search (
    Component                        : in COMPONENT_TYPE;
    Search_Array                     : in ARRAY_TYPE;
    Order_Strategy                   : in DATA_ORDER_TYPE := NOT_ORDERED)
    return BOOLEAN;
  --| Purpose
  --| This overloading of function Search should be used when
  --| the user only wants to know if the component exists or not.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  SEARCH_UTILITIES.SEARCH                         .  SPEC
  -- .                                                  .
  -- ....................................................
  function Search (
    Component                        : in COMPONENT_TYPE;
    Search_Array                     : in ARRAY_TYPE;
    Order_Strategy                   : in DATA_ORDER_TYPE := NOT_ORDERED;
    No_Match_Index                   : in INDEX_TYPE      := INDEX_TYPE'LAST)
    return INDEX_TYPE;
  --| Purpose
  --| This overloading of function Search should be used when
  --| the component is definitely known to exist and only the location
  --| is required.  (Note that No_Match_Index may be used to return a
  --| no match index value... but this won't work in all cases.)
  --|
  --| Exceptions (none)
  --| Notes (none)

end Search_Utilities;
