--::::::::::
--set.spc
--::::::::::
-- ************************************************
-- *                                              *
-- *  SET_PKG                                     *  SPEC
-- *                                              *
-- ************************************************
with Lists;
pragma Elaborate (Lists);
generic
    type ELEM_TYPE is private;
    with function Equal (E1, E2: ELEM_TYPE) return BOOLEAN is "=";
package Set_Pkg is
--| Purpose
--| This package provides the set abstract data type.  All standard set
--| operations are provided.  Standard mathematical set notation is
--| employed to describe the effects of the operations.
--|
--| The component type, and an equality relation used for membership
--| tests, are generic formals of the package. The implementation isn't
--| particularly fast, since the only available information about the
--| component type is the equality relation. However, this shouldn't be a
--| concern unless the sets become large or speed becomes important.
--| See scalar_set_pkg, hashed_set_pkg and ordered_set_pkg for other
--| implementations.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| Programmer: Ron Kownacki, Intermetrics
--| One of a family of set packages:

    type SET is private;

-- Exceptions:

    No_More: exception;       -- Raised on incorrect use of an iterator.

-- Iterators:
  
    type MEMBERS_ITER is private;  -- Members of a set in arbitrary order

-- Constructors:

  -- ...............................................
  -- .                                             .
  -- .  SET_PKG.CREATE                             .  SPEC
  -- .                                             .
  -- ...............................................
  function Create return SET;
  --| Purpose
  --| Return {}.  This operation is not strictly necessary, since an
  --| uninitialized set object is viewed as the empty set.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...............................................
  -- .                                             .
  -- .  SET_PKG.INSERT                             .  SPEC
  -- .                                             .
  -- ...............................................
  procedure Insert (S: in out SET;
                    E: in     ELEM_TYPE);
  --| Purpose
  --| Insert the element, e, into the set, s.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...............................................
  -- .                                             .
  -- .  SET_PKG.DELETE                             .  SPEC
  -- .                                             .
  -- ...............................................
  procedure Delete (S: in out SET;
                    E: in     ELEM_TYPE);
  --| Purpose
  --| If e is in s, then remove e from s.  Otherwise, no effect.
  --|
  --| Exceptions (none)
  --| Notes (none)
      
  -- ...............................................
  -- .                                             .
  -- .  SET_PKG.INTERSECT                          .  SPEC
  -- .                                             .
  -- ...............................................
  function Intersect (S1, S2: SET) return SET;
  --| Purpose
  --| Return {e | member(s1, e) and member(s2, e)}.
  --|
  --| Exceptions (none)
  --| Notes (none)
      
  -- ...............................................
  -- .                                             .
  -- .  SET_PKG.UNION                              .  SPEC
  -- .                                             .
  -- ...............................................
  function Union (S1, S2: SET) return SET;
  --| Purpose
  --| Return {e | member(s1, e) or member(s2, e)}.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...............................................
  -- .                                             .
  -- .  SET_PKG.COPY                               .  SPEC
  -- .                                             .
  -- ...............................................
  function Copy (S: SET) return SET;
  --| Purpose
  --| Returns a copy of s.  Subsequent changes to s will not be
  --| visible through the application of operations to the copy of s.
  --| Assignment or parameter passing without copying will result
  --| in a single set value being shared among objects.
  --| The assignment operation is used to transfer the values of
  --| the elem_type components of s; consequently, changes in these
  --| values may be observable through both sets if these types are
  --| access types, or if they contain access type components.
  --|
  --| Exceptions (none)
  --| Notes (none)
  
-- Query Operations

  -- ...............................................
  -- .                                             .
  -- .  SET_PKG.EQUAL                              .  SPEC
  -- .                                             .
  -- ...............................................
  function Equal (S1, S2: SET) return BOOLEAN;
  --| Purpose
  --| Return (for all e: elem_type (member(s1, e) iff member(s2, e))).
  --| Note that (s1 = s2) implies equal(s1, s2) holds for all time.
  --| "=" is object equality, equal is state equality.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...............................................
  -- .                                             .
  -- .  SET_PKG.IS_EMPTY                           .  SPEC
  -- .                                             .
  -- ...............................................
  function Is_Empty (S: SET) return BOOLEAN;
  --| Purpose
  --| Return s = {}.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...............................................
  -- .                                             .
  -- .  SET_PKG.IS_MEMBER                          .  SPEC
  -- .                                             .
  -- ...............................................
  function Is_Member (S: SET; E: ELEM_TYPE) return BOOLEAN;
  --| Purpose
  --| Return true iff e is a member of s.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...............................................
  -- .                                             .
  -- .  SET_PKG.SIZE                               .  SPEC
  -- .                                             .
  -- ...............................................
  function Size (S: SET) return NATURAL;
  --| Purpose
  --| Return |s|, the cardinality of s.
  --|
  --| Exceptions (none)
  --| Notes (none)

-- Iterators

  -- ...............................................
  -- .                                             .
  -- .  SET_PKG.MAKE_MEMBERS_ITER                  .  SPEC
  -- .                                             .
  -- ...............................................
  function Make_Members_Iter (S: SET) return MEMBERS_ITER;
  --| Purpose
  --| Create and return a members iterator based on s.  This object
  --| can then be used in conjunction with the more function and the
  --| next procedure to iterate over the members of s in some
  --| arbitrary order.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...............................................
  -- .                                             .
  -- .  SET_PKG.MORE                               .  SPEC
  -- .                                             .
  -- ...............................................
  function More (Iter: MEMBERS_ITER) return BOOLEAN;
  --| Purpose
  --| Return true iff the members iterator has not been exhausted.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...............................................
  -- .                                             .
  -- .  SET_PKG.NEXT                               .  SPEC
  -- .                                             .
  -- ...............................................
  procedure Next (Iter: in out MEMBERS_ITER;
                  E:    out    ELEM_TYPE);
  --| Purpose
  --| Let iter be based on the set, s.  Successive calls of next
  --| will return the members of s in some arbitrary order.
  --| After all members have been returned, then the procedure will
  --| raise no_more.
  --| Requires:
  --| s must not be changed between the invocations of
  --| make_nodes_iterator(g) and next.
  --|
  --| Exceptions
  --|   no_more
  --|
  --| Notes (none)

-- Heap management

  -- ...............................................
  -- .                                             .
  -- .  SET_PKG.DESTROY                            .  SPEC
  -- .                                             .
  -- ...............................................
  procedure Destroy (S: in out SET);
  --| Purpose
  --| Return space consumed by the set value associated with object
  --| s to the heap.  If other objects share the same set value, then
  --| further use of these objects is erroneous.  Components of type
  --| elem_type, if they are access types, are not garbage collected.
  --| It is the user's responsibility to dispose of these objects.
  --| s is set to {}.
  --|
  --| Exceptions (none)
  --| Notes (none)

private
    package List_Pkg is new Lists (ELEM_TYPE, Equal);
    use List_Pkg;
    
    type SET is new LIST;

      -- Representation Invariants:
      --   None; all lists are legal representations of sets.
      -- Abstraction Function:  A: representation --> set
      --   A(null) = create.
      --   A(attach(r, e)) = insert(A(r), e).
      --   Sufficient since all lists can be generated by null, attach.
      --
      --     Note that this implementation allows faster insertion and
      -- membership testing than if duplicate insertions of an element
      -- caused a check to ensure that each element is only kept once in
      -- the list.  This implies that deleting an element always involves
      -- a scan of the entire list.
    
    type MEMBERS_ITER is new LIST;

      -- For a set, s, make returns members_iter(copy(list(s))).
      -- More(iter) returns true iff list(iter) isn't empty.
      -- Next(iter) returns the first element in list(iter).  Before doing 
      -- this, it removes all occurrences of this element from list(iter).

end Set_Pkg;
