--::::::::::
--slist.spc
--::::::::::
-- *****************************************************************
-- *                                                               *
-- *  SINGLY_LINKED_LIST                                           *  SPEC
-- *                                                               *
-- *****************************************************************
generic
  type LIST_ELEMENT is private;
package Singly_Linked_List is
--| Purpose
--| This package provides an abstraction for a singly linked list.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| Author: Richard Conn

-- Types
  type LIST_TYPE is limited private;

-- Exceptions
  End_Error : exception;

  -- .............................................................
  -- .                                                           .
  -- .  SINGLY_LINKED_LIST.EMPTY                                 .  SPEC
  -- .                                                           .
  -- .............................................................
  function Empty (List : LIST_TYPE) return BOOLEAN;
  --| Purpose
  --| Indicates whether the list contains any elements.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .............................................................
  -- .                                                           .
  -- .  SINGLY_LINKED_LIST.NULL_NODE                             .  SPEC
  -- .                                                           .
  -- .............................................................
  function Null_Node (List : LIST_TYPE) return BOOLEAN;
  --| Purpose
  --| Indicates whether the "current pointer" references an element
  --| in the list.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .............................................................
  -- .                                                           .
  -- .  SINGLY_LINKED_LIST.HEAD_NODE                             .  SPEC
  -- .                                                           .
  -- .............................................................
  function Head_Node (List : LIST_TYPE) return BOOLEAN;
  --| Purpose
  --| Indicates whether the "current pointer" references the head
  --| of the list.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .............................................................
  -- .                                                           .
  -- .  SINGLY_LINKED_LIST.TAIL_NODE                             .  SPEC
  -- .                                                           .
  -- .............................................................
  function Tail_Node (List : LIST_TYPE) return BOOLEAN;
  --| Purpose
  --| Indicates whether the "current pointer" references the tail
  --| of the list.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .............................................................
  -- .                                                           .
  -- .  SINGLY_LINKED_LIST.CURRENT_ELEMENT                       .  SPEC
  -- .                                                           .
  -- .............................................................
  function Current_Element (List : LIST_TYPE) return LIST_ELEMENT;
  --| Purpose
  --| Returns the value of the element referenced by the "current pointer".
  --| Raises End_Error if Null_Node(List) = True.
  --|
  --| Exceptions
  --|   End_Error
  --|
  --| Notes (none)

  -- .............................................................
  -- .                                                           .
  -- .  SINGLY_LINKED_LIST.FIRST                                 .  SPEC
  -- .                                                           .
  -- .............................................................
  procedure First (List : in out LIST_TYPE);
  --| Purpose
  --| Positions the "current pointer" at the head of the list
  --| (even if the list is empty).
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .............................................................
  -- .                                                           .
  -- .  SINGLY_LINKED_LIST.NEXT                                  .  SPEC
  -- .                                                           .
  -- .............................................................
  procedure Next (List : in out LIST_TYPE);
  --| Purpose
  --| Positions the "current pointer" at the next element in the list.
  --| After the last element in the list, Null_Node(List) becomes True.
  --| Raises End_Error if Null_Node(List) = True.
  --|
  --| Exceptions
  --|   End_Error
  --|
  --| Notes (none)

  -- .............................................................
  -- .                                                           .
  -- .  SINGLY_LINKED_LIST.INSERT_AFTER                          .  SPEC
  -- .                                                           .
  -- .............................................................
  procedure Insert_After (List    : in out LIST_TYPE;
                          Element : LIST_ELEMENT);
  --| Purpose
  --| Inserts an element after the "current pointer".
  --| If Null_Node(List) = True the element is appended after
  --| the tail element.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .............................................................
  -- .                                                           .
  -- .  SINGLY_LINKED_LIST.INSERT_BEFORE                         .  SPEC
  -- .                                                           .
  -- .............................................................
  procedure Insert_Before (List    : in out LIST_TYPE;
                           Element : LIST_ELEMENT);
  --| Purpose
  --| Inserts an element before the "current pointer".
  --| If Null_Node(List) = True the element is prepended before
  --| the head element.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- .............................................................
  -- .                                                           .
  -- .  SINGLY_LINKED_LIST.DELETE_ELEMENT                        .  SPEC
  -- .                                                           .
  -- .............................................................
  procedure Delete_Element (List : in out LIST_TYPE);
  --| Purpose
  --| Deletes the element referenced by the "current pointer" from the list.
  --| Upon deletion, the "current pointer" references the element after the
  --| deleted element.
  --| Raises End_Error if Null_Node(List) = True.
  --|
  --| Exceptions
  --|   End_Error
  --|
  --| Notes (none)

  -- .............................................................
  -- .                                                           .
  -- .  SINGLY_LINKED_LIST.MODIFY                                .  SPEC
  -- .                                                           .
  -- .............................................................
  generic
    with procedure Transformation (Element : in out LIST_ELEMENT);
  procedure Modify (List : LIST_TYPE);
  --| Purpose
  --| Permits modification of the element referenced by the "current pointer"
  --| where the modification doesn't require external values (e.g.
  --| incrementing a field of the element).
  --| Raises End_Error if Null_Node(List) = True.
  --|
  --| Exceptions
  --|   End_Error
  --|
  --| Notes (none)

  -- .............................................................
  -- .                                                           .
  -- .  SINGLY_LINKED_LIST.UPDATE                                .  SPEC
  -- .                                                           .
  -- .............................................................
  generic
    type UPDATE_INFORMATION is private;
    with procedure Transformation (Element     : in out LIST_ELEMENT;
                                   Information : UPDATE_INFORMATION);
  procedure Update (List        : LIST_TYPE;
                    Information : UPDATE_INFORMATION);
  --| Purpose
  --| Permits modification of the element referenced by the "current pointer"
  --| where the modification requires external values (e.g. assigning a value
  --| to a field of the element).
  --| Raises End_Error if Null_Node(List) = True.
  --|
  --| Exceptions
  --|   End_Error
  --|
  --| Notes (none)

-- Pragmas
  pragma Inline (Empty, Null_Node, Head_Node, Tail_Node, Current_Element);
  pragma Inline (Modify, Update);

private
  type NODE;
  type NODE_ACCESS is access NODE;
  type NODE is
    record
      Element : LIST_ELEMENT;
      Next    : NODE_ACCESS;
    end record;
  type LIST_TYPE is
    record
      Head     : NODE_ACCESS;
      Tail     : NODE_ACCESS;
      Previous : NODE_ACCESS;
      Current  : NODE_ACCESS;
    end record;

end Singly_Linked_List;
