--::::::::::
--sort.spc
--::::::::::
-- *****************************************************
-- *                                                   *
-- *  SORT_UTILITIES                                   *  SPEC
-- *                                                   *
-- *****************************************************
with System;  -- predefined package SYSTEM
generic
  type COMPONENT_TYPE is private;  -- type of the data components
  type INDEX_TYPE     is (<>);     -- type of array index
  type ARRAY_TYPE is array (INDEX_TYPE range <>) of COMPONENT_TYPE;
  with function "<" (Left,Right : in COMPONENT_TYPE) return BOOLEAN is <>;
  with function Equal (Left,Right : in COMPONENT_TYPE) return BOOLEAN is "=";
package Sort_Utilities is
--| Purpose
--| Sort_Utilities is a generic sorting package. The Sort subprograms
--| will sort a one dimensional array of any component type that supports
--| assignment, equality, and inequality (private types) indexed by
--| discrete type components. The default linear order is ascending order
--| but may be overridden by the user. The default sort algorithm,
--| Quicksort (non-recursive), may also be overridden.
--|
--| Note that the component type can be a record type. The Sort subprograms
--| are not restricted to simple data types. If records are to be sorted,
--| then the formal generic subprogram parameter "<" must be
--| specified with by a linear order, e.g., a function provided
--| as an actual generic subprogram parameter at instantiation.
--|
--| Note that the component type can be an access type (which can
--| point to other objects, improving sort efficiency). If access types
--| are to be sorted, then the formal generic subprogram parameter "<"
--| must be specified by a linear order (see example #3 below).
--| Since access types can be sorted, the Sort routine below can be
--| used to sort limited types and unconstrained types (designated by
--| an access type).
--|
--| For data in which equality does not truly apply (i.e., real types)
--| one can use the Equal function to specify an equality operation.
--| Hence, one can decide that two numbers are "close enough" to be
--| equal (see example #4 below).
--|
--| The number of comparisons and exchanges made to sort the array
--| can be returned. These numbers should give some indication on how
--| much work was actually performed by the sorting algorithms. These
--| numbers can also be used to compare the relative efficiency
--| of the sorting algorithms.
--|
--| This package can be used to sort data on external devices. The user
--| should use this package to sort a subset of the external data, then
--| use a merge operation on all sorted subsets. For example, if the
--| system can only hold 1000 components in RAM, but you need to sort
--| 3000 components, bring in components #1-1000 and sort them using this
--| routine, and then write them to a file. Next do the same with
--| components #1001-2000, and finally with components #2001-3000. Now
--| merge the three sorted files using a merge package.
--|
--| One of the Sort subprograms is a function which can be used to sort
--| an array and test it against another in an inline expression. This
--| can be useful when comparing the contents of two arrays which may be
--| equal, but not at the identical indices. This will be most useful for
--| comparing the equality of sets implemented as arrays (see example #5
--| below).
--|
--| Other Sort subprograms allow the user to maintain the original state
--| of the array by returning a new array that is sorted. These subprograms
--| will be useful in cases where sorting is required, but the original
--| (unsorted) data must be preserved.
--|
--| Design of this package has been documented in the IEEE Computer
--| Society Second International Conference on Ada Applications and
--| Environments proceedings. Contact the IEEE or the author for a copy
--| of the paper. This paper is forthcoming in a special issue of IEEE
--| Software also.
--|
--| Initialization Exceptions (none)
--|
--| Notes
--| See the explanation below for details on the different
--| sort algorithms available and their respective merits.
--| There are also examples at the end of this specification
--| on the use of this package.
--|
--| Modifications
--| Author: Goeff Mendal, Stanford University

  type SORT_ALGORITHM_TYPE is (Quicksort, Recursive_Quicksort, Bsort,
    Bubble_Sort, Bubble_Sort_with_Quick_Exit, Selection_Sort, Heapsort,
    Insertion_Sort, Merge_Sort);
  -- Users can specify the type of sorting algorithm they want by
  -- specifying an enumeration literal from the type above. The default
  -- algorithm, Quicksort (non-recursive), generally performs best.
  --
  -- One note about stability of the algorithms: only the Bubble Sorts
  -- and Insertion Sort are stable algorithms. Thus, they are the
  -- only algorithms that preserve the ordering of equal components
  -- without use of a selector function. In all cases, a selector
  -- function may be specified to introduce stability into the
  -- sorting algorithms (see example #3 below).
  --
  -- Quicksort:   O(NlogN). Is most efficient when used with large, unsorted
  --              arrays. Uses an explicit stack to maintain state and
  --              partitions. Instable. This is the default algorithm.
  -- Recursive_Quicksort:  O(NlogN). Is most efficient when used with large,
  --              unsorted arrays. Recursive nature may introduce significant
  --              memory overhead for very large arrays. Instable.
  -- Bsort:       O(NlogN). Is most efficient when used with large arrays
  --              that are already sorted, partially sorted, or sorted in
  --              reverse. Recursive. Instable.
  -- Bubble_Sort: O(N**2). Is most efficient when used with small
  --              arrays that are almost already sorted. Non-recursive.
  --              Brute force. Low memory requirements. Stable.
  -- Bubble_Sort_with_Quick_Exit: O(N**2). Is most efficient when
  --              used with small arrays that are almost already sorted.
  --              Non-recursive. Same as bubble sort above except brute
  --              force is limited. Stable.
  -- Selection_Sort: O(N**2). Is most efficient when used with
  --              small arrays in which the Component_Type is a
  --              record type. Non-recursive. Brute force. Instable.
  -- Heapsort:    O(NlogN). Is most efficient when used with
  --              large, unsorted arrays. Non-recursive. Very low
  --              memory requirements. Instable.
  -- Insertion_Sort: O(N**2). Is most efficient when used with
  --              small arrays that are almost already sorted. Non-
  --              recursive. Brute force. Stable.
  -- Merge_Sort:  O(NlogN). Is most efficient when used with medium-large
  --              arrays. Non-recursive. Instable. Uses an auxiliary array
  --              to perform merging.


  type PERFORMANCE_INSTRUMENTATION_TYPE is range -1 .. SYSTEM.MAX_INT;
  -- This type declaration should be used to specify the
  -- instrumentation analysis results that can be returned by the Sort
  -- subprograms below. -1 is only returned if an overflow in calculations
  -- has occurred. The Sort subprograms will still sort the array if an
  -- overflow in instrumentation analysis data calculations
  -- occurs.

  Sort_Arrays_Length_Mismatch : exception;
  -- This exception is raised during execution of the Sort
  -- subprograms which take two arrays as parameters.  These two arrays
  -- must be of the same length.

  -- ...................................................
  -- .                                                 .
  -- .  SORT_UTILITIES.VERSION                         .  SPEC
  -- .                                                 .
  -- ...................................................
  function Version return STRING;
  --| Purpose
  --| Returns the version number of this package.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  SORT_UTILITIES.SORT                            .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Sort (
    Sort_Array             : in out ARRAY_TYPE;
    Number_of_Comparisons,
    Number_of_Exchanges    :    out PERFORMANCE_INSTRUMENTATION_TYPE;
    Sort_Algorithm         : in     SORT_ALGORITHM_TYPE := Quicksort);
  --| Purpose
  --| The following procedure will sort a one dimensional array of
  --| components. It can sort in ascending/descending order or any
  --| user-defined order. It can sort components of any type that
  --| support equality, inequality, and assignment (private types).
  --| The array indices can be of any discrete type. The number of
  --| comparisons and exchanges can also be returned.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ...................................................
  -- .                                                 .
  -- .  SORT_UTILITIES.SORT                            .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Sort (
    Sort_Array     : in out ARRAY_TYPE;
    Sort_Algorithm : in     SORT_ALGORITHM_TYPE := Quicksort);
  --| Purpose
  --| This overloading of procedure Sort should be specified
  --| when no instrumentation analysis data are required.
  --|
  --| Exceptions (none)
  --| Notes (none)
    
  -- ...................................................
  -- .                                                 .
  -- .  SORT_UTILITIES.SORT                            .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Sort (
    Unsorted_Array         : in     ARRAY_TYPE;
    Sorted_Array           :    out ARRAY_TYPE;
    Number_of_Comparisons,
    Number_of_Exchanges    :    out PERFORMANCE_INSTRUMENTATION_TYPE;
    Sort_Algorithm         : in     SORT_ALGORITHM_TYPE := Quicksort);
  --| Purpose
  --| The following overloading of procedure Sort should be used when
  --| the original data must be preserved and instrumentation analysis
  --| results are required.
  --|
  --| Exceptions
  --|   Sort_Arrays_Length_Mismatch is raised if Unsorted_Array
  --|                               and Sorted_Array are not
  --|                               the same length
  --|
  --| Notes (none)
  
  -- ...................................................
  -- .                                                 .
  -- .  SORT_UTILITIES.SORT                            .  SPEC
  -- .                                                 .
  -- ...................................................
  procedure Sort (
    Unsorted_Array : in     ARRAY_TYPE;
    Sorted_Array   :    out ARRAY_TYPE;
    Sort_Algorithm : in     SORT_ALGORITHM_TYPE := Quicksort);
  --| Purpose
  --| The following overloading of procedure Sort should be used when
  --| the original data must be preserved and no instrumentation analysis
  --| results are required.
  --|
  --| Exceptions
  --|   Sort_Arrays_Length_Mismatch is raised if Unsorted_Array
  --|                               and Sorted_Array are not
  --|                               the same length
  --|
  --| Notes (none)
    
  -- ...................................................
  -- .                                                 .
  -- .  SORT_UTILITIES.SORT                            .  SPEC
  -- .                                                 .
  -- ...................................................
  function Sort (
    Sort_Array     : in ARRAY_TYPE;
    Sort_Algorithm : in SORT_ALGORITHM_TYPE := Quicksort)
    return Array_Type;
  --| Purpose
  --| This overloading of function Sort should be used when
  --| sorting is required in an inline expression.

end Sort_Utilities;
