--::::::::::
--sscan.spc
--::::::::::
-- **********************************************
-- *                                            *
-- *  STRING_SCANNER                            *  SPEC
-- *                                            *
-- **********************************************
with String_Pkg;
use String_Pkg;
package String_Scanner is
--| Purpose
--| Functions for scanning tokens from strings.
--|
--| This package provides a set of functions used to scan tokens from
--| strings.  After the function make_Scanner is called to convert a string
--| into a string Scanner, the rest of the functions may be called to scan
--| various tokens from the string.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--| Modifications
--| Source: Bill Toscano and Michael Gordon, Intermetrics

  type SCANNER is private;

  Out_Of_Bounds : exception;
      -- Raised when a operation is attempted on a
      -- Scanner that has passed the end
  Scanner_Already_Marked : exception;
      -- Raised when a Mark is attemped on a Scanner
      -- that has already been marked

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.MAKE_SCANNER                .  SPEC
  -- .                                            .
  -- ..............................................
  function Make_Scanner (S : in STRING_TYPE) return SCANNER;
  --| Purpose
  --| Construct a Scanner from S.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.DESTROY_SCANNER             .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Destroy_Scanner (T : in out SCANNER);
  --| Purpose
  --| Free space occupied by the Scanner.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.MORE                        .  SPEC
  -- .                                            .
  -- ..............................................
  function More (T : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff additional characters remain to be scanned.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.FORWARD                     .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Forward (T : in SCANNER);
  --| Purpose
  --| Advance the scanner position.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.BACKWARD                    .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Backward (T : in SCANNER);
  --| Purpose
  --| Bump back the scanner position.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.GET                         .  SPEC
  -- .                                            .
  -- ..............................................
  function Get (T : in SCANNER) return CHARACTER;
  --| Purpose
  --| Return character at the current Scanner position.
  --| The scanner position remains unchanged.
  --|
  --| Exceptions
  --|   Out_Of_Bounds
  --|
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.NEXT                        .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Next (T : in     SCANNER;
                  C :    out CHARACTER);
  --| Purpose
  --| Return character at the current Scanner position and update
  --| the position.
  --|
  --| Exceptions
  --|   Out_Of_Bounds
  --|
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.POSITION                    .  SPEC
  -- .                                            .
  -- ..............................................
  function Position (T : in SCANNER) return POSITIVE;
  --| Purpose
  --| Return a positive integer indicating the current Scanner position,
  --|
  --| Exceptions
  --|   Out_Of_Bounds
  --|
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.GET_STRING                  .  SPEC
  -- .                                            .
  -- ..............................................
  function Get_String (T : in SCANNER) return STRING_TYPE;
  --| Purpose
  --| Return a String_Type corresponding to the contents of the Scanner
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.GET_REMAINDER               .  SPEC
  -- .                                            .
  -- ..............................................
  function Get_Remainder (T : in SCANNER) return STRING_TYPE;
  --| Purpose
  --| Return a String_Type starting at the current index of the Scanner
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.MARK                        .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Mark (T : in SCANNER);
  --| Purpose
  --| Mark the current index for possible future use.
  --|
  --| Exceptions
  --|   Scanner_Already_Marked
  --|
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.RESTORE                     .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Restore (T : in SCANNER);
  --| Purpose
  --| Restore the index to the previously marked value
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_WORD                     .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Word (T : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff Scanner is at the start of a word.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_WORD                   .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_word (T      : in SCANNER;
                       Found  : out BOOLEAN;
                       Result : out STRING_TYPE;
                       Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan T for a sequence of non-blank 
  --| characters.  If at least one is found, return Found => TRUE, 
  --| Result => <the characters>.
  --| Otherwise return Found => FALSE and Result is unpredictable.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_NUMBER                   .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Number (T : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff Scan_Number would return a non-null string (Scanner is
  --| at a decimal digit).
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_NUMBER                 .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Number (T      : in SCANNER;
                         Found  : out BOOLEAN;
                         Result : out STRING_TYPE;
                         Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan T for a sequence of digits.
  --| If at least one is found, return Found => TRUE, Result => <the digits>.
  --| Otherwise return Found => FALSE and Result is unpredictable.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_NUMBER                 .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Number (T      : in SCANNER;
                         Found  : out BOOLEAN;
                         Result : out INTEGER;
                         Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan T for a sequence of digits.
  --| If at least one is found, return Found => TRUE, Result => <the digits>.
  --| Otherwise return Found => FALSE and Result is unpredictable.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_SIGNED_NUMBER            .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Signed_Number (T : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff Scan_Signed_Number would return a non-null
  --| string and Scanner is at a decimal digit or sign (+/-).
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_SIGNED_NUMBER          .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Signed_Number (T      : in SCANNER;
                                Found  : out BOOLEAN;
                                Result : out STRING_TYPE;
                                Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan T for a sequence of digits preceeded with optional sign.
  --| If at least one digit is found, return Found => TRUE, 
  --| Result => <the digits>.
  --| Otherwise return Found => FALSE and Result is unpredictable.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_SIGNED_NUMBER          .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Signed_Number (T      : in SCANNER;
                                Found  : out BOOLEAN;
                                Result : out INTEGER;
                                Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan T for a sequence of digits preceeded with optional sign.
  --| If at least one digit is found, return Found => TRUE, 
  --| Result => <the digits>.
  --| Otherwise return Found => FALSE and Result is unpredictable.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_SPACE                    .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Space (T : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff Scan_Space would return a non-null string.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_SPACE                  .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Space (T      : in SCANNER;
                        Found  : out BOOLEAN;
                        Result : out STRING_TYPE);
  --| Purpose
  --| Scan T past all white space (spaces
  --| and tabs.  If at least one is found, return Found => TRUE,
  --| Result => <the characters>.
  --| Otherwise return Found => FALSE and Result is unpredictable.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SKIP_SPACE                  .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Skip_Space (T : in SCANNER);
  --| Purpose
  --| Scan T past all white space (spaces and tabs).  
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_ADA_ID                   .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Ada_Id (T : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff Scan_Ada_Id would return a non-null string.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_ADA_ID                 .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Ada_Id (T      : in SCANNER;
                         Found  : out BOOLEAN;
                         Result : out STRING_TYPE;
                         Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan T for a valid Ada identifier.
  --| If one is found, return Found => TRUE, Result => <the characters>.
  --| Otherwise return Found => FALSE and Result is unpredictable.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_QUOTED                   .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Quoted (T : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff T is at a quoted string (eg. ... "Hello" ...).
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_QUOTED                 .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Quoted (T      : in SCANNER;
                         Found  : out BOOLEAN;
                         Result : out STRING_TYPE;
                         Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan at T for an opening quote
  --| followed by a sequence of characters and ending with a closing
  --| quote.  If successful, return Found => TRUE, Result => <the characters>.
  --| Otherwise return Found => FALSE and Result is unpredictable.
  --| A pair of quotes within the quoted string is converted to a single quote.
  --| The outer quotes are stripped. 
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_ENCLOSED                 .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Enclosed (B : in CHARACTER;
                        E : in CHARACTER;
                        T : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff T as encosed by B and E (eg. ... [ABC] ...).
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_ENCLOSED               .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Enclosed (B      : in CHARACTER;
                           E      : in CHARACTER;
                           T      : in SCANNER;
                           Found  : out BOOLEAN;
                           Result : out STRING_TYPE;
                           Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan at T for an enclosing character
  --| followed by a sequence of characters and ending with an enclosing character.
  --| If successful, return Found => TRUE, Result => <the characters>.
  --| Otherwise return Found => FALSE and Result is unpredictable.
  --| The enclosing characters are stripped. 
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_SEQUENCE                 .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Sequence (Chars : in STRING_TYPE;
                        T     : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff T is at some character of Chars.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_SEQUENCE                 .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Sequence (Chars : in STRING;
                        T     : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff T is at some character of Chars.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_SEQUENCE               .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Sequence (Chars  : in STRING_TYPE;
                           T      : in SCANNER;
                           Found  : out BOOLEAN;
                           Result : out STRING_TYPE;
                           Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan T for a sequence of characters C such that C appears in 
  --| Char.  If at least one is found, return Found => TRUE, 
  --| Result => <the characters>.
  --| Otherwise return Found => FALSE and Result is unpredictable.
  --| Skip means to skip white characters before scanning.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --| Scan_Sequence("0123456789", S, Index, Found, Result)
  --| is equivalent to Scan_Number(S, Index, Found, Result)
  --| but is less efficient.

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_SEQUENCE               .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Sequence (Chars  : in STRING;
                           T      : in SCANNER;
                           Found  : out BOOLEAN;
                           Result : out STRING_TYPE;
                           Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan T for a sequence of characters C such that C appears in 
  --| Char.  If at least one is found, return Found => TRUE, 
  --| Result => <the characters>.
  --| Otherwise return Found => FALSE and Result is unpredictable.
  --| Skip means to skip white characters before scanning.
  --|
  --| Exceptions (none)
  --|
  --| Notes
  --| Scan_Sequence("0123456789", S, Index, Found, Result)
  --| is equivalent to Scan_Number(S, Index, Found, Result)
  --| but is less efficient.

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_NOT_SEQUENCE             .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Not_Sequence (Chars : in STRING_TYPE;
                            T     : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff T is not at some character of Chars.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_NOT_SEQUENCE             .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Not_Sequence (Chars : in STRING;
                            T     : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff T is not at some character of Chars.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_NOT_SEQUENCE           .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Not_Sequence (Chars  : in STRING_TYPE;
                               T      : in SCANNER;
                               Found  : out BOOLEAN;
                               Result : out STRING_TYPE;
                               Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan T for a sequence of characters C such that C does not appear
  --| in Chars.  If at least one such C is found, return Found => TRUE, 
  --| Result => <the characters>.
  --| Otherwise return Found => FALSE and Result is unpredictable.
  --| Skip means to skip white characters before scanning.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_NOT_SEQUENCE           .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Not_Sequence (Chars  : in STRING;
                               T      : in SCANNER;
                               Found  : out BOOLEAN;
                               Result : out STRING_TYPE;
                               Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan T for a sequence of characters C such that C does not appear
  --| in Chars.  If at least one such C is found, return Found => TRUE, 
  --| Result => <the characters>.
  --| Otherwise return Found => FALSE and Result is unpredictable.
  --| Skip means to skip white characters before scanning.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_LITERAL                  .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Literal (Chars : in STRING_TYPE;
                       T     : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff T is at literal Chars.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_LITERAL                  .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Literal (Chars : in STRING;
                       T     : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff T is at literal Chars.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_LITERAL                .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Literal (Chars  : in STRING_TYPE;
                          T      : in SCANNER;
                          Found  : out BOOLEAN;
                          Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan T for a literal Chars such that Char matches the sequence
  --| of characters in T.  If found, return Found => TRUE, 
  --| Otherwise return Found => FALSE
  --| Skip means to skip white characters before scanning.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_LITERAL                .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Literal (Chars  : in STRING;
                          T      : in SCANNER;
                          Found  : out BOOLEAN;
                          Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan T for a litral Chars such that Char matches the sequence
  --| of characters in T.  If found, return Found => TRUE, 
  --| Otherwise return Found => FALSE
  --| Skip means to skip white characters before scanning.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_NOT_LITERAL              .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Not_Literal (Chars : in STRING;
                           T     : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff T is not at literal Chars
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.IS_NOT_LITERAL              .  SPEC
  -- .                                            .
  -- ..............................................
  function Is_Not_Literal (Chars : in STRING_TYPE;
                           T     : in SCANNER) return BOOLEAN;
  --| Purpose
  --| Return TRUE iff T is not at literal Chars
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_NOT_LITERAL            .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Not_Literal (Chars  : in STRING;
                              T      : in SCANNER;
                              Found  : out BOOLEAN;
                              Result : out STRING_TYPE;
                              Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan T for a literal Chars such that Char does not match the
  --| sequence of characters in T.  If found, return Found => TRUE, 
  --| Otherwise return Found => FALSE
  --| Skip means to skip white characters before scanning.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................
  -- .                                            .
  -- . STRING_SCANNER.SCAN_NOT_LITERAL            .  SPEC
  -- .                                            .
  -- ..............................................
  procedure Scan_Not_Literal (Chars  : in STRING_TYPE;
                              T      : in SCANNER;
                              Found  : out BOOLEAN;
                              Result : out STRING_TYPE;
                              Skip   : in BOOLEAN := FALSE);
  --| Purpose
  --| Scan T for a litral Chars such that Char does not match the
  --| sequence of characters in T.  If found, return Found => TRUE, 
  --| Otherwise return Found => FALSE
  --| Skip means to skip white characters before scanning.
  --|
  --| Exceptions (none)
  --| Notes (none)

private
    type SCAN_RECORD is
      record
        Text  : STRING_TYPE;       -- Copy of string being scanned
        Index : POSITIVE := 1;     -- Current position of Scanner
        Mark  : NATURAL := 0;      -- Mark
      end record;
    type SCANNER is access SCAN_RECORD;

end String_Scanner;
