--::::::::::
--stack.spc
--::::::::::
-- **************************************************
-- *                                                *
-- *  STACK_PKG                                     *  SPEC
-- *                                                *
-- **************************************************
with Lists;
generic
    type ELEM_TYPE is private;
package Stack_Pkg is
--| Purpose
--| This package provides the stack abstract data type.  Element type is
--| a generic formal parameter to the package.  There are no explicit
--| bounds on the number of objects that can be pushed onto a given stack.
--| All standard stack operations are provided.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| Programmer: Ron Kownacki, Intermetrics

    type STACK is private;
    
-- Exceptions:
  
    Uninitialized_Stack: exception;
        -- Raised on attempt to manipulate an uninitialized stack object.
        -- The initialization operations are create and copy.

    Empty_Stack: exception;
        -- Raised by some operations when empty.

-- Constructors:

  -- ..............................................................
  -- .                                                            .
  -- .  STACK_PKG.CREATE                                          .  SPEC
  -- .                                                            .
  -- ..............................................................
  function Create return STACK;
  --| Purpose
  --| Return the empty stack.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ..............................................................
  -- .                                                            .
  -- .  STACK_PKG.PUSH                                            .  SPEC
  -- .                                                            .
  -- ..............................................................
  procedure Push (S: in out STACK; E: in ELEM_TYPE);
  --| Purpose
  --| Push e onto the top of s.
  --| Raises uninitialized_stack iff s has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_stack
  --|
  --| Notes (none)

  -- ..............................................................
  -- .                                                            .
  -- .  STACK_PKG.POP                                             .  SPEC
  -- .                                                            .
  -- ..............................................................
  procedure Pop (S: in out STACK);
  --| Purpose
  --| Pops the top element from s, and throws it away.
  --| Raises empty_stack iff s is empty.
  --| Raises uninitialized_stack iff s has not been initialized.
  --|
  --| Exceptions
  --|   empty_stack
  --|   uninitialized_stack
  --|
  --| Notes (none)

  -- ..............................................................
  -- .                                                            .
  -- .  STACK_PKG.POP                                             .  SPEC
  -- .                                                            .
  -- ..............................................................
  procedure Pop (S: in out STACK; E: out ELEM_TYPE);
  --| Purpose
  --| Pops the top element from s, returns it as the e parameter.
  --| Raises empty_stack iff s is empty.
  --| Raises uninitialized_stack iff s has not been initialized.
  --|
  --| Exceptions
  --|   empty_stack
  --|   uninitialized_stack
  --|
  --| Notes (none)
      
  -- ..............................................................
  -- .                                                            .
  -- .  STACK_PKG.COPY                                            .  SPEC
  -- .                                                            .
  -- ..............................................................
  function Copy (S: STACK) return STACK;
  --| Purpose
  --| Return a copy of s.
  --| Stack assignment and passing stacks as subprogram parameters
  --| result in the sharing of a single stack value by two stack
  --| objects; changes to one will be visible through the others.
  --| copy can be used to prevent this sharing.
  --| Raises uninitialized_stack iff s has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_stack
  --|
  --| Notes (none)
      
-- Queries:

  -- ..............................................................
  -- .                                                            .
  -- .  STACK_PKG.TOP                                             .  SPEC
  -- .                                                            .
  -- ..............................................................
  function Top (S: STACK) return ELEM_TYPE;
  --| Purpose
  --| Return the element on the top of s.  Raises empty_stack iff s is
  --| empty.
  --| Raises uninitialized_stack iff s has not been initialized.
  --|
  --| Exceptions
  --|   empty_stack
  --|   uninitialized_stack
  --|
  --| Notes (none)
      
  -- ..............................................................
  -- .                                                            .
  -- .  STACK_PKG.SIZE                                            .  SPEC
  -- .                                                            .
  -- ..............................................................
  function Size (S: STACK) return NATURAL;
  --| Purpose
  --| Return the current number of elements in s.
  --| Raises uninitialized_stack iff s has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_stack
  --|
  --| Notes (none)

  -- ..............................................................
  -- .                                                            .
  -- .  STACK_PKG.IS_EMPTY                                        .  SPEC
  -- .                                                            .
  -- ..............................................................
  function Is_Empty (S: STACK) return BOOLEAN;
  --| Purpose
  --| Return true iff s is empty.
  --| Raises uninitialized_stack iff s has not been initialized.
  --|
  --| Exceptions
  --|   uninitialized_stack
  --|
  --| Notes (none)

-- Heap Management:

  -- ..............................................................
  -- .                                                            .
  -- .  STACK_PKG.DESTROY                                         .  SPEC
  -- .                                                            .
  -- ..............................................................
  procedure Destroy (S: in out STACK);
  --| Purpose
  --| Return the space consumed by s to the heap.  No effect if s is
  --| uninitialized.  In any case, leaves s in uninitialized state.
  --|
  --| Exceptions (none)
  --| Notes (none)

private
    package Elem_List_Pkg is new Lists (ELEM_TYPE);
    subtype ELEM_LIST is Elem_List_Pkg.LIST;

    type STACK_REC is
        record
            Size: NATURAL := 0;
            Elts: ELEM_LIST := Elem_List_Pkg.Create;
        end record;
    type STACK is access STACK_REC;
    -- Let an instance of the representation type, r, be denoted by the
    -- pair, <size, elts>.  Dot selection is used to refer to these
    -- components.
    --
    -- Representation Invariants:
    --     r /= null
    --     elem_list_pkg.length(r.elts) = r.size.
    --
    -- Abstraction Function:
    --     A(<size, elem_list_pkg.create>) = stack_pkg.create.
    --     A(<size, elem_list_pkg.attach(e, l)>) = push(A(<size, l>), e).

end Stack_Pkg;
