--::::::::::
--testlog.spc
--::::::::::
-- **************************************************
-- *                                                *
-- *  Test_Log                                      *  SPEC
-- *                                                *
-- **************************************************
package Test_Log is
--| Purpose
--| A Test Log is a log of test activity.  Expected and
--| Actual test results can be reported to it via the Compare
--| routines, and Test Log can build a summary report of
--| the test results.  It keeps track of the number of
--| tests and the number of errors detected.
--|
--| Test_Log provides a number of Compare routines
--| that compare one value with another and two counters.
--| The Test Counter is incremented each time a Compare
--| routine is called and the Error counter is
--| incremented each time the comparison does not work
--| out.
--|
--| Test_Log performs its operations in one of three modes
--| which may be selected by calling the Set_Mode routine:
--|   SILENT           all results of calls to Compare are
--|                    logged internally and no display is
--|                    generated except when the Report
--|                    routine is called
--|   VERBOSE          all results of calls to Compare are
--|                    displayed on the console
--|   REPORT_TO_FILE   same as VERBOSE, except the results
--|                    are written to a file rather than to
--|                    the console
--|
--| A fourth "mode" is USER_SELECTABLE, which results in a
--| prompt being displayed to the user and the user selecting
--| one of the modes SILENT, VERBOSE, or REPORT_TO_FILE.
--|
--| Initialization Exceptions (none)
--| Notes
--|    The Test and Error counters are initially set to
--| zero.  They may be reset to zero at any time by
--| calling the Reset procedure.
--|    The Test and Error counters are of type NATURAL,
--| so care should be exercised to see that no more
--| tests than NATURAL'LAST are done before a Reset.
--|
--| Modifications
--| 2/27/91  Richard Conn   Initial Version and Release

  REPORT_FILE_ERROR : exception;
  -- raised if output report file cannot be created

  Test_Log_File : constant STRING := "testlog.rpt";
  -- Name of test log file (see next comment)

  type MODE is (SILENT, VERBOSE, REPORT_TO_FILE, USER_SELECTABLE);
  -- The Test Log can run silently, displaying a summary report
  -- at the end, or verbosely, displaying each comparison as it
  -- is done.  The REPORT_TO_FILE mode is the same as VERBOSE,
  -- but the output is sent to Test_Log_File rather than the
  -- console.  USER_SELECTABLE causes the user to be prompted at
  -- the console and manually select the SILENT, VERBOSE, or
  -- REPORT_TO_FILE modes.

  type TEST_RESULT is (FAIL, PASS);
  -- Values of the result of a test

  -- ..................................................
  -- .                                                .
  -- .  Test_Log.Set_Mode                             .  SPEC
  -- .                                                .
  -- ..................................................
  procedure Set_Mode (To : in MODE);
  --| Purpose
  --| The mode of operation is set to the indicated mode.
  --| See the discussion above for a description of the
  --| modes.
  --|
  --| Exceptions (none)
  --| Notes
  --|   If this routine is not called, the default mode
  --| is SILENT.

  -- ..................................................
  -- .                                                .
  -- .  Test_Log.Set_Test_ID_Field_Width              .  SPEC
  -- .                                                .
  -- ..................................................
  procedure Set_Test_ID_Field_Width (To : in NATURAL := 10);
  --| Purpose
  --| Set the length of a test ID to be output (up to 60).
  --| Any test ID string shorter than this length will be
  --| padded with spaces.  Any test ID string longer than
  --| this length will be output in full.
  --|
  --| Exceptions (none)
  --| Notes
  --|   If this routine is not called, the field width is
  --| automatically set to the default value.

  -- ..................................................
  -- .                                                .
  -- .  Test_Log.Set_String_Field_Width               .  SPEC
  -- .                                                .
  -- ..................................................
  procedure Set_String_Field_Width (To : in NATURAL := 20);
  --| Purpose
  --| Set the length of a string to be output (up to 60).
  --| Any string shorter than this length will be padded
  --| with spaces.  Any string longer than this length
  --| will be output in full.
  --|
  --| Exceptions (none)
  --| Notes
  --|   If this routine is not called, the field width is
  --| automatically set to the default value.

  -- ..................................................
  -- .                                                .
  -- .  Test_Log.Set_Integer_Field_Width              .  SPEC
  -- .                                                .
  -- ..................................................
  procedure Set_Integer_Field_Width (To : in NATURAL := 20);
  --| Purpose
  --| Set the length of an integer to be output.
  --| If the integer requires more space than this,
  --| the necessary space will be taken.
  --|
  --| Exceptions (none)
  --| Notes
  --|   If this routine is not called, the field width is
  --| automatically set to the default value.

  -- ..................................................
  -- .                                                .
  -- .  Test_Log.Set_Float_Field_Width                .  SPEC
  -- .                                                .
  -- ..................................................
  procedure Set_Float_Field_Width
    (Before_Decimal : in NATURAL := 2;
     After_Decimal  : in NATURAL := 5;
     In_Exponent    : in NATURAL := 4);
  --| Purpose
  --| Set the length of the fields of a floating point
  --| value to be output.  If In_Exponent is non-zero,
  --| scientific notation is used; if In_Exponent is
  --| zero, fixed point notation is used.
  --|
  --| Exceptions (none)
  --| Notes
  --|   If this routine is not called, the field widths are
  --| automatically set to the default values.

  -- ..................................................
  -- .                                                .
  -- .  Test_Log.Reset                                .  SPEC
  -- .                                                .
  -- ..................................................
  procedure Reset;
  --| Purpose
  --| The Reset routine resets the test and error counters.
  --| It need not be called the first time this package's
  --| routines are used since these counters come up
  --| initialized.
  --|
  --| Exceptions (none)

  -- ..................................................
  -- .                                                .
  -- .  Test_Log.Compare                              .  SPEC
  -- .                                                .
  -- ..................................................
  procedure Compare(Test_ID         : in STRING;
                    Expected_Result : in STRING;
                    Actual_Result   : in STRING);
  procedure Compare(Test_ID         : in STRING;
                    Expected_Result : in INTEGER;
                    Actual_Result   : in INTEGER);
  procedure Compare(Test_ID         : in STRING;
                    Expected_Result : in FLOAT;
                    Actual_Result   : in FLOAT;
                    Tolerance       : in FLOAT);
  --| Purpose
  --| These routines compare the two values (x1 and x2) for
  --| equality (except in the case of F1 and F2, which are
  --| compared by abs(F1-F2)<Tolerance).  If these values
  --| are equal or within tolerance, then only the
  --| Test counter is incremented.  If these values are
  --| not equal or within tolerance, the Test counter and
  --| Error counter are incremented and the Test_ID is
  --| displayed.
  --|
  --| If the Mode (see the Set_Mode procedure) is SILENT,
  --| the results are not shown.  If the Mode is VERBOSE
  --| or REPORT_TO_FILE, then the Test_ID, the Expected_Result,
  --| the Actual_Result, and the result of the comparison
  --| (FAIL or PASS) is written to the console (VERBOSE)
  --| or the output file Test_Log_File (REPORT_TO_FILE).
  --|
  --| Exceptions (none)

  -- ..................................................
  -- .                                                .
  -- .  Test_Log.Enter_Test_Result                    .  SPEC
  -- .                                                .
  -- ..................................................
  procedure Enter_Test_Result
      (Test_ID : in STRING;
       Result  : in TEST_RESULT);
  --| Purpose
  --| This routine enters Result as though a Compare call
  --| was made.  This is the same as calling one of the
  --| Compare routines, but the result of the comparison
  --| is the input value to this routine and no comparison
  --| is actually done.  This is useful when a test does
  --| not generate a value as a result, such as when the
  --| test expects an exception to be raised.
  --|
  --| See the Purpose section of the Compare routines
  --| for more information.
  --|
  --| Exceptions (none)

  -- ..................................................
  -- .                                                .
  -- .  Test_Log.Error_Count                          .  SPEC
  -- .                                                .
  -- ..................................................
  function Error_Count return NATURAL;
  --| Purpose
  --| Error_Count returns the value of the Error Counter.
  --|
  --| Exceptions (none)

  -- ..................................................
  -- .                                                .
  -- .  Test_Log.Test_Count                           .  SPEC
  -- .                                                .
  -- ..................................................
  function Test_Count return NATURAL;
  --| Purpose
  --| Test_Count returns the value of the Test Counter.
  --|
  --| Exceptions (none)

  -- ..................................................
  -- .                                                .
  -- .  Test_Log.Write                                .  SPEC
  -- .                                                .
  -- ..................................................
  procedure Write(Message : in STRING := "");
  --| Purpose
  --| Write the message to the console followed by a
  --| New Line.
  --|
  --| Exceptions (none)

  -- ..................................................
  -- .                                                .
  -- .  Test_Log.Report                               .  SPEC
  -- .                                                .
  -- ..................................................
  procedure Report(Message : in STRING := "");
  --| Purpose
  --| Print a report showing the values of the Test and
  --| Error Counters.  If Message is not null, it is
  --| printed, indented, before the counter values.
  --|
  --| Exceptions (none)

  -- ..................................................
  -- .                                                .
  -- .  Test_Log.Close                                .  SPEC
  -- .                                                .
  -- ..................................................
  procedure Close;
  --| Purpose
  --| Close the Test Log.  If a Test_Log_File is open,
  --| it is closed.
  --|
  --| Exceptions (none)

end Test_Log;
