--::::::::::
--tod.spc
--::::::::::
-- ****************************************************
-- *                                                  *
-- *  TOD_UTILITIES                                   *  SPEC
-- *                                                  *
-- ****************************************************
with Calendar;  -- Predefined (internal representation) TOD package.
package TOD_Utilities is
--| Purpose
--| This package will provide direct conversion from an external
--| time/date string to the internal Ada CALENDAR.TIME representation
--| and vice versa.  Most free format external representations are
--| supported.  Components of an external format include:
--|   Year, Month and Day (as numbers and strings), Hour, Minutes,
--|   and Seconds
--| As long as the external representation can be parsed unambiguously,
--| this package should be able to handle the conversion.  Examples of
--| legal external formats:
--|   7pm Fr March 12, 1982
--|   15 Dec. 84 12:36PM
--|   YESTERDAY 3PM
--|   6/01/83          <-- defaults to 12:00:00AM
--|   3:45AM           <-- defaults to the current date
--|   18:07:35         <-- defaults to the current date
--|   8-26             <-- defaults to 12:00:00AM of the current year
--|   friday           <-- defaults to 12:00:00AM of the current or next
--|                        future Friday
--| Examples of illegal external representations:
--|   2/31/84          <-- February never has a 31st day
--|   12:3605/01/84    <-- too tough to parse (nondeterminstic)
--|   3/8423:00:00     <-- too tough to parse (nondeterminstic)
--|   3:54:29AMTues    <-- too tough to parse (nondeterminstic)
--|   Nov 1983         <-- must always include day number in the date
--|   Sun 8/3/84       <-- 8/3/84 was a Friday
--|
--| Optional periods may be placed after ABBREVIATED day/month names.
--|
--| All external formats are converted to upper case, so there are no
--| problems with specifying mixed and/or lower case input.  All
--| results are returned in upper case by default (which can be overridden
--| by specifying lower case or mixed case).
--|
--| Special external formats: TODAY, TOMORROW, YESTERDAY, NOW
--| TODAY is equivalent to 12AM of the current date.  TOMORROW and
--| YESTERDAY are equivalent to the next/previous date.  NOW is
--| equivalent to calling the function CALENDAR.CLOCK.
--|
--| Defaults:
--|   If the year is omitted, it defaults to the current year.  If the
--|   time is omitted, it defaults to 12:00:00AM.  If the day name and no
--|   date is specified, the current or next future date is assumed.  If
--|   only the time is specified, the current date is assumed.  If the
--|   minutes and/or seconds are not specified in the time, they default
--|   to zero.  If the year is given in short format (1 or 2 digits) then
--|   it defaults to the current century.
--|
--| BNF for the external representation:
--|   {<special_format> [<time>] |
--|    [<time>] <special_format> |
--|    <day_string> &|* <date> &|* <time>}
--|
--|   <special_format> ::= {TODAY | TOMORROW | YESTERDAY | NOW}
--|
--|   <day_string> ::= SU|NDAY, MO|NDAY, ..., SA|TURDAY
--|
--|   <date> ::= {<month_number><sep1><day_number>[<sep1><year_number>] |
--|               <month_name><sep2><day_number>[<sep2><year_number>] |
--|               <day_number><sep2><month_name>[<sep2><year_number>] |
--|               <full_year_number><sep2><month_name><sep2><day_number> |
--|               <full_year_number><sep2><day_number><sep2><month_name>}
--|
--|   <time> ::= {<hour>':'<minutes>[':'<seconds>][<AM_PM>] |
--|               <AMPM_hour><AM_PM>}
--|
--|   <month_number> ::= 1 .. 12
--|   <month_name> ::= JAN|UARY, FEB|RUARY, ..., DEC|EMBER
--|   <day_number> ::= 1 .. 31
--|   <year_number> ::= {<short_year_number> | <full_year_number>}
--|   <short_year_number> ::= [0]0 .. 99    <-- for century 2000
--|                           [0]1 .. 99    <-- for century 2100
--|   <full_year_number> ::= 1901 .. 2099
--|   <sep1> ::= {'-'|'/'}
--|   <sep2> ::= {<sep1> | {' ' | ','} ...}
--|
--|   <hour> ::= [0]0 .. 24
--|   <AMPM_hour> ::= [0]1 .. 12
--|   <minutes> ::= 00 .. 59
--|   <seconds> ::= 00 .. 59
--|   <AM_PM> ::= {"AM" | "PM"}
--|
--|   Notes on the BNF above:
--|     Items in angle brackets must be separated by at least one
--|     blank and/or comma when they appear with exactly one space
--|     between them.
--|
--|     However, items in angle brackets which are not separated by
--|     exactly one blank have a more rigid syntax, and must be followed
--|     precisely as specified in the BNF.
--|
--|     Some characters/strings are enclosed in quotes to emphasize that
--|     they are explicit, and not metasymbols.  When specifying an
--|     external TOD_String, do NOT include the quotes.
--|
--|     The AM/PM indicator may be left off the time if at least the
--|     hours and minutes are specified.  If only the hour is specified,
--|     it must be in the range 01 .. 12 and must have the AM/PM
--|     indicator following it.  If the AM/PM indicator is left off a
--|     time format, AM is assumed unless the hour is in the range
--|     13 .. 23.  If the AM/PM indicator is included, the hour must
--|     be in the range 01 .. 12.
--|
--|     Notation:
--|       {...|...|...}    -- Select exactly one alternative.
--|       [...]            -- Optional.
--|       &|               -- Select one or the other or both,
--|       &|*              -- Same as &| with the extension of selecting
--|                           the items in any order.
--|       ' '              -- Encloses a character literal.
--|       " "              -- Encloses a string.
--|       < >              -- Encloses a non-terminal symbol.
--|       ...              -- Denotes a repeatable field.
--|       |                -- Separates alternatives and denotes legal
--|                        -- abbreviations.
--|
--| Initialization Exceptions (none)
--| Notes (none)
--|
--| Modifications
--| Author:  Geoff Mendal, Stanford University

  External_TOD_Representation_Length : constant POSITIVE := 38;
  subtype EXTERNAL_TOD_REPRESENTATION_TYPE is STRING (
    1 .. External_TOD_Representation_Length);
  -- This type should be used to retrieve an external TOD
  -- representation from the CALENDAR.TIME representation.

  type TYPE_SET is (UPPER_CASE, lower_case, Mixed_Case);
  -- This type should be used to specify the type set of an
  -- external representation returned by the internal-to-external
  -- function below.

  -- ..................................................
  -- .                                                .
  -- .  TOD_UTILITIES.VERSION                         .  SPEC
  -- .                                                .
  -- ..................................................
  function Version return STRING;
  --| Purpose
  --| Returns the version number of this package.
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  TOD_UTILITIES.CONVERT                           .  SPEC
  -- .                                                  .
  -- ....................................................
  function Convert (
    TOD_Value       : in CALENDAR.TIME;
    Default_Setting : in TYPE_SET := UPPER_CASE)
    return EXTERNAL_TOD_REPRESENTATION_TYPE;
  --| Purpose
  --| The following function will take the CALENDAR.TIME representation
  --| and return an external representation. The external representation
  --| has the following format:
  --|   Columns  1 ..  9 : Day as a string
  --|   Columns 11 .. 12 : Day as a number
  --|   Columns 14 .. 22 : Month as a string
  --|   Columns 24 .. 27 : year number
  --|   Columns 29 .. 38 : time in AM/PM format
  --|   All unused columns are blank
  --|
  --|  Example string returned:
  --|    "THURSDAY  09 AUGUST    1984 05:19:05PM"
  --|
  --| Exceptions (none)
  --| Notes (none)

  -- ....................................................
  -- .                                                  .
  -- .  TOD_UTILITIES.NOW                               .  SPEC
  -- .                                                  .
  -- ....................................................
  function Now (Default_Setting : in TYPE_SET := UPPER_CASE)
    return EXTERNAL_TOD_REPRESENTATION_TYPE;
  --| Purpose
  --| This function is a convenience, equivalent to calling
  --| the above Convert function with an argument of
  --| CALENDAR.CLOCK.  The current time and date are
  --| returned as specified for Convert above.
  --|
  --| Exceptions (none)
  --| Notes
  --|    Same as Convert(Calendar.Clock)

  -- ....................................................
  -- .                                                  .
  -- .  TOD_UTILITIES.CONVERT                           .  SPEC
  -- .                                                  .
  -- ....................................................
  function Convert (TOD_String : in STRING) return CALENDAR.TIME;
  --| Purpose
  --| This function will take an external TOD representation
  --| and return the CALENDAR.TIME representation.  The external
  --| representation can be any STRING object that conforms to
  --| the BNF given above.
  --|
  --| Exceptions (see below)
  --| Notes (none)

  Duplication_Error,                          -- "5/25/61 May 25 1961"
  Date_Error,                                 -- "2/31/75"
  Month_Number_Error,                         -- "13/1/1960"
  Year_Error,                                 -- "1/1/1900"
  Day_Number_Error,                           -- "1/32/1984"
  Day_Date_Error,                             -- "Sunday 8/3/84"
  Month_Missing_Error,                        -- "1961 25"
  Day_Number_Missing_Error,                   -- "1961 May"
  Hour_Error,                                 -- "25:00:00"
  Minute_Error,                               -- "23:61:00"
  Second_Error,                               -- "23:59:60"
  Time_String_Error,                          -- "1:05:05:PM"
  Abbreviation_Error,                         -- "Sept.emb. 5"
  External_Representation_Error : exception;  -- "blah blah blah"
  -- These exceptions will be raised if the input to the
  -- above function cannot be parsed unambiguously.  Also, this function
  -- traps CALENDAR.TIME_ERROR and instead raises the exception
  -- Date_Error below in its place.

end TOD_Utilities;
