
--\\BALANCED_TREES_VIS.ADA
--< PACKAGE : Balanced_Trees.
--  AUTHOR : SSgt R. Kirchner
--           SSgt C. Rasmussen
--           Sgt  D. Hamm
--  LAST MODIFIED : 12 Nov 85
--  BY : Lt Brooke
--  CHANGES MADE : Made Position_in_Tree a PRIVATE type (it used to
--                 be a LIMITED type).
--<
--< PACKAGE DESCRIPTION :
--<
--<      Provides the neccessary functions to create, use and
--< maintain balanced-tree indexing.  The trees created by this
--< package must have a minimum node size of Four keys. (Anything
--< less than four will cause errors in Procedure Delete_Key at the
--< present time.)
--<
--<      The package is generic, instantiated for the Type_Definition
--< you want ASSOCIATED with the STRING data stored within the trees.
--< The string data evaluated within the trees are known as KEYS, and
--< the generic parameter type associated with each KEY is called ITEMS.
--< The data is stored as strings because of the need for searching
--< on partial values.  When creating a new_tree, a length of the
--< strings to be stored within that tree must be specified.  All
--< values inserted into or deleted from that tree must be of that
--< length, and searching for a value of greater length is not allowed.
--< The Exception KEY_LENGTH_ERROR is raised in such cases.
--<
--< PACKAGE DEPENDENCIES :
--<
--<      None.
--<
generic
  type Items is private;
  No_Of_Keys : Positive := 7;
    -- Must not be less than four or Program_Error is raised.

package Balanced_Trees is

  subtype Keys  is String;

  type Trees            is private;
  type Position_In_Tree is private;


  Key_Length_Error   : exception;
  Key_Not_Found      : exception;
  Key_Already_Exists : exception;
  End_Of_Tree        : exception;

  function New_Tree (With_Key_Length : Integer) return Trees;
        -- DESCRIPTION : Creates a tree for strings of KEY_LENGTH.
        -- INPUTS      : With_Key_Length - length of strings.
        -- OUTPUTS     : Trees type pointing to the newly createed tree.
        -- EXCEPTIONS  : None.

  procedure Delete_Tree (Tree : in out Trees);
        -- DESCRIPTION : Deletes a tree and deallocates the space
        --               previously used by that tree.
        -- INPUTS      : Tree - tree to be deleted.
        -- OUTPUTS     : Null Trees type.
        -- EXCEPTIONS  : None.

  function Item_Where (Key_Is : Keys; In_Tree : Trees) return Items;
        -- DESCRIPTION : Returns the item associated to the first
        --               instance (partial Key searches) of the
        --               Key in a Tree.
        -- INPUTS      : In_Tree - The tree to be searched.
        --               Key_Is - The key to be searched for.
        -- OUTPUTS     : The keys related item (if key is found).
        -- EXCEPTIONS  : Key_Length_Error when Using_Key is of greater
        --               length than strings stored in tree.
        --               Key_Not_Found when Using_Key is not found.

  procedure Delete_Key (With_Value : Keys; From : Trees);
        -- DESCRIPTION : Deletes a key from a tree.
        -- INPUTS      : With_Value - Key to be deleted.
        --               From - Tree where key is to be deleted.
        -- OUTPUTS     : None.
        -- EXCEPTIONS  : Key_Length_Error when Key_Value is not the same
        --               length as keys stored in tree.
        --               Key_Not_Found when Key_Value is not found.

  procedure Insert (Key_Value : Keys; And_Item : Items; Into : Trees);
        -- DESCRIPTION : Inserts a key into a tree.
        -- INPUTS      : Key_Value - Key to be inserted.
        --               And_Item - Item to be associated with Key_Value
        --               Into - Tree where Key_Value is inserted.
        -- OUTPUTS     : None.
        -- EXCEPTIONS  : Key_Length_Error when Key_Value is not the same
        --               length as keys stored in tree.
        --               Key_Already_Exists when Key_Value already
        --               exists in the tree.

  procedure Get_First (From            : Trees;
		       Giving_Position : out Position_In_Tree;
		       Giving_Item     : out Items);
        -- DESCRIPTION : Returns the FIRST position within a tree
        --               and the item at that position.
        -- INPUTS      : From - Tree you want the first
        --               position/item from.
        -- OUTPUTS     : Giving_Position - First Position_In_Tree.
        --               Giving_Item - Item associated with the
        --               Giving_Position.
        -- EXCEPTIONS  : End_Of_Tree when tree has no entries.

  procedure Get_First_Key (From            : Trees;
			   With_Value      : Keys;
			   Giving_Position : out Position_In_Tree;
			   Giving_Item     : out Items);
        -- DESCRIPTION : Returns the position/item assosiated with the
        --               first instance of With_Value.
        -- INPUTS      : From - Tree you want the position/item from.
        --               With_Value - Key you are looking for the
        --               first instance of.
        -- OUTPUTS     : Giving_Position - Position_In_Tree of
        --               With_Value.
        --               And_Item - Item associated with the Key.
        -- EXCEPTIONS  : Key_Length_Error when Key_Value is of greater
        --               length than values stored in tree.
        --               Key_Not_Found when With_Value is not found.

  procedure Get_Last (From            : Trees;
		      Giving_Position : out Position_In_Tree;
		      Giving_Item     : out Items);
        -- DESCRIPTION : Returns the LAST position within a tree
        --               and the item at that position.
        -- INPUTS      : From - Tree you want the last
        --               position/item from
        -- OUTPUTS     : Giving_Position - Last Position_In_Tree.
        --               And_Item - Item associated with the
        --               Giving_Position.
        -- EXCEPTIONS  : End_Of_Tree when tree has no entries.

  procedure Get_Last_Key (From            : Trees;
			  With_Value      : Keys;
			  Giving_Position : out Position_In_Tree;
			  Giving_Item     : out Items);
        -- DESCRIPTION : Returns the position/item associated with the
        --               last instance of With_Value.
        -- INPUTS      : From - Tree you want position/item from.
        --               With_Value - Key you are looking for the
        --               last instance of.
        -- OUTPUTS     : Giving_Position - Position_In_Tree of
        --               With_Value.
        --               Giving_Item - Item associated with the key.
        -- EXCEPTIONS  : Key_Length_Error when Key_Value is of greater
        --               length than keys stored in tree.
        --               Key_Not_Found when Key_Value is not found.

  procedure Get_Next (From        : in out Position_In_Tree;
		      Giving_Item : out Items);
        -- DESCRIPTION : Allows you to get the NEXT sequential position
        --               in a tree and the item at that position.
        -- INPUTS      : From - your current position within a tree.
        -- OUTPUTS     : From - the NEXT position in the tree.
        --               Giving_Item  - The item at the NEXT position.
        -- EXCEPTIONS  : End_Of_Tree when there is not another position.

  procedure Get_Prior (From        : in out Position_In_Tree;
		       Giving_Item : out Items);
        -- DESCRIPTION : Allows you to get the PRIOR sequential position
        --               in a tree and the item at that position.
        -- INPUTS      : From - Your current position within a tree.
        -- OUTPUTS     : From - The PRIOR position in the tree.
        --               Giving_Item  - The item at the PRIOR position.
        -- EXCEPTIONS  : End_Of_Tree when there is not a previous position.

  procedure Change_Item (For_Key : Keys; In_Tree : Trees; To : Items);
        -- DESCRIPTION : Changes the item associated with a key.
        -- INPUTS      : For_Key - Key of item to be changed.
        --               In_Tree - Tree where item is to be changed.
        --               To - New item to replace the old item.
        -- OUTPUTS     : None.
        -- EXCEPTIONS  : Key_Length_Error when Key_Value is not the same
        --               length as keys stored in tree.
        --               Key_Not_Found when Key_Value is not found.

  procedure Change_Item (At_Position : Position_In_Tree; To : Items);
        -- DESCRIPTION : Changes the item at a certain tree_position.
        -- INPUTS      : At_Position - Position of item to be changed.
        --               To - New item to replace the old item.
        -- OUTPUTS     : None.
        -- EXCEPTIONS  : None.

  function Key_At (Position : Position_In_Tree) return Keys;
        -- DESCRIPTION : Returns the key located at a position within
        --               a tree.
        -- INPUTS      : Position - Position where the key is at.
        -- OUTPUTS     : The key at that position.
        -- EXCEPTIONS  : None.

  function Item_At (Position : Position_In_Tree) return Items;
        -- DESCRIPTION : Returns the item located at a position within
        --               a tree.
        -- INPUTS      : Position - Position where the item is at.
        -- OUTPUTS     : The item at that position.
        -- EXCEPTIONS  : None.

  function Inclusive_Subtree (Of_Tree  : Trees;
			      From_Key : Keys;
			      To_Key   : Keys) return Trees;
        -- DESCRIPTION : Creates a tree from part of another tree of
        --               everythig between two keys (INCLUSIVE).
        -- INPUTS      : Of_Tree - Initial tree.
        --               From_Key - Starting key of new tree
        --               To_Key - Ending key of new tree.
        -- OUTPUTS     : Trees type pointing to the new tree.
        -- EXCEPTIONS  : Key_Length_Error when either keys are greater
        --               length than keys stored in the original tree.

  function Exclusive_Subtree (Of_Tree  : Trees;
			      From_Key : Keys;
			      To_Key   : Keys) return Trees;
        -- DESCRIPTION : Creates a tree from part of another tree of
        --               everything between two keys (EXCLUSIVE).
        -- INPUTS      : Of_Tree - Initial tree.
        --               From_Key - Starting key of new tree
        --               To_Key - Ending key of new tree.
        -- OUTPUTS     : New tree.
        -- EXCEPTIONS  : Key_Length_Error when either keys are greater
        --               length than keys stored in the original tree.

private
  type Acc_Keys  is access Keys;
    -- Pointer to a Key.
  type Acc_Items is access Items;
    -- Pointer to a Item.
  type Datas is
    record
      Key  : Acc_Keys;
      Item : Acc_Items;
    end record;
      -- A record containing a matched pair of a Key and a Item.
  type Acc_Data   is access Datas;
    -- A pointer to a Datas containg the Key and Item
  type Tree_Array is array (1 .. No_Of_Keys + 1) of Trees;
    -- An Array of pointers.
  type Data_Array is array (1 .. No_Of_Keys) of Acc_Data;
    -- An Array of pointers to Datas.
  type Node;
  type Trees is access Node;
    -- A pointer to a Node.
  type Node is
    record
      Mother  : Trees;
      Pointer : Tree_Array;
      Data    : Data_Array;
    end record;
      -- A record containing a pointer to the parent Node, an array
      -- of pointers to lower Nodes and an array of pointers to the
      -- Data within the Node.
  type Position_In_Tree is
    record
      Node_In_Tree     : Trees;
      Position_In_Node : Integer;
    end record;
      -- A record containing a pointer to a node and a position
      -- within that node.
end Balanced_Trees;
