--\\VARIABLE_LISTS.ADA
-- last modified : 15 Nov 85
-- by            : TCB
-- changes made  : Added final touches to documentation.
--
--
with Unchecked_Deallocation,
     Text_Io;  -- This can be pulled if Self_Test is also pulled.

package body Variable_Lists is
--
--
--
-- Some notes on implementation:
--
-- The primary author took the cheap way out.  In order to avoid
-- constantly checking to see if there are any blocks at all in a
-- particular list, he insures that there is ALWAYS exactly ONE
-- empty block at the end of a list. e.g., if the block size is 5,
-- then when a list is declared, it will start with 0 Elements, but
-- one block of 5 empty slots. As Elements are added, when this last
-- empty block is used, a new empty block is added; so when item # 1
-- is added to the list, it is put in slot number 1 of the "old" last
-- block and a "new" last block is created. So there are 10 slots, only
-- the first of which is filled. As more Elements are added, it is not
-- until the last block is again touched (Element # 6) that a new block
-- is again allocated.  Similarly, as Elements are deleted, it is not
-- until there are TWO blank blocks at the end that one is deallocated.
-- However, since we always keep 1 blank one there, as soon as another
-- one BECOMES empty we know we can deallocate the last block.
-- This technique may waste a little space, but it makes everyday use
-- of lists run much more efficiently.
--
-- It may be expedient to include "Total_Elements : natural := 0" as
-- a component of the Lists type in the private part, and maitain this
-- value in procedures Add, Remove, and Empty.  This will make Item_At
-- and Number_In run much more efficiently.  However, this redundancy
-- of information may cause inconsistencies if there is an abnormal
-- termination.
--
--

  procedure Flush is new Unchecked_Deallocation
       (Object => List_Block, Name => Block_Ptr);


  procedure Deallocate (Ptr : in out Block_Ptr) is
	-- This is needed because Unchecked_Deallocation of a NULL
	-- access object raises nasty errors in the ROLM / DG  ADE.
  begin
    if Ptr /= null then
      Flush (Ptr);
    end if;
  end Deallocate;


  function Block_Number (N : Positive; Within : Lists) return Block_Ptr is
    List    : Lists renames Within;
    Current : Block_Ptr;
  begin
    Current := List.First_Block;

    for Block in 2 .. N loop
      Current := Current.Next;
    end loop;

    return Current;
  exception
    when Constraint_Error =>  raise Arent_That_Many_Elements;
  end Block_Number;


  function Block_Of (N : Positive) return Positive is
  begin
    return ((N - 1) / Block_Size) + 1;
  end Block_Of;


  function Last_Block_In (List : Lists) return Block_Ptr is
  begin
    return Block_Number (List.Number_Blocks, Within => List);
  end Last_Block_In;


  procedure Reclaim (From_Here : in out Block_Ptr) is
	-- Recovers ALL the blocks from this point on, not just
	-- the one pointed to in FROM_HERE.
    Current_Block, Next_Block : Block_Ptr;
  begin
    Current_Block := From_Here;
    Next_Block := Current_Block.Next;

    while Next_Block /= null loop
      Deallocate (Current_Block);
      Current_Block := Next_Block;
      Next_Block := Next_Block.Next;
    end loop;

    Deallocate (Current_Block);
    From_Here := null;
  exception
    when Constraint_Error =>  -- From_Here is ALREADY null, dummy.
      null;
  end Reclaim;


  function Row_Of (N : Positive) return Positive is
  begin
    return ((N - 1) mod Block_Size) + 1;
  end Row_Of;


  procedure Add (Item : Element; Onto : in out Lists) is
    Last_Block : Block_Ptr;
    List       : Lists renames Onto;
    Last_Entry : Count_Range renames List.Last_Block_Upto;
  begin
    Last_Block := Last_Block_In (List);
    Last_Entry := Last_Entry + 1; -- this gets a C_E if we try to put
                                  -- something in the last slot of a
                                  -- block.
    Copy (Item, Into => Last_Block.Item (Last_Entry));
  exception
    when Constraint_Error =>
              -- filled up the block. need a new one.
      Copy (Item, Into => Last_Block.Item (Block_Size));
      Last_Block.Next := new List_Block;
      Last_Entry := 0;
      List.Number_Blocks := List.Number_Blocks + 1;
  end Add;


  procedure Copy (Value : Lists; Into : in out Lists) is
    Source_List : Lists renames Value;
    Destination : Lists renames Into;
  begin
    Empty (Destination);

    for Current in 1 .. Number_In (Source_List) loop
      Add (Item_At (Current, Within => Source_List), Onto => Destination);
    end loop;
  end Copy;


  procedure Empty (List : in out Lists) is
  begin
    List.Last_Block_Upto := 0;
    List.Number_Blocks := 1;
    Reclaim (From_Here => List.First_Block.Next);
  end Empty;


  function Item_At (Number : Positive; Within : Lists) return Element is
    List : Lists renames Within;
  begin
    if Number <= Number_In( List ) then
      return Block_Number (Block_Of (Number), Within => List).Item
	      (Row_Of (Number));
    else
      raise Arent_That_Many_Elements;
    end if;
  end Item_At;


  function Number_In (List : Lists) return Natural is
  begin
    return (List.Number_Blocks - 1) * Block_Size + List.Last_Block_Upto;
  end Number_In;


  procedure Remove (Number : Positive; From : in out Lists) is
    List       : Lists renames From;
    Last_Block : Block_Ptr;
  begin
    if Number <= Number_In (List) then
      for Current_Item in Number + 1 .. Number_In (List) loop
	Replace (Current_Item - 1, Within => List,
		 By => Item_At (Current_Item, Within => List));
      end loop;

      List.Last_Block_Upto := List.Last_Block_Upto - 1;
    else
      raise Arent_That_Many_Elements;
    end if;
  exception
    when Constraint_Error =>
               -- now have an empty block.
      List.Last_Block_Upto := Count_Range'Last;
      Last_Block := Last_Block_In (List);
      Reclaim (Last_Block);
      List.Number_Blocks := List.Number_Blocks - 1;
      Last_Block_In (List).Next := null;
  end Remove;


  procedure Replace (Number : Positive;
		     Within : in out Lists;
		     By     : Element) is
    List  : Lists renames Within;
    Value : Element renames By;
  begin
    if Number <= Number_In ( List ) then
      Copy (Value,
	    Into =>
	      Block_Number (Block_Of (Number), Within => List).Item
	         (Row_Of (Number)));
    else
      raise Arent_That_Many_Elements;
    end if;
  end Replace;

  procedure Self_Test is
    List         : Lists;
    Error_Caught : Boolean := False;

    procedure Error_Is (Message : String) is
    begin
      Text_Io.Put_Line ("*** Variable_lists Self Test : " & Message);
      Error_Caught := True;
    end Error_Is;

  begin
    if Number_In (List) /= 0 then
      Error_Is ("Number_In not 0");

    elsif List.Number_Blocks /= 1 then
      Error_Is ("Number of Blocks not 1");

    elsif List.First_Block = null then
      Error_Is ("First_Block is null");
    end if;

    if Error_Caught then
      raise Program_Error;
    end if;
  end Self_Test;

begin
  Self_Test;
end Variable_Lists;
