--\\VARIABLE_LISTS_VIS.ADA
--< PACKAGE :        Variable_Lists
--  author        : Thomas C. Brooke
--  created on    : 3 Jul 85
--  last modified : 5 Jul 85
--  by            : TCB
--<
--< PACKAGE DESCRIPTION:
--<
--< This package provides true variable length lists.
--<
--< The items in a list are ordered and are given an ordinal number
--< starting at 1;  e.g. the first item added to a list is item # 1,
--< the second is item # 2, etc. Items can be retrieved directly by
--< their position number: >>> Item_At( 45, within => list_a ) <<<.
--< The length of a list is the Number_In that list, so iteration
--< can occur: >>> 1 .. Number_In( list_b ) <<< will reach every element
--< in list_b.  Items may be removed (although this process is in-
--< efficient). Lists may be emptied, this process is efficient.
--< Any item can be directly replaced by another element.
--<
--< Note that the implementation is a linked list of blocks.  This
--< makes it a relatively efficient infinite length, but there is
--< still direct retrieval of any single element.  The user may
--< specify the size of blocks as best fits his needs, if he so
--< desires.
--<
--< Lists are created merely by declaring an object of type Lists.
--< This new list is initially empty, so no explicit initialization
--< is required.
--<
--< Choosing the Block_By size may greatly affect performance.  The
--< larger Block_By is, the fewer links have to be traversed to get
--< to elements near the end of the list; however, a good deal more
--< space is wasted on the last partially full block.  A smaller
--< Block_By will waste less space, but may slow down processing in
--< the back end of the lists.
--<
--< PACKAGE DEPENDENCIES : none.
--
--
--
generic
  type Element is limited private;
	-- creates lists of Element, where Element can be any
        -- CONSTRAINED subtype.  Note that this specifically
        -- prohibits variant records, even with a default
        -- descriminant value.

  with procedure Copy (Value : Element; Into : in out Element);
	-- how to copy one Element into another.

  Block_By : Positive := 50;
	-- what size to make blocks of Elements.

package Variable_Lists is

  type Lists is limited private; -- because it may use access types.
                                 -- and we may have Lists of Lists.

  Arent_That_Many_Elements : exception;

  procedure Add       (Item : Element; Onto : in out Lists);
	-- DESCRIPTION : Adds ITEM to the list ONTO.
	-- INPUTS      : Item - The Element to be added to a list.
	--               Onto - The list to which ITEM is added.
	-- OUTPUTS     : Onto - the new list with the ITEM appended
	-- EXCEPTIONS  : should never raise any.

  procedure Copy      (Value : Lists; Into : in out Lists);
	-- DESCRIPTION : Empties the destination list INTO, then
	--               copies the contents of VALUE into INTO.
	-- INPUTS      : Value - the source list to copy
	--               Into  - the destination where it is copied into
	-- OUTPUTS     : Into  - the resulting list.
	-- EXCEPTIONS  : should not raise any.

  procedure Empty     (List : in out Lists);
	-- DESCRIPTION : Empties the LIST
	-- INPUTS      : List - the list to be emptied.
	-- OUTPUTS     : List - the resulting empty list.
	-- EXCEPTIONS  : should never raise any.

  function  Item_At   (Number : Positive; Within : Lists) return Element;
	-- DESCRIPTION : Returns the Element at position NUMBER in the
	--               list WITHIN.
	-- INPUTS      : Number - the ordinal position of the Element.
	--               Within - the list from which it is retrieved.
	-- OUTPUTS     : the Element at that position.
	-- EXCEPTIONS  : Arent_That_Many_Elements when NUMBER is
	--               greater than the Number_In the list.

  function  Number_In (List : Lists) return Natural;
	-- DESCRIPTION : Returns the number of Elements in LIST
	-- INPUTS      : List - the list to check the number in.
	-- OUTPUTS     : the number of Elements in LIST
	-- EXCEPTIONS  : should never raise any.

  procedure Remove    (Number : Positive; From : in out Lists);
	-- DESCRIPTION : Deletes Element NUMBER from the list FROM.
	-- INPUTS      : Number - Ordinal # of the Element to remove.
	--               From   - the list to remove it from.
	-- OUTPUTS     : From   - the list after the Element is removed.
	-- EXCEPTIONS  : Arent_That_Many_Elements when Number is
	--               greater than the Number_In the list FROM.

  procedure Replace   (Number : Positive;
		       Within : in out Lists;
		       By     : Element);
	-- DESCRIPTION : Replaces the Element at NUMBER in the list
	--               WITHIN by the Element BY.
	-- INPUTS      : Number - Ordinal number of Element to replace.
	--               Within - the list within which it is replaced.
	--               By     - the Element it is replaced by.
	-- OUTPUTS     : Within - the list with the Element replaced.
	-- EXCEPTIONS  : Arent_That_Many_Elements when Number is
	--               greater than the Number_In the list WITHIN.


private

  Block_Size : Positive renames Block_By;
  type Array_Of_Elements is array (1 .. Block_Size) of Element;
  type List_Block;
  type Block_Ptr is access List_Block;
  type List_Block is
    record
      Item : Array_Of_Elements;
      Next : Block_Ptr;
    end record;

  subtype Count_Range is Natural range 0 .. Block_Size - 1;

  type Lists is
    record
      Last_Block_Upto : Count_Range := 0;
      Number_Blocks   : Positive := 1;
      First_Block     : Block_Ptr := new List_Block;
    end record;

end Variable_Lists;
