generic
   
   type Key_Type is limited private;    -- Any user-supplied data type; the user-supplied LESS_THAN and EQUAL 
                                        --   functions, below, must consider every aspect of the data type in 
                                        --   performing their comparisons.  Non-key information is to be kept 
                                        --   separately, in the user-supplied NON_KEY_ITEM_CONTAINER data structure 
                                        --   which will be linked to each key value.

   type Non_Key_Item_Type is limited private;   -- The type of information to be associated with a key value.  

   type Non_Key_Item_Container is limited private;    -- This user-supplied data type will be used as the data
                                                      --   structure in which the non-key information is stored.  
                                                      --   It must be able to dynamically contain any natural 
                                                      --   number (0..infinity) of items.
                                                      --
                                                      -- The primary function of the B_PLUS_TREE is to add to,
                                                      --   delete from, and present for user analysis the instance 
                                                      --   of this data type which corresponds to a specified key value,
                                                      --   such that key values can be accessed either sequentially 
                                                      --   (in ascending or descending order) or randomly (via an 
                                                      --   automatically balanced tree structure) on demand.  This
                                                      --   implementation provides for very rapid access to the key
                                                      --   value which was most recently accessed, thus speeding up
                                                      --   sequences of requests involving the same key value.
                                                      --
                                                      -- Unfortunately, due to limitations of Ada, there is no way 
                                                      --   that this package can accept an arbitrary user-defined 
                                                      --   data type and still enforce the restriction that THE USER
                                                      --   MUST NOT MODIFY ANY INSTANCE OF THIS DATA STRUCTURE WHICH
                                                      --   THE B_PLUS_TREE PRESENTS FOR THE USER'S (READ-ONLY) ANALYSIS.
                                                      --   ALL MODIFICATION OF THIS DATA STRUCTURE MUST BE PERFORMED 
                                                      --   ONLY BY THE PROCEDURES SUPPLIED IN THIS PACKAGE!!!


   with function Empty (This_Non_Key_Item_Container : in Non_Key_Item_Container) return Boolean is <>;

   with procedure Assign (To_Non_Key_Item_Container   : in out Non_Key_Item_Container;
                          From_Non_Key_Item_Container : in     Non_Key_Item_Container) is <>;

   with procedure Insert (Container    : in out Non_Key_Item_Container;
                          Non_Key_Item : in     Non_Key_Item_Type      ) is <>;

   with procedure Delete (Container    : in out Non_Key_Item_Container;
                          Non_Key_Item : in     Non_Key_Item_Type      ) is <>;    

   with procedure Destroy_Contents (This_Non_Key_Item_Container : in out Non_Key_Item_Container) is <>;

      -- This procedure must destroy everything in the container, 
      --   in preparation for the destruction of the container itself.

   with procedure Assign (Target_Key : in out Key_Type;
                          Source_Key : in     Key_Type) is <>;

   with function Less_Than (First_Key  : in Key_Type;                       
                            Second_Key : in Key_Type) return Boolean is <>;   -- Ada won't allow a default to "<"..

   with function Equal (First_Key  : in Key_Type;
                        Second_Key : in Key_Type) return Boolean is <>;

package Generic_B_Plus_Tree is

   type B_Plus_Tree is limited private;

   Key_Does_Not_Exist_In_This_B_Plus_Tree            : EXCEPTION;
   No_Preceding_Key_Exists_In_This_B_Plus_Tree       : EXCEPTION;
   No_Following_Key_Exists_In_This_B_Plus_Tree       : EXCEPTION;
   No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree : EXCEPTION;


   type Pointer_To_B_Plus_Tree is access B_Plus_Tree;

   ---------------------------------------------------------------------------------------------------------------------------
   procedure Destroy (Targeted_Object : in out Pointer_To_B_Plus_Tree);

      -- Unlike Unchecked_Deallocation, this procedure will properly destroy the B_Plus_Tree being pointed to...
   ---------------------------------------------------------------------------------------------------------------------------
   procedure Insert_Item (Targeted_B_Plus_Tree : in out B_Plus_Tree;
                          Key_Value            : in     Key_Type;
                          Non_Key_Information  : in     Non_Key_Item_Type);
   ---------------------------------------------------------------------------------------------------------------------------
   procedure Delete_Item (Targeted_B_Plus_Tree : in out B_Plus_Tree;
                          Key_Value            : in     Key_Type;
                          Non_Key_Information  : in     Non_Key_Item_Type);

      -- Raises Key_Does_Not_Exist_In_This_B_Plus_Tree 
      --   or No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree, when appropriate...
   ---------------------------------------------------------------------------------------------------------------------------
   function Key_Exists (Targeted_B_Plus_Tree : in B_Plus_Tree;
                        Search_Key           : in Key_Type    ) return Boolean;
   ---------------------------------------------------------------------------------------------------------------------------
   function Number_Of_Keys_Stored (Targeted_B_Plus_Tree : in B_Plus_Tree) return Natural;
   ---------------------------------------------------------------------------------------------------------------------------
   function Get_Item_Container (Targeted_B_Plus_Tree : in B_Plus_Tree;
                                Search_Key           : in Key_Type    ) return Non_Key_Item_Container;

      -- This procedure involves copying the entire container... if your Non_Key_Item_Type is quite large,
      --   it may be advisable to implement it as a pointer to the "real" structures, thus reducing the copying
      --   burden per instance of the Non_Key_Item_Type to that of a single pointer.

      -- Raises Key_Does_Not_Exist_In_This_B_Plus_Tree 
      --   or No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree, when appropriate...
   ---------------------------------------------------------------------------------------------------------------------------
   function Get_First_Key (Targeted_B_Plus_Tree : in B_Plus_Tree) return Key_Type;

      -- Raises No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree when appropriate...
   ---------------------------------------------------------------------------------------------------------------------------
   function Get_Last_Key (Targeted_B_Plus_Tree : in B_Plus_Tree) return Key_Type;

      -- Raises No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree when appropriate...
   ---------------------------------------------------------------------------------------------------------------------------
   function A_Preceding_Key_Exists (Targeted_B_Plus_Tree : in B_Plus_Tree;
                                    Search_Key           : in Key_Type     ) return Boolean;
   ---------------------------------------------------------------------------------------------------------------------------
   function Get_Preceding_Key (Targeted_B_Plus_Tree : in B_Plus_Tree;
                               Search_Key           : in Key_Type     ) return Key_Type;

      -- Raises Key_Does_Not_Exist_In_This_B_Plus_Tree 
      --   or No_Preceding_Key_Exists_In_This_B_Plus_Tree
      --   or No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree, when appropriate...
   ---------------------------------------------------------------------------------------------------------------------------
   function A_Following_Key_Exists (Targeted_B_Plus_Tree : in B_Plus_Tree;
                                    Search_Key           : in Key_Type     ) return Boolean;
   ---------------------------------------------------------------------------------------------------------------------------
   function Get_Following_Key (Targeted_B_Plus_Tree : in B_Plus_Tree;
                               Search_Key           : in Key_Type     ) return Key_Type;

      -- Raises Key_Does_Not_Exist_In_This_B_Plus_Tree 
      --   or No_Following_Key_Exists_In_This_B_Plus_Tree 
      --   or No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree, when appropriate...
   ---------------------------------------------------------------------------------------------------------------------------
   procedure Delete_Key (Targeted_B_Plus_Tree : in out B_Plus_Tree;
                         Search_Key           : in     Key_Type);

      -- Raises Key_Does_Not_Exist_In_This_B_Plus_Tree 
      --   or No_Keys_Currently_Exist_In_This_Empty_B_Plus_Tree, when appropriate...
      --
      -- The Non_Key_Item_Container associated with this key will be emptied via the Destroy_Contents procedure.
   ---------------------------------------------------------------------------------------------------------------------------
   procedure Exchange (First_B_Plus_Tree  : in out B_Plus_Tree;
                       Second_B_Plus_Tree : in out B_Plus_Tree);

      -- Exchanges the values of First_B_Plus_Tree and Second_B_Plus_Tree in O(1) time.
   ---------------------------------------------------------------------------------------------------------------------------
   procedure Assign (To_B_Plus_Tree   : in out B_Plus_Tree;
                     From_B_Plus_Tree : in     B_Plus_Tree);
   ---------------------------------------------------------------------------------------------------------------------------
   procedure Destroy (Targeted_B_Plus_Tree : in out B_Plus_Tree);

      -- Destroys all keys and all associated containers, and renders the tree Empty.
   ---------------------------------------------------------------------------------------------------------------------------
   function Empty (Targeted_B_Plus_Tree: in B_Plus_Tree) return Boolean;
   ---------------------------------------------------------------------------------------------------------------------------

private

   type B_Plus_Tree_Descriptor;

   type B_Plus_Tree is access B_Plus_Tree_Descriptor;   

end Generic_B_Plus_Tree;


