------------------------------------------------------------------------
--
-- Indexed sequential files in Ada: a didactical example
--
-- See also: Ada-Belgium Newsletter, Vol.4, November 1996
--           Proceedings 1996 Ada-Belgium Seminar
--           "Ada: the solution to integrate technologies"
--           Brussels, November 22, 1996
--
------------------------------------------------------------------------
--
-- This file contains three Ada packages
--     - the specification of "Aux_io_exceptions".
--     - the specification of "Indexed_io".
--     - the body of "Indexed_io".
--
-- For the Gnat-compiler, the three packages should be put in three
-- different files, named respectively "Auxioexc.ads", "Indexeio.ads"
-- and "Indexeio.adb".
-- The files
--     indexeio.ads
--     indexeio.adb
--     auxioexc.ads
-- should be copied in the directory C:\DJGPP\ADAINC
--
-- After compilation, the files
--     indexeio.o
--     auxioexc.o
--     auxioexc.ali
-- should be put in the directory C:\DJGPP\LIB\ADALIB
--
-- Information concerning the invokation of the compiler or the linker
-- can be found in the Ada-Belgium repository.
--
------------------------------------------------------------------------
--
-- Marc A Gobin
-- Bleukeveldlaan 7
-- 3080 Tervuren
-- Tel : +32(2)767.69.52
-- Fax : +32(2)767.18.90
-- E-mail : Marc.Gobin@info.rma.ac.be
--
------------------------------------------------------------------------

package aux_io_exceptions is
   key_error : exception ;
   existence_error : exception ;
end aux_io_exceptions ;

------------------------------------------------------------------------

with io_exceptions, aux_io_exceptions, Ada.Direct_io ;
generic
   type element_type is private ;
package indexed_io is
   type file_type is limited private ;
   type file_mode is (in_file, inout_file, out_file) ;
   type relation_type is (greater_equal, equal, greater) ;
   procedure create (file : in out file_type ;
		     mode : in file_mode := inout_file ;
		     name : in string := "" ;
		     form : in string) ;
   procedure open (file : in out file_type ;
		   mode : in file_mode ;
		   name : in string ;
		   form : in string := "") ;
   procedure close (file : in out file_type) ;
   procedure delete (file : in out file_type) ;
   procedure reset (file : in out file_type ;
		    mode : in file_mode ;
		    key_number : in integer := 0) ;
   procedure reset (file : in out file_type ; key_number : in integer := 0) ;
   function mode (file : file_type) return file_mode ;
   function name (file : file_type) return string ;
   function form (file : file_type) return string ;
   function is_open (file : file_type) return boolean ;
   procedure read (file : in out file_type ; item : out element_type) ;
   generic
      type key_type is private ;
      default_key_number : integer := 0 ;
   procedure read_by_key (file : in out file_type ;
			  item : out element_type ;
			  key : in key_type ;
			  key_number : in integer := default_key_number ;
			  relation : in relation_type := equal) ;
   procedure write (file : in out file_type ; item : in element_type) ;
   procedure update (file : in out file_type ; item : in element_type) ;
   procedure delete_element (file : in out file_type) ;
   function end_of_file (file : file_type) return boolean ;
   status_error : exception renames io_exceptions.status_error ;
   mode_error : exception renames io_exceptions.mode_error ;
   name_error : exception renames io_exceptions.name_error ;
   use_error : exception renames io_exceptions.use_error ;
   device_error : exception renames io_exceptions.device_error ;
   end_error : exception renames io_exceptions.end_error ;
   data_error : exception renames io_exceptions.data_error ;
   key_error : exception renames aux_io_exceptions.key_error ;
   existence_error : exception renames aux_io_exceptions.existence_error ;
private
   subtype block is string (1..512) ;
   package dirio is new Ada.Direct_io (block) ;
   type file_tab is array (0..9) of dirio.file_type ;
   type key_prop is array (0..9) of integer ;
   type block_nr_typ is array (1..20) of integer ;
   type acc_block is access block ;
   type block_acc_typ is array (1..20) of acc_block ;
   type file_type is record
      data_file : dirio.file_type ;
      key_file : file_tab ;
      name : string (1..52) ;
      namlen, punt, size, formlen, num_keys, cur_key_no,
              num_rec, data_recup : integer ;
      mode : file_mode ;
      form : block ;
      key_length, key_position, key_levels, key_recup : key_prop ;
      block_nr, char_nr : block_nr_typ ;
      block_acc : block_acc_typ ;
      init : boolean := true ;
      opened : boolean := false ;
   end record ;
end indexed_io ;

------------------------------------------------------------------------

with Text_io, Ada.integer_Text_io ; use Text_io, Ada.integer_text_io ;
with Ada.Direct_io, Ada.Unchecked_Conversion ;
package body indexed_io is
   key_num_int, dummy : integer ;
   zone : block ;
   four_i : integer ;
   subtype str4 is string (1..4) ;
   key_num : character ;
   function string_to_int is new Ada.Unchecked_Conversion (str4, integer) ;
   function int_to_string is new Ada.Unchecked_Conversion (integer, str4) ;
   function get_endpt (file : in file_type ; level : integer) return integer is
      zone4 : string (1..4) ;
   begin
      zone4 := int_to_string (0) ;
      zone4(1..2) := file.block_acc(level).all(1..2) ;
      return string_to_int(zone4) ;
   end get_endpt ;

   procedure create (file : in out file_type ;
		     mode : in file_mode := inout_file ;
		     name : in string := "" ;
		     form : in string) is
      name1 : string (1..name'length+4) := name & "    " ;
      procedure analyse (form : in string) is  -- analyses form and fills
         debut : integer := 1 ;                -- the appropriate variables
         longueur : integer := form'length ;
         word : string (form'range) ;
         procedure get_word (eerste : in integer; lengte : out integer;
                             woord : out string) is
            ini, last, into : integer ;
         begin
            ini := eerste ;
            while ini <= form'length and then form(ini) = ' ' loop
               ini := ini + 1 ;
            end loop ;
            last := ini ; into := 0 ;
            if ini > form'length then
               lengte := 0 ;
            elsif form (ini) = ';' then
               woord (1) := ';' ;
               lengte := 1 ;
               debut := ini + 1 ;
            else
               while last <= form'length and then (form (last) /= ' ' and
                               form (last) /= ';') loop
                  into := into + 1 ;
                  if form (last) in 'a' .. 'z' then
                     woord (into) := character'val
                                    (character'pos (form (last)) - 32) ;
                  else
                     woord (into) := form (last) ;
                  end if ;
                  last := last + 1 ;
               end loop ;
               debut := last ;
               lengte := into ;
            end if ;

         procedure test_word (txt : in string; msg : in string) is
         begin
            if longueur /= txt'length or else word (1 .. longueur) /= txt then
               put_line (msg) ;
               raise use_error ;
            end if ;
         end test_word ;
         procedure test_word (karak : in character; msg : in string) is
         begin
            if longueur /= 1 or else word (1) /= karak then
               put_line (msg) ;
               raise use_error ;
            end if ;
         end test_word ;
      begin
         get_word (1, longueur, word) ;
         test_word ("RECORD", """Record"" missing in form.") ;
         get_word (debut, longueur, word) ;
         test_word (';', "';' missing in form.") ;
         get_word (debut, longueur, word) ;
         test_word ("SIZE", """Size"" missing for Record in form.") ;
         get_word (debut, longueur, word) ;
         begin
            get (word (1..longueur), file.size, dummy) ;
         exception
            when data_error | constraint_error =>
               put_line ("Wrong size value in form.") ;
               raise use_error ;
         end ;
         get_word (debut, longueur, word) ;
         test_word (';', "';' missing in form.") ;
         get_word (debut, longueur, word) ;
         key_num := '0' ;
         KEY_LOOP : loop
            key_num_int := character'pos(key_num) - 48 ;
            test_word ("KEY", """Key"" " & key_num & " missing in form.") ;
            get_word (debut, longueur, word) ;
            test_word (key_num, "Wrong key number in form.") ;
            get_word (debut, longueur, word) ;
            test_word (';', "';' missing in form.") ;
            get_word (debut, longueur, word) ;
            test_word ("LENGTH", """Length"" missing for Key " &
                          key_num & " in form.") ;
            get_word (debut, longueur, word) ;
            begin
               get (word (1..longueur), file.key_length(key_num_int), dummy) ;
               if file.key_length(key_num_int) > 100 then
                        raise constraint_error ; end if ;
            exception
               when data_error | constraint_error =>
                  put_line ("Wrong length value for Key " &
                               key_num & " in form.") ;
                  raise use_error ;
            end ;
            get_word (debut, longueur, word) ;
            test_word (';', "';' missing in form.") ;
            get_word (debut, longueur, word) ;
            test_word ("POSITION", """Position"" missing for Key " &
                             key_num & " in form.") ;
            get_word (debut, longueur, word) ;
            begin
               get (word (1..longueur), file.key_position(key_num_int), dummy) ;
               if file.key_position(key_num_int) +
                               file.key_length(key_num_int) > file.size then
                  raise constraint_error ;
               end if ;
            exception
               when data_error | constraint_error =>
                  put_line ("Wrong position value for key " &
                              key_num & " in form.") ;
                  raise use_error ;
            end ;
            get_word (debut, longueur, word) ;
            test_word (';', "';' missing in form.") ;
            get_word (debut, longueur, word) ;
            key_num := character'succ (key_num) ;
         exit key_loop when word (1) = '$' ;
         end loop key_loop ;
         file.num_keys := character'pos (key_num) - 48 ;
      end analyse ;
   begin
      if file.opened then raise status_error ; end if ;
      file.namlen := name'length ;
      while file.namlen > 0 and then name1(file.namlen) = ' ' loop
         file.namlen := file.namlen - 1 ;
      end loop ;
      if file.namlen = 0 then raise dirio.name_error ; end if ;
      file.name := (others => ' ') ;
      file.name (1 .. file.namlen) := name1 (1 .. file.namlen) ;
      file.punt := file.namlen ;
      while file.punt > 0 and then name1(file.punt) /= '.' loop
         file.punt := file.punt - 1 ;
      end loop ;
      if file.punt = 0 then
         file.punt := file.namlen + 1 ;
         file.namlen := file.namlen + 4 ;
         name1 (file.punt) := '.' ;
      end if ;
      analyse (form & '$') ;
      dirio.create (file.data_file, dirio.out_file, name) ;
      for i in 0 .. file.num_keys - 1 loop
         dirio.create (file.key_file(i), dirio.out_file,
                       name1(1..file.punt) & "id" & character'val(i+48)) ;
         file.key_levels(i) := 1 ;
         file.key_recup(i) := 0 ;
         dirio.reset (file.key_file(i), dirio.inout_file) ;
      end loop ;
      file.mode := mode ;
      file.form := (others => ' ') ;
      file.formlen := form'length ;
      if file.formlen > 512 then raise use_error ; end if ;
      file.form (1 .. form'length) := form ;
      file.data_recup := 512 * 3 + 1 ;
      zone (1..52) := file.name ;
      zone (53..56) := int_to_string (file.namlen) ;
      zone (57..60) := int_to_string (file.punt) ;
      zone (61..64) := int_to_string (file.size) ;
      zone (65..68) := int_to_string (file.formlen) ;
      zone (69..72) := int_to_string (file.num_keys) ;
      zone (73..76) := int_to_string (file.num_rec) ;
      zone (77..80) := int_to_string (file.data_recup) ;
      for i in 0 .. file.num_keys - 1 loop
         four_i := i * 4 ;
         zone(four_i + 113 .. four_i + 116) :=
                   int_to_string (file.key_length(i)) ;
         zone(four_i + 153 .. four_i + 156) :=
                   int_to_string (file.key_position(i)) ;
         zone(four_i + 193 .. four_i + 196) :=
                   int_to_string (file.key_levels(i)) ;
         zone(four_i + 233 .. four_i + 236) :=
                   int_to_string (file.key_recup(i)) ;
      end loop ;
      dirio.write (file.data_file, zone) ;
      dirio.write (file.data_file, file.form) ;
      zone := (others => ' ') ;
      zone(1..4) := int_to_string(0) ;
      dirio.write (file.data_file, zone) ;
      dirio.reset (file.data_file, dirio.inout_file) ;
      file.cur_key_no := 0 ;
      file.num_rec := 0 ;
      file.init := true ;
      file.opened := true ;
   end create ;

   procedure open (file : in out file_type ;
		   mode : in file_mode ;
		   name : in string ;
		   form : in string := "") is
      zone4 : str4 ;
   begin
      if file.opened then raise status_error ; end if ;
      dirio.open (file.data_file,
                  dirio.file_mode'value (file_mode'image(mode)), name) ;
      dirio.read (file.data_file, zone) ;
      dirio.read (file.data_file, file.form) ;
      file.name := zone (1..52) ;
      zone4 := zone(53..56) ; file.namlen := string_to_int(zone4) ;
      zone4 := zone(57..60) ; file.punt := string_to_int (zone4) ;
      zone4 := zone(61..64) ; file.size := string_to_int (zone4) ;
      zone4 := zone(65..68) ; file.formlen := string_to_int (zone4) ;
      zone4 := zone(69..72) ; file.num_keys := string_to_int (zone4) ;
      zone4 := zone(73..76) ; file.num_rec := string_to_int (zone4) ;
      zone4 := zone(77..80) ; file.data_recup := string_to_int (zone4) ;
      for i in 0 .. file.num_keys - 1 loop
         four_i := i * 4 ;
         zone4 := zone(four_i + 113 .. four_i + 116) ;
            file.key_length(i) := string_to_int (zone4) ;
         zone4 := zone(four_i + 153 .. four_i + 156) ;
            file.key_position(i) := string_to_int (zone4) ;
         zone4 := zone(four_i + 193 .. four_i + 196) ;
            file.key_levels(i) := string_to_int (zone4) ;
         zone4 := zone(four_i + 233 .. four_i + 236) ;
            file.key_recup(i) := string_to_int (zone4) ;
         dirio.open (file.key_file(i),
                dirio.file_mode'value (file_mode'image(mode)),
                file.name(1..file.punt-1) & ".id" & character'val(i+48)) ;
      end loop ;
      file.mode := mode ;
      file.cur_key_no := 0 ;
      file.init := true ;
      file.opened := true ;
   end open ;

   procedure save (file : in file_type) is
      zone : block := (others => ' ') ;
   begin
      zone (1..52) := file.name ;
      zone (53..56) := int_to_string (file.namlen) ;
      zone (57..60) := int_to_string (file.punt) ;
      zone (61..64) := int_to_string (file.size) ;
      zone (65..68) := int_to_string (file.formlen) ;
      zone (69..72) := int_to_string (file.num_keys) ;
      zone (73..76) := int_to_string (file.num_rec) ;
      zone (77..80) := int_to_string (file.data_recup) ;
      for i in 0 .. file.num_keys - 1 loop
         four_i := i * 4 ;
         zone(four_i + 113 .. four_i + 116) :=
                   int_to_string (file.key_length(i)) ;
         zone(four_i + 153 .. four_i + 156) :=
                   int_to_string (file.key_position(i)) ;
         zone(four_i + 193 .. four_i + 196) :=
                   int_to_string (file.key_levels(i)) ;
         zone(four_i + 233 .. four_i + 236) :=
                   int_to_string (file.key_recup(i)) ;
      end loop ;
      dirio.write (file.data_file, zone, 1) ;
   end save ;

   procedure search (file : in out file_type; given_key : in string;
                        key_no : in integer; equal : out boolean;
                        largest : out boolean) is
      -- zoekt een gegeven key in een een bestand, voor elk niveau
      -- worden de parameters bepaald
      zone : str4 ;
      beg, endpt, len : integer ;
      zonebidon : block ;
   begin
      largest := false ;
      equal := false ;
      len := file.key_length(key_no) + 4 ;
      file.block_nr (1) := 1 ;
      NIVO : for level in 1 .. file.key_levels (key_no) loop
         dirio.read (file.key_file(key_no), zonebidon,
                     dirio.count(file.block_nr(level))) ;
         if file.block_acc (level) = null then
            file.block_acc (level) := new block' (zonebidon) ;
         else
            file.block_acc (level).all := zonebidon ;
         end if ;
         endpt := get_endpt (file, level) ;
         beg := 3 ;
         SRC : loop
            if beg >= endpt then
               largest := true ;
               beg := beg - len ;
               exit SRC ;
            else
               if file.block_acc(level).all
                     (beg..beg+len-5) >= given_key then
                  equal := file.block_acc(level).all
                     (beg..beg+file.key_length(key_no)-1) = given_key ;
                  exit SRC ;
               end if ;
               beg := beg + len ;
            end if ;
         end loop SRC ;
         file.char_nr (level) := beg ;
         zone := file.block_acc(level).all(beg+len-4..beg+len-1) ;
         file.block_nr(level+1) := string_to_int (zone) ;
      end loop NIVO ;
   end search ;

   procedure close (file : in out file_type) is
   begin
      save (file) ;
      if not file.opened then raise status_error ; end if ;
      dirio.close (file.data_file) ;
      for i in 0 .. file.num_keys - 1 loop
         dirio.close (file.key_file(i)) ;
      end loop ;
      file.opened := false ;
   end close ;

   procedure delete (file : in out file_type) is
   begin
      if not file.opened then raise status_error ; end if ;
      dirio.delete (file.data_file) ;
      for i in 0 .. file.num_keys - 1 loop
         dirio.delete (file.key_file(i)) ;
      end loop ;
   end delete ;

   procedure reset (file : in out file_type ;
		    mode : in file_mode ;
		    key_number : in integer := 0) is
   begin
      if not file.opened then raise status_error ; end if ;
      file.mode := mode ;
      file.cur_key_no := key_number ;
      file.init := true ;
   exception
      when others =>
         if file.opened then save (file) ; end if ;
         raise ;
   end reset ;

   procedure reset (file : in out file_type ;
                         key_number : in integer := 0) is
   begin
      if not file.opened then raise status_error ; end if ;
      file.cur_key_no := key_number ;
      file.init := true ;
   exception
      when others =>
         if file.opened then save (file) ; end if ;
         raise ;
   end reset ;

   function mode (file : file_type) return file_mode is
   begin
      if not file.opened then raise status_error ; end if ;
      return file.mode ;
   end mode ;

   function name (file : file_type) return string is
   begin
      if not file.opened then raise status_error ; end if ;
      return file.name(1..file.namlen) ;
   end name ;

   function form (file : file_type) return string is
   begin
      if not file.opened then raise status_error ; end if ;
      return file.form(1..file.formlen) ;
   end form ;

   function is_open (file : file_type) return boolean is
   begin
      return file.opened ;
   end is_open ;

   procedure read (file : in out file_type ; item : out element_type) is
      dum_key : string (1 .. file.key_length(file.cur_key_no)) ;
      equal, largest : boolean ;
      level, endpt, beg, len, ptr, bloknr, char : integer ;
      zone4 : str4 ;
      subtype strsiz is string (1..file.size) ;
      strzone : strsiz ;
      blk : block ;
      function string_to_elem is new
                Ada.Unchecked_Conversion (strsiz, element_type) ;
   begin
      if not file.opened then raise status_error ; end if ;
      if file.mode = out_file then raise mode_error ; end if ;
      len := file.key_length(file.cur_key_no) + 4 ;
      if file.init then
         file.init := false ;
         dum_key := (others => character'val(0)) ;
         search (file, dum_key, file.cur_key_no, equal, largest) ;
         if largest then raise end_error ; end if ;
      else -- not file.init
         level := file.key_levels(file.cur_key_no) ;
         REPEAT : loop
            endpt := get_endpt(file, level) ;
            exit REPEAT when file.char_nr(level) + len < endpt ;
            if level = 1 then raise end_error ; end if ;
            level := level - 1 ;
         end loop REPEAT ;
         file.char_nr(level) := file.char_nr(level) + len ;
         while level /= file.key_levels(file.cur_key_no) loop
            beg := file.char_nr(level) ;
            zone4 := file.block_acc(level).all (beg+len-4..beg+len-1) ;
            file.block_nr(level+1) := string_to_int (zone4) ;
            level := level + 1 ;
            dirio.read (file.key_file(file.cur_key_no),
                        file.block_acc(level).all,
                        dirio.count(file.block_nr(level))) ;
            file.char_nr(level) := 3 ;
        end loop ;
      end if ;
      beg := file.char_nr(file.key_levels(file.cur_key_no)) ;
      zone4 := file.block_acc(file.key_levels(file.cur_key_no)).all
                             (beg+len-4..beg+len-1) ;
      bloknr := string_to_int (zone4) / 512 ;
      char := string_to_int (zone4) rem 512 ;
      ptr := 1 ;
      TFT : loop
         dirio.read (file.data_file, blk, dirio.count(bloknr)) ;
         exit TFT when 512 - char >= file.size - ptr ;
         strzone (ptr..ptr+512-char) := blk (char..512) ;
         ptr := ptr + 513 - char ;
         char := 1 ;
         bloknr := bloknr + 1 ;
      end loop TFT ;
      strzone (ptr..file.size) := blk (char..char+file.size-ptr) ;
      item := string_to_elem (strzone) ;
   exception
      when others =>
         if file.opened then save (file) ; end if ;
         raise ;
   end read ;

   procedure read_by_key (file : in out file_type ;
			  item : out element_type ;
			  key : in key_type ;
			  key_number : in integer := default_key_number ;
			  relation : in relation_type := equal) is
      pos : integer := file.key_position(key_number) + 1 ;
      len : integer := file.key_length(key_number) ;
      ptr, bloknr, char, beg : integer ;
      subtype key_str_type is string (1..len) ;
      key_str : key_str_type ;
      gelijk, groter : boolean ;
      subtype strsiz is string (1..file.size) ;
      strzone : strsiz ;
      zone4 : str4 ;
      blk : block ;
      function string_to_elem is new
                Ada.Unchecked_Conversion (strsiz, element_type) ;
      function elem_to_string is new
                Ada.Unchecked_Conversion (element_type, strsiz) ;
      function key_to_string is new
                Ada.Unchecked_Conversion (key_type, key_str_type) ;
   begin
      if not file.opened then raise status_error ; end if ;
      if file.mode = out_file then raise mode_error ; end if ;
      if key_type'size / 8 /= file.key_length(key_number) then
         raise key_error ;
      end if ;
      file.cur_key_no := key_number ;
      key_str := key_to_string (key) ;
      search (file, key_str, key_number, gelijk, groter) ;
      file.init := false ;
      if groter or (not gelijk and relation = equal) then
         raise existence_error ;
      end if ;
      if gelijk and relation = greater then
         begin
           ZOEK : loop
              read (file, item) ;
              strzone := elem_to_string (item) ;
              exit ZOEK when strzone (pos..pos+len-1) /= key_str ;
           end loop ZOEK ;
         exception
           when end_error => raise existence_error ;
         end ;
      end if ;
      beg := file.char_nr(file.key_levels(file.cur_key_no)) ;
      zone4 := file.block_acc(file.key_levels(file.cur_key_no)).all
                             (beg+len..beg+len+3) ;
      bloknr := string_to_int (zone4) / 512 ;
      char := string_to_int (zone4) rem 512 ;
      ptr := 1 ;
      TFT : loop
         dirio.read (file.data_file, blk, dirio.count(bloknr)) ;
         exit TFT when 512 - char >= file.size - ptr ;
         strzone (ptr..ptr+512-char) := blk (char..512) ;
         ptr := ptr + 513 - char ;
         char := 1 ;
         bloknr := bloknr + 1 ;
      end loop TFT ;
      strzone (ptr..file.size) := blk (char..char+file.size-ptr) ;
      item := string_to_elem (strzone) ;
   exception
      when others =>
         if file.opened then save (file) ; end if ;
         raise ;
   end read_by_key ;

   procedure erase (file :in out file_type ; key : in string ;
                    key_no : in integer) is
      len : integer := file.key_length (key_no) + 4 ;
      level, endpt : integer ;
      zone4 : str4 ;
      previous : string (1..len-4) ;
      blk : block := (others => ' ') ;
      equal, largest : boolean ;
   begin
      search(file, key, key_no, equal, largest) ;
      level := file.key_levels(key_no) ;
      LUS : loop
         endpt := get_endpt (file, level) ;
         file.block_acc(level).all(file.char_nr(level)..512-len) :=
            file.block_acc(level).all(file.char_nr(level)+len..512) ;
         endpt := endpt - len ;
         zone4 := int_to_string (endpt) ;
         file.block_acc(level).all(1..2) := zone4(1..2) ;
         dirio.write(file.key_file(key_no), file.block_acc(level).all,
                     dirio.count(file.block_nr(level))) ;
         if file.char_nr(level) < endpt then return ; end if ;
         exit LUS when endpt > 3 ;
         file.block_acc(level).all(1..4) :=
                  int_to_string(file.key_recup(key_no)) ;
         file.key_recup(key_no) := file.block_nr(level) ;
         dirio.write(file.key_file(key_no), file.block_acc(level).all,
                     dirio.count(file.block_nr(level))) ;
         if level = 1 then
            file.key_levels(key_no) := 1 ;
            return ;
         end if ;
         level := level - 1 ;
      end loop LUS ;
      previous := file.block_acc(level).all(endpt-len..endpt-5) ;
      search (file, previous, key_no, equal, largest) ;
      while level /= 1 loop
         level := level - 1 ;
         file.block_acc(level).all
                   (file.char_nr(level)..file.char_nr(level)+len-5) :=
                   previous ;
         dirio.write (file.key_file(key_no), file.block_acc(level).all,
                      dirio.count(file.block_nr(level))) ;
         endpt := get_endpt(file, level) ;
         if file.char_nr(level) + len /= endpt then return ; end if ;
      end loop ;
      LUS1 : while file.key_levels(key_no) > 1 loop
         endpt := get_endpt (file, level) ;
         exit LUS1 when endpt /= len + 3 ;
         dirio.write (file.key_file(key_no), file.block_acc(2).all, 1) ;
         file.key_levels(key_no) := file.key_levels(key_no) - 1 ;
         for i in 1 .. file.key_levels(key_no) loop
            file.block_acc(i).all := file.block_acc(i+1).all ;
            file.block_nr(i) := file.block_nr(i+1) ;
            file.char_nr(i) := file.char_nr(i+1) ;
         end loop ;
         blk(1..4) := int_to_string(file.key_recup(key_no)) ;
         file.key_recup(key_no) := file.block_nr(1) ;
         dirio.write(file.key_file(key_no), blk,
                     dirio.count(file.block_nr(1))) ;
         file.block_nr(1) := 1 ;
         --   block_acc(file.key_levels(key_no)+1) may be deallocated
      end loop LUS1 ;
   end erase ;

   procedure insert (file : in out file_type ; key : in string ;
                     key_no : in integer) is
      pos : integer := file.key_position(key_no) + 1 ;
      len : integer := file.key_length(key_no) + 4 ;
      aant, level, endpt, endpt2, bln, crn : integer ;
      subtype key_type is string (1 .. len) ;
      key1, key2 : key_type ;
      blk : block ;
      zone4 : str4 ;
      zone : string (1..510+len) ;
      two_keys, equal, largest : boolean ;
   begin
      key1 := key ;
      if file.num_rec = 0 then
         blk := (others => ' ') ;
         zone4 := int_to_string (len+3) ;
         blk(1..2) := zone4(1..2) ;
         blk(3..len+2) := key1 ;
         dirio.write(file.key_file(key_no), blk, 1) ;
         return ;
      end if ;
      search (file, key1 (1 .. len - 4), key_no, equal, largest) ;
      level := file.key_levels(key_no) ;
      two_keys := false ;
      LUSA : loop
         if largest then
            endpt := get_endpt(file, level) ;
            if endpt + len <= 513 then       -- er is nog plaats
               file.block_acc(level).all(endpt..endpt+len-1) := key1 ;
               endpt := endpt + len ;
               zone4 := int_to_string (endpt) ;
               file.block_acc(level).all(1..2) := zone4(1..2) ;
               dirio.write(file.key_file(key_no),
                           file.block_acc(level).all,
                           dirio.count(file.block_nr(level))) ;
               while level > 1 loop
                  level := level - 1 ;
                  endpt := get_endpt(file, level) ;
                  file.block_acc(level).all(endpt-len..endpt-5) :=
                        key1(1..len-4) ;
                  dirio.write(file.key_file(key_no),
                              file.block_acc(level).all,
                              dirio.count(file.block_nr(level))) ;
               end loop ;
               exit LUSA ;
            end if ;
                  -- er is geen plaats meer
            bln := file.key_recup(key_no) ;
            if bln = 0 then
               bln := integer(dirio.size(file.key_file(key_no))) + 1 ;
            else
               dirio.read(file.key_file(key_no), blk, dirio.count(bln)) ;
               zone4(1..4) := blk(1..4) ;
               file.key_recup(key_no) := string_to_int(zone4) ;
            end if ;
            blk := (others => ' ') ;
            zone4 := int_to_string(len+3) ;
            blk (1..2) := zone4(1..2) ;
            blk(3..len+2) := key1 ;
            dirio.write (file.key_file(key_no), blk, dirio.count(bln)) ;
            zone4 := int_to_string(bln) ;
            key1(len-3..len) := zone4 ;
            two_keys := false ;
            if level = 1 then
               key2 := key1 ;
               key1(1..len-4) :=
                           file.block_acc(level).all(endpt-len..endpt-5) ;
            end if ;
         else    -- not largest
            endpt := get_endpt(file, level) ;
            if endpt + len <= 513 then       -- er is nog plaats
               crn := file.char_nr(level) ;
               file.block_acc(level).all(crn+len..512) :=
                      file.block_acc(level).all(crn..512-len) ;
               file.block_acc(level).all(crn..crn+len-1) := key1 ;
               if two_keys then
                  file.block_acc(level).all(crn+len..crn+len*2-1) :=
                        key2 ;
               end if ;
               zone4 := int_to_string(endpt+len) ;
               file.block_acc(level).all(1..2) := zone4(1..2) ;
               dirio.write(file.key_file(key_no),
                           file.block_acc(level).all,
                           dirio.count(file.block_nr(level))) ;
               exit LUSA ;
            end if ;
                         -- er is geen plaats meer
            bln := file.key_recup(key_no) ;
            if bln = 0 then
               bln := integer(dirio.size(file.key_file(key_no))) + 1 ;
            else
               dirio.read(file.key_file(key_no), blk, dirio.count(bln)) ;
               zone4(1..4) := blk(1..4) ;
               file.key_recup(key_no) := string_to_int(zone4) ;
            end if ;
            aant := (510 + len) / (len * 2) ;
            crn := file.char_nr(level) ;
            zone(1..crn-3) := file.block_acc(level).all(3..crn-1) ;
            zone(crn-2..crn+len-3) := key1 ;
            if two_keys then
               zone(crn+len-2..510+len) := key2 &
                              file.block_acc(level).all(crn+len..512) ;
            else
               zone(crn+len-2..510+len) :=
                              file.block_acc(level).all(crn..512) ;
            end if ;
            endpt := endpt - (aant - 1) * len ;
            file.block_acc(level).all := (others => ' ') ;
            zone4 := int_to_string(endpt) ;
            file.block_acc(level).all(1..2) := zone4(1..2) ;
            file.block_acc(level).all(3..endpt-1) := zone(1..endpt-3) ;
            dirio.write(file.key_file(key_no),
                   file.block_acc(level).all,
                   dirio.count(file.block_nr(level))) ;
            blk := (others => ' ') ;
            endpt2 := aant * len + 3 ;
            zone4 := int_to_string (endpt2) ;
            blk(1..2) := zone4(1..2) ;
            blk(3..endpt2-1) := zone(endpt-2..endpt2+endpt-6) ;
            dirio.write(file.key_file(key_no), blk, dirio.count(bln)) ;
            zone4 := int_to_string(file.block_nr(level)) ;
            key1 := zone(endpt-len-2..endpt-7) & zone4 ;
            zone4 := int_to_string(bln) ;
            key2 := blk(endpt2-len..endpt2-5) & zone4 ;
            two_keys := true ;
         end if ;
         if level = 1 then
            bln := file.key_recup(key_no) ;
            if bln = 0 then
               bln := integer(dirio.size(file.key_file(key_no))) + 1 ;
            else
               dirio.read(file.key_file(key_no), blk, dirio.count(bln)) ;
               zone4(1..4) := blk(1..4) ;
               file.key_recup(key_no) := string_to_int(zone4) ;
            end if ;
            dirio.read(file.key_file(key_no), blk, 1) ;
            dirio.write(file.key_file(key_no), blk, dirio.count(bln)) ;
            blk := (others => ' ') ;
            zone4 := int_to_string(len * 2 + 3) ;
            blk (1..2) := zone4(1..2) ;
            zone4 := int_to_string(bln) ;
            key1(len-3..len) := zone4 ;
            blk(3..len+2) := key1 ;
            blk(len+3..len*2+2) := key2 ;
            dirio.write(file.key_file(key_no), blk, 1) ;
            file.key_levels(key_no) := file.key_levels(key_no) + 1 ;
            exit LUSA ;
         end if ;
         level := level - 1 ;
      end loop LUSA ;
   end insert ;

   procedure write (file : in out file_type ; item : in element_type) is
      subtype strelem  is string (1 .. file.size) ;
      function elem_to_str is new Ada.Unchecked_Conversion
                  (element_type, strelem) ;
      string_item : strelem ;
      pos : integer := file.key_position(0) + 1 ;
      len : integer := file.key_length(0) + 4 ;
      bln, crn, bltemp, rest, beg, begitem, endpt : integer ;
      neu, equal, largest : boolean ;
      blk, bidon : block ;
      zone4 : str4 ;
   begin
      if not file.opened then raise status_error ; end if ;
      if file.mode = in_file then raise mode_error ; end if ;
      if item'size / 8 /= file.size then raise use_error ; end if ;
      string_item := elem_to_str (item) ;
      if file.num_rec /= 0 then
         search (file, string_item (pos .. pos+len-5), 0, equal, largest) ;
         if equal then raise key_error ; end if ;
      end if ;
      crn := (file.data_recup - 1) rem 512 + 1 ;
      bln := (file.data_recup - crn) / 512 ;
      -- aanpassing van data_recup
      dirio.read (file.data_file, blk, dirio.count(bln)) ;
      if crn < 510 then        --  alles staat in blk
         zone4 := blk (crn..crn+3) ;
      else
         dirio.read (file.data_file, bidon, dirio.count(bln+1)) ;
         zone4 := blk (crn..512) & bidon (1..crn-509) ;
      end if ;
      file.data_recup := string_to_int (zone4) ;
      neu := file.data_recup = 0 ;
      if neu then
         file.data_recup := bln * 512 + crn + file.size ;
      end if ;
      rest := file.size ;   -- nog te schrijven karakters
      beg := crn ;          -- vanaf waar in te vullen
      bltemp := bln ;       -- nr van te schrijven blok
      begitem := 1 ;        -- vanaf waar te copieren
      while 513 - beg < rest loop
         blk (beg .. 512) := string_item (begitem .. begitem + 512 - beg) ;
         dirio.write (file.data_file, blk, dirio.count(bltemp)) ;
         bltemp := bltemp + 1 ;
         if not neu then
            dirio.read (file.data_file, blk, dirio.count(bltemp)) ;
         else
            blk := (others => ' ') ;
         end if ;
         rest := rest - 513 + beg ;
         begitem := begitem + 513 - beg ;
         beg := 1 ;
      end loop ;
      endpt := beg + file.size - begitem ;
      blk (beg .. endpt) := string_item (begitem .. file.size) ;
      if neu then
         zone4 := int_to_string (0) ;
         if endpt < 509 then
            blk (endpt+1..endpt+4) := zone4 ;
         else
            blk (endpt+1..512) := zone4 (1..512-endpt) ;
            bidon := (others => ' ') ;
            bidon (1..endpt-508) := zone4 (513-endpt..4) ;
            dirio.write (file.data_file, bidon, dirio.count(bltemp+1)) ;
         end if ;
      end if ;
      dirio.write (file.data_file, blk, dirio.count(bltemp)) ;
      zone4 := int_to_string (bln * 512 + crn) ;
      for key_no in 0 .. file.num_keys - 1 loop
         pos := file.key_position(key_no) + 1 ;
         len := file.key_length(key_no) + 4 ;
         insert (file, string_item(pos..pos+len-5)&zone4, key_no) ;
      end loop ;
      file.init := true ;
      file.num_rec := file.num_rec + 1 ;
   exception
      when others =>
         if file.opened then save (file) ; end if ;
         raise ;
   end write ;

   procedure update (file : in out file_type ; item : in element_type) is
      subtype strelem  is string (1 .. file.size) ;
      function elem_to_str is new Ada.Unchecked_Conversion
                  (element_type, strelem) ;
      old_item, string_item : strelem ;
      pos : integer := file.key_position(0) + 1 ;
      len : integer := file.key_length(0) + 4 ;
      level, bln, crn, bltemp, rest, beg, begitem, endpt : integer ;
      equal, largest : boolean ;
      blk : block ;
      zone4 : str4 ;
   begin
      if not file.opened then raise status_error ; end if ;
      if file.mode /= inout_file then raise mode_error ; end if ;
      if item'size / 8 /= file.size then raise use_error ; end if ;
      if file.num_rec = 0 then raise use_error ; end if ;
      string_item := elem_to_str (item) ;
      search (file, string_item (pos .. pos+len-5), 0, equal, largest) ;
      if not equal then raise existence_error ; end if ;
      level := file.key_levels(0) ;
      crn := file.char_nr(level) ;
      zone4 := file.block_acc(level).all(crn+len-4..crn+len-1) ;
      bltemp := string_to_int(zone4) ;
      crn := (bltemp - 1) rem 512 + 1 ;
      bln := (bltemp - crn) / 512 ;
      dirio.read(file.data_file, blk, dirio.count(bln)) ;
      rest := file.size ;
      beg := crn ;          -- vanaf waar in te vullen
      bltemp := bln ;       -- nr van te schrijven blok
      begitem := 1 ;        -- vanaf waar te copieren
      while 513 - beg < rest loop
         old_item(begitem..begitem+512-beg) := blk(beg..512) ;
         blk(beg..512) := string_item(begitem..begitem+512-beg) ;
         dirio.write (file.data_file, blk, dirio.count(bltemp)) ;
         bltemp := bltemp + 1 ;
         dirio.read (file.data_file, blk, dirio.count(bltemp)) ;
         rest := rest - 513 + beg ;
         begitem := begitem + 513 - beg ;
         beg := 1 ;
      end loop ;
      endpt := beg + file.size - begitem ;
      old_item(begitem..file.size) := blk(beg..endpt) ;
      blk(beg..endpt) := string_item(begitem..file.size) ;
      dirio.write (file.data_file, blk, dirio.count(bltemp)) ;
      zone4 := int_to_string(bln*512+crn) ;
      for key_no in 1 .. file.num_keys - 1 loop
         pos := file.key_position(key_no) + 1 ;
         len := file.key_length(key_no) + 4 ;
         if string_item(pos..pos+len-5) /= old_item(pos..pos+len-5) then
            erase(file, old_item(pos..pos+len-5), key_no) ;
            insert (file, string_item(pos..pos+len-5)&zone4, key_no) ;
         end if ;
      end loop ;
      file.init := true ;
   exception
      when others =>
         if file.opened then save (file) ; end if ;
         raise ;
   end update ;

   procedure delete_element (file : in out file_type) is
      level : integer := file.key_levels(file.cur_key_no) ;
      crn : integer := file.char_nr(level) ;
      len : integer := file.key_length(file.cur_key_no) + 4 ;
      pos, ptr, char, bloknr, bln, pointer : integer ;
      zone4 : string (1..4) ;
      blk : block ;
      strzone : string (1..file.size) ;
      gelijk, groter : boolean ;
   begin
      if not file.opened then raise status_error ; end if ;
      if file.mode = in_file then raise mode_error ; end if ;
      if file.init then raise use_error ; end if ;
      zone4 := file.block_acc(level).all(crn+len-4..crn+len-1) ;
      pointer := string_to_int(zone4) ;
      crn := pointer rem 512 ;
      bln := (pointer - crn) / 512 ;
      bloknr := bln ;
      char := crn ;
      ptr := 1 ;
      TFT : loop
         dirio.read (file.data_file, blk, dirio.count(bloknr)) ;
         exit TFT when 512 - char >= file.size - ptr ;
         strzone (ptr..ptr+512-char) := blk (char..512) ;
         ptr := ptr + 513 - char ;
         char := 1 ;
         bloknr := bloknr + 1 ;
      end loop TFT ;
      strzone (ptr..file.size) := blk (char..char+file.size-ptr) ;
      zone4 := int_to_string(file.data_recup) ;
      dirio.read(file.data_file, blk, dirio.count(bln)) ;
      if crn < 510 then
         blk(crn..crn+3) := zone4 ;
      else
         blk(crn..512) := zone4(1..513-crn) ;
         dirio.write(file.data_file, blk, dirio.count(bln)) ;
         bln := bln + 1 ;
         dirio.read(file.data_file, blk, dirio.count(bln)) ;
         blk(1..crn-509) := zone4(514-crn..4) ;
      end if ;
      dirio.write(file.data_file, blk, dirio.count(bln)) ;
      file.data_recup := pointer ;
      for key_no in 0 .. file.num_keys - 1 loop
         pos := file.key_position(key_no) + 1 ;
         len := file.key_length(key_no) ;
         search (file, strzone(pos..pos+len-1), key_no, gelijk, groter) ;
         erase(file, strzone(pos..pos+len-1), key_no) ;
      end loop ;
      file.init := true ;
   exception
      when others =>
         if file.opened then save (file) ; end if ;
         raise ;
   end delete_element ;

   function end_of_file (file : file_type) return boolean is
      zone4 : str4 ;
   begin
      if not file.opened then raise status_error ; end if ;
      if file.mode = out_file then raise mode_error ; end if ;
      if file.init then
         return file.num_rec = 0 ;
      end if ;
      for level in 1 .. file.key_levels(file.cur_key_no) loop
         zone4 := int_to_string (file.char_nr(level)+
                        file.key_length(file.cur_key_no) + 4) ;
         if zone4 (1..2) /= file.block_acc(level).all(1..2) then
            return false ;
         end if ;
      end loop ;
      return true ;
   exception
      when others =>
         if file.opened then save (file) ; end if ;
         raise ;
   end end_of_file ;

end indexed_io ;

------------------------------------------------------------------------
