with Dbms_Declarations, Dbms_Errors;

generic
   type Short_Numeric is range <>;
   type Long_Numeric is range <>;
   type Real_Type is digits <>;
   type Date_Type is private;

package Dbms_Template is

   subtype Dbms_Short_Numeric is Short_Numeric;
   subtype Dbms_Long_Numeric  is Long_Numeric;
   subtype Dbms_Real_Type is Real_Type;
   subtype Dbms_Date_Type is Date_Type;


   type Table is limited private;
   type Table_Record is limited private;
   type Record_Field is limited private;

   type Create_Fields is limited private;


   procedure Impart;
   procedure Depart;


   package Create_Fields_Builder is

      procedure Initialize
          (Fields           : out Create_Fields;
           Number_Of_Fields : in Dbms_Declarations.Dbms_Field_Count);

      procedure Set (Table_Fields  : in out Create_Fields;
                     To_Field_Name : in String;
                     To_Field_Type : in Dbms_Declarations.Field_Kind);

      procedure Dispose_Of (Fields : in out Create_Fields);

      Size_Error : exception;

   end Create_Fields_Builder;



   package Table_Operations is

      Table_Name_Length : constant := 75;

      subtype Table_Name is String (1 .. Table_Name_Length);
      subtype Record_Count is Long_Numeric;
	
      procedure Create (Table_Name   : in String;
                        Using_Fields : in Create_Fields);

      procedure Open (Table_Name   : in String;
                      Opened_Table : out Table;
                      Options : in Dbms_Declarations.Open_Options);
      procedure Close (The_Table : in out Table);

      function Table_Exists (Table_Name : in String) return Boolean;

      procedure Erase_Table (Table_Name : in String);
      procedure Delete_Table (Table_Name : in String);

      procedure Copy_Table (From_Table : in String;
                            To_Table   : in String);

      procedure Rename_Table (From_Table : in String;
                              To_Table   : in String);

      function Is_Table_Name (For_Table : in Table) return Table_Name;

      function Is_Number_Of_Records (In_Table : in Table)
         return Record_Count;

   end Table_Operations;



   package Record_Operations is

      subtype Record_Number is Long_Numeric;

      procedure Initialize_Table_Record (For_Table : in Table;
                                         Using     : out Table_Record);

      procedure Dispose_Of (The_Table_Record : in out Table_Record);

      procedure Erase (The_Table_Record : in out Table_Record);

      procedure Goto_First_Record (In_Table : in out Table);

      procedure Goto_Next_Record (In_Table : in out Table);

      procedure Goto_Previous_Record (In_Table : in out Table);

      procedure Goto_Last_Record (In_Table : in out Table);

      procedure Get_Current_Record (From_Table : in Table;
                                    Into       : in out Table_Record);

      procedure Append (The_Record : in Table_Record;
                        To         : in out Table);

      procedure Update_Current_Record (In_Table    : in out Table;
                                       With_Record : in Table_Record);

      procedure Delete_Current_Record (From_Table : in Table);

      procedure Goto_Record (Rec_Num  : in Record_Number;
                             In_Table : in Table);

      procedure Insert (Into_Table : in Table;
                        The_Record : in Table_Record);

      function Is_Current_Record_Number (In_Table : in Table)
         return Record_Number;

   end Record_Operations;



   package Field_Operations is

      Field_Name_Length : constant := 25;
      subtype Field_Name is String (1 .. Field_Name_Length);

      procedure Initialize_Record_Field
          (In_Table         : in Table;
           Using_Field_Name : in String;
           The_Record_Field : out Record_Field);

      procedure Erase (The_Record_Field : in out Record_Field;
                       In_Record        : in Table_Record);

      procedure Put_Short (The_Value  : in Dbms_Short_Numeric;
                           Into_Field : in out Record_Field;
                           Of_Record  : in Table_Record);

      procedure Put_Long (The_Value  : in Dbms_Long_Numeric;
                          Into_Field : in out Record_Field;
                          Of_Record  : in Table_Record);

      procedure Put_String (The_Value  : in String;
                            Into_Field : in out Record_Field;
                            Of_Record  : in Table_Record);

      procedure Put_Real (The_Value  : in Dbms_Real_Type;
                          Into_Field : in out Record_Field;
                          Of_Record  : in Table_Record);

      procedure Put_Date (The_Value  : in Dbms_Date_Type;
                          Into_Field : in out Record_Field;
                          Of_Record  : in Table_Record);

      procedure Get_Short (The_Value  : out Dbms_Short_Numeric;
                           From_Field : in  out Record_Field;
                           Of_Record  : in  Table_Record);

      procedure Get_Long (The_Value  : out Dbms_Long_Numeric;
                          From_Field : in  out Record_Field;
                          Of_Record  : in  Table_Record);

      procedure Get_String (The_Value  : out String;
                            From_Field : in  out Record_Field;
                            Of_Record  : in  Table_Record);

      procedure Get_Real (The_Value  : out Dbms_Real_Type;
                          From_Field : in  out Record_Field;
                          Of_Record  : in  Table_Record);

      procedure Get_Date (The_Value  : out Dbms_Date_Type;
                          From_Field : in  out Record_Field;
                          Of_Record  : in  Table_Record);

      function Is_Number_Of_Fields (In_Table : in Table)
         return Dbms_Declarations.Dbms_Field_Count;

      function Is_Field_Blank (The_Field : in Record_Field;
                               In_Record : in Table_Record)
         return Boolean;

      function Is_Field_Name (For_Field : in Record_Field;
                              In_Table : in Table)
         return Field_Name;

   end Field_Operations;



   package Index_Operations is

      subtype Index_Fields_Count is Natural;

      procedure Create_Index
          (On_Table               : in String;
           Kind                   : in Dbms_Declarations.Index_Kind;
           Number_Of_Index_Fields : in Dbms_Declarations.
              Dbms_Field_Count);

      procedure Delete_Index (Index_Name : in String;
                              ID         : in Dbms_Declarations.Index_ID);

      function Is_Number_Of_Index_Fields (For_Table : in Table)
                  return Index_Fields_Count;

   end Index_Operations;



   package Search_Operations is

      procedure Search_On_Index
          (For_Table            : in Table;
           Using                : in Table_Record;
           Number_Of_Key_Fields : in Dbms_Declarations.
              Dbms_Field_Count;
           Kind : in Dbms_Declarations.Search_Kind);

      procedure Search_On_Field
          (For_Table   : in Table;
           Using       : in Table_Record;
           Using_Value : in Record_Field;
           Kind        : in Dbms_Declarations.Search_Kind);

   end Search_Operations;


   function Is_Error return Dbms_Errors.Error_Kind;


   Dbms_Error        : exception;
   Operation_Failure : exception;



private

--   type Unsigned is  range 0 .. 65_535;
--   for Unsigned'SIZE use 16;

   type Unsigned is new Integer;

   type Table is new Unsigned;
   type Table_Record is new Unsigned;
   type Record_Field is new Unsigned;

   type Create_Fields_Structure;
   type Create_Fields is access Create_Fields_Structure;

end Dbms_Template;

