with Dbms_Declarations, Dbms, Dbms_Errors;
use Dbms;

package body Parts_Inventory_Database is

   Table_Name : constant String := "Parts";

   Parts_Inventory        : Table;
   Parts_Inventory_Record : Table_Record;

   Part_Number_Field      : Record_Field;
   Customer_Number_Field  : Record_Field;
   Quantity_Field         : Record_Field;
   Transaction_Date_Field : Record_Field;

   procedure Initialize_Data_Areas;
   procedure Release_Data_Areas;

   function "="
       (Left, Right : in Dbms_Errors.Error_Kind) return Boolean
       renames Dbms_Errors."=";

   use
      Table_Operations,
      Record_Operations,
      Field_Operations,
      Index_Operations,
      Search_Operations;

   procedure Append (Using : in Data_Record_Description) is
      D : Dbms_Date.Date_Declaration.Date renames Using.The_Date;
      use Dbms_Date, Date_Conversion;
   begin
      Put_String (The_Value  => String (Using.The_Part_Number),
                  Into_Field => Part_Number_Field,
                  Of_Record  => Parts_Inventory_Record);
      Put_String (The_Value  => String (Using.The_Customer_Number),
                  Into_Field => Customer_Number_Field,
                  Of_Record  => Parts_Inventory_Record);
      Put_Long
          (The_Value  => Dbms.Dbms_Long_Numeric (Using.The_Quantity),
           Into_Field => Quantity_Field,
           Of_Record  => Parts_Inventory_Record);
      Put_Date
          (The_Value => Convert_To_Date (D.The_Month, D.The_Day,
                                         D.The_Year),
           Into_Field => Transaction_Date_Field,
           Of_Record  => Parts_Inventory_Record);
      Append (The_Record => Parts_Inventory_Record,
              To         => Parts_Inventory);
   end Append;

   procedure Close is
   begin
      Release_Data_Areas;
      Close (Parts_Inventory);
   end Close;

   procedure Find_Inventory_Record
       (For_Part_Number     : in Part_Number;
        For_Customer_Number : in Customer_Number) is
      use Dbms, Search_Operations, Record_Operations;
   begin
      Put_String (The_Value  => String (For_Part_Number),
                  Into_Field => Part_Number_Field,
                  Of_Record  => Parts_Inventory_Record);
      Put_String (The_Value  => String (For_Customer_Number),
                  Into_Field => Customer_Number_Field,
                  Of_Record  => Parts_Inventory_Record);
      Search_On_Index (For_Table            => Parts_Inventory,
                       Using => Parts_Inventory_Record,
                       Number_Of_Key_Fields => 2,
                       Kind => Dbms_Declarations.First);
      Get_Current_Record (From_Table => Parts_Inventory,
                          Into       => Parts_Inventory_Record);
   end Find_Inventory_Record;

   procedure Open is
      Table_Options : Dbms_Declarations.Open_Options :=
         (Save_Every_Change => True, Id => 0);
   begin
      if not Table_Exists (Table_Name) then
         declare
            Table_Fields : Create_Fields;
            use Dbms_Declarations, Create_Fields_Builder;
         begin
            Initialize (Table_Fields, 4);
            Set (Table_Fields, "Part Number", (Alphanumeric, 8));
            Set (Table_Fields, "Cust Number", (Alphanumeric, 8));
            Set (Table_Fields, "Quantity", (Kind => Long));
            Set (Table_Fields, "Date", (Kind => Date));
            Create (Table_Name, Table_Fields);
            Create_Index (Table_Name, Primary, 2);
            Dispose_Of (Table_Fields);
         end;
      end if;
      Open (Table_Name, Parts_Inventory, Table_Options);
      Initialize_Data_Areas;
   end Open;

   procedure Initialize_Data_Areas is
   begin
      Initialize_Table_Record (For_Table => Parts_Inventory,
                               Using     => Parts_Inventory_Record);
      Initialize_Record_Field (In_Table         => Parts_Inventory,
                               Using_Field_Name => "Part Number",
                               The_Record_Field => Part_Number_Field);
      Initialize_Record_Field
          (In_Table         => Parts_Inventory,
           Using_Field_Name => "Cust Number",
           The_Record_Field => Customer_Number_Field);
      Initialize_Record_Field (In_Table         => Parts_Inventory,
                               Using_Field_Name => "Quantity",
                               The_Record_Field => Quantity_Field);
      Initialize_Record_Field
          (In_Table         => Parts_Inventory,
           Using_Field_Name => "Date",
           The_Record_Field => Transaction_Date_Field);
   end Initialize_Data_Areas;

   function Is_Current_Record return Data_Record_Description is
      T : Data_Record_Description;
   begin
      return
         Data_Record_Description'
            (Is_The_Part_Number, Is_The_Customer_Number,
             Is_The_Quantity, Is_The_Date);
   end Is_Current_Record;

   function Is_The_Part_Number return Part_Number is
      Part_Num : Part_Number;
   begin
      Dbms.Field_Operations.Get_String
          (The_Value  => String (Part_Num),
           From_Field => Part_Number_Field,
           Of_Record  => Parts_Inventory_Record);
      return Part_Num;
   end Is_The_Part_Number;

   function Is_The_Customer_Number return Customer_Number is
      Customer_Num : Customer_Number;
   begin
      Dbms.Field_Operations.Get_String
          (The_Value  => String (Customer_Num),
           From_Field => Customer_Number_Field,
           Of_Record  => Parts_Inventory_Record);
      return Customer_Num;
   end Is_The_Customer_Number;

   function Is_The_Quantity return Quantity is
      Q : Dbms_Long_Numeric;
   begin
      Dbms.Field_Operations.Get_Long
          (The_Value  => Q,
           From_Field => Quantity_Field,
           Of_Record  => Parts_Inventory_Record);
      return Quantity (Q);
   end Is_The_Quantity;

   function Is_The_Date return Dbms_Date.Date_Declaration.Date is
      use Dbms_Date;
      L : Long_Integer;
      M : Month;
      D : Day;
      Y : Year;
   begin
      Dbms.Field_Operations.Get_Date
          (The_Value  => L,
           From_Field => Transaction_Date_Field,
           Of_Record  => Parts_Inventory_Record);
      Date_Conversion.Split (L, M, D, Y);
      return Dbms_Date.Date_Declaration.Date'(M, D, Y);
   end Is_The_Date;

   procedure Release_Data_Areas is
   begin
      Dispose_Of (The_Table_Record => Parts_Inventory_Record);
   end Release_Data_Areas;

   function Match_All return Boolean is
   begin
      return True;
   end Match_All;


   package body Customers_To_Part_Query_Template is

      The_Part_Number : Part_Number := (others => ' ');
      Find_First      : Boolean     := False;

      procedure Find_Next is
         Search_Method : Dbms_Declarations.Search_Kind;
      begin
      Find_Record_Loop:
         loop
            if Find_First then
               Search_Method := Dbms_Declarations.First;
               Find_First    := False;
            else
               Search_Method := Dbms_Declarations.Next;
            end if;
            Search_On_Index (For_Table            => Parts_Inventory,
                             Using => Parts_Inventory_Record,
                             Number_Of_Key_Fields => 1,
                             Kind                 => Search_Method);
            Get_Current_Record (From_Table => Parts_Inventory,
                                Into       => Parts_Inventory_Record);
            exit Find_Record_Loop when Is_Match;
         end loop Find_Record_Loop;
      exception
         when Dbms_Error =>
            if Is_Error = Dbms_Errors.Record_Not_Found then
               raise End_Of_Data;
            else
               raise Program_Error;
            end if;
      end Find_Next;

      procedure Initialize_Query (For_Part : in Part_Number) is
      begin
         The_Part_Number := For_Part;
         Find_First      := True;
      end Initialize_Query;

   end Customers_To_Part_Query_Template;


   package body Sequential_Search_Template is

      procedure Find_First is
      begin
         null;
      end Find_First;
      procedure Find_Next is
      begin
         null;
      end Find_Next;

   end Sequential_Search_Template;


end Parts_Inventory_Database;
