
-------------------------------------------------------------------------------
--                                                                           --
--  Component
--
--  $Id: read_file.adb,v 1.5 1998/12/30 19:34:26 obry Exp $
--
--  Author    : Pascal Obry
--  e-mail    : pascal_obry@csi.com
--
--  Machine/System Compiled/Run on : SPARC SUN-OS, Windows NT -GNAT
--
--                                                                           --
-------------------------------------------------------------------------------


with Ada.Unchecked_Deallocation;
with Ada.Text_IO;
with Ada.Strings.Maps;
with Ada.Strings.Fixed;

package body Read_File is

   use Ada.Text_IO;
   use Ada.Strings;

   type String_Pointer is access String;
   type Character_Set_Pointer is access Maps.Character_Set;

   type Field_Record is
      record
         Content   : String_Pointer;
         Separator : Character;
      end record;

   Default_Separators : constant String := ' ' & ASCII.HT;

   Input_Buffer   : String (1 .. Max_Line_Lenght);
   Input_Filename : String_Pointer;

   File           : aliased File_Type;
   Input_File     : File_Access := File'Unchecked_Access;
   --  Unchecked_Access is used there because we can't have a dandling
   --  pointer. File and Input_File are at the same level and they are
   --  declared (and use) only in the package body.

   Separators   : Character_Set_Pointer
                := new Maps.Character_Set'(Maps.To_Set (Default_Separators));

   Fields       : array (1 .. Max_Field) of Field_Record;
   NF           : Natural;
   LN           : Natural;

   ----------
   -- Open --
   ----------

   procedure Open (Name : in String) is
   begin
      if Name = "-" then
         Input_File := Ada.Text_IO.Standard_Input;
      else
         Open (File => File, Name => Name, Mode => In_File);
      end if;
      Input_Filename := new String'(Name);
      LN := 0;
      NF := 0;
   end Open;

   -----------
   -- Close --
   -----------

   procedure Close is
   begin
      if Input_Filename.all /= "-" then
         Close (File);
      end if;
   end Close;

   --------------------
   -- Set_Separators --
   --------------------

   procedure Set_Separators (To : in String) is

      procedure Free is new Ada.Unchecked_Deallocation
        (Maps.Character_Set, Character_Set_Pointer);

   begin
      Free (Separators);
      Separators := new Maps.Character_Set'(Maps.To_Set (To));
   end Set_Separators;

   -----------
   -- Fetch --
   -----------

   procedure Fetch is

      -----------------------
      -- Free_Current_Line --
      -----------------------

      procedure Free_Current_Line is

         procedure Free is new Ada.Unchecked_Deallocation
           (String, String_Pointer);

      begin
         for I in 1 .. NF loop
            Free (Fields (I).Content);
         end loop;
      end Free_Current_Line;

      Start, Stop : Natural;
      Line_Length : Natural;

   begin -- Fetch
      Free_Current_Line;

      Get_Line (File => Input_File.all,
                Item => Input_Buffer,
                Last => Line_Length);

      NF    := 0;
      Start := Input_Buffer'First;

      loop
         Stop := Fixed.Index
           (Source => Input_Buffer (Start .. Line_Length),
            Set    => Separators.all);
         exit when Stop = 0;

         NF := NF + 1;
         Fields (NF) :=
           (Content   => new String'(Input_Buffer (Start .. Stop - 1)),
            Separator => Input_Buffer (Stop));

         Start := Stop + 1;
      end loop;

      --  last field
      NF := NF + 1;
      if Start <= Line_Length then
         Fields (NF) :=
           (Content   => new String'(Input_Buffer (Start .. Line_Length)),
            Separator => ASCII.LF);
      else
         Fields (NF) := (Content   => new String'(""),
                         Separator => ASCII.LF);
      end if;

      LN := LN + 1;

   end Fetch;

   -----------------
   -- End_OF_File --
   -----------------

   function End_Of_File
     return Boolean is
   begin
      return Ada.Text_IO.End_Of_File (Input_File.all);
   end End_Of_File;

   -------------------
   -- Is_Empty_Line --
   -------------------

   function Is_Empty_Line
     return Boolean is
   begin
      return NF = 1 and then Fields (1).Content'Length = 0;
   end Is_Empty_Line;

   ----------------------
   -- Number_Of_Fields --
   ----------------------

   function Number_Of_Fields
     return Natural is
   begin
      return NF;
   end Number_Of_Fields;

   -----------------
   -- Line_Number --
   -----------------

   function Line_Number
     return Natural is
   begin
      return LN;
   end Line_Number;

   -------------------------
   -- Separator_For_Field --
   -------------------------

   function Separator_For_Field (N : in Field_Range)
      return Character is
   begin
      return Fields (N).Separator;
   end Separator_For_Field;

   -----------
   -- Field --
   -----------

   function Field (N : in Field_Range)
      return String is
   begin
      return Fields (N).Content.all;
   end Field;

   -----------
   -- Field --
   -----------

   function Field (N : in Field_Range)
      return Integer is

      package Int_IO is new Integer_IO (Integer);
      Result : Integer;
      Last   : Positive;

   begin
      Int_IO.Get (From => Fields (N).Content.all,
                  Item => Result,
                  Last => Last);
      return Result;
   end Field;

   -----------
   -- Field --
   -----------

   function Field (N : in Field_Range)
      return Float is

      package Flt_IO is new Float_IO (Float);
      Result : Float;
      Last   : Positive;

   begin
      Flt_IO.Get (From => Fields (N).Content.all,
                  Item => Result,
                  Last => Last);
      return Result;
   end Field;

   -------------------
   -- Field_Of_Type --
   -------------------

   function Field_Of_Type (N : in Field_Range)
      return T is

      package Enm_IO is new Enumeration_IO (T);
      Result : T;
      Last   : Positive;

   begin
      Enm_IO.Get (From => Fields (N).Content.all,
                  Item => Result,
                  Last => Last);
      return Result;
   end Field_Of_Type;

end Read_File;
